/* sa_head.h  include file; global variables for standalone CirclePack */

#ifndef SA_HEAD_H
#define SA_HEAD_H

#include "sa_types.h"
#include "sa_proto.h"
#include "sa_hooks.h"
#include "sys/types.h"
#include "socketfun.h"
#define NUM_PACKS 3		/* number of packs and canvases allowed */
#define DISP_GRP 250            /* size to break up long display lists */
#include "StandAlone_proto.h"

#define IN_OUT_PORTNUM 33333    /* port number for in_out_fp socket */ 
#define DATA_PORTNUM 44444      /* port number for display data socket */ 

/* ================ color stuff (eventually depends on GUI) === */

extern unsigned long *colors,*color; /* color table and ptr to my colors. */

/* ============= global variables for CirclePack ===================== */
/* (see cp_types.h for global variables needed by the cp library) */

extern IOBUF *msg_out_fp;       /* stdout or msg socket mimicking stdout */
extern IOBUF *msg_in_fp;        /* stdin or socket for stdin */
extern IOBUF *client_to[NUM_PROC]; /* to remote process */
extern IOBUF *client_from[NUM_PROC];  /* from remote process */

extern FILE *post_fp;		/* file for canvas postscript output */
extern FILE *custom_fp;	/* file for "custom" postscript output */
extern FILE *log_fp;           /* file for log in cmd mode */

extern struct hostent *host;    /* host on which C code is running */

extern char CPVersion[64];     /* CirclePack Version number */
extern char CPHelp[256];       /* location of help file */
extern char rgb_file[NAME_MAX]; /* name of file of rbg colors */
extern char user_name[64];     /* store for use in postscript output */
extern char packing_dir_name[NAME_MAX]; /* directory for packings */
extern char print_cmd[NAME_MAX]; /* cmd for printing */
extern char ps_viewer[NAME_MAX]; /* post script popup viewer command */
extern char ps_file_name[NAME_MAX]; /* custom postscript output */
extern char print_file_name[NAME_MAX]; /* printing postscript file */
extern char *home_dir;		/* store user home directory */
extern char *working_dir;	/* store current working directory */
extern char *ftn_keys[12];     /* cmd strings for ftn keys 1--10 (see packrc.txt) */
extern char cmd_history[BUFSIZE]; /* stores last command for reuse */
// extern char path[NAME_MAX]; 	/* path name for packing directory */
extern char *emsgbuf,*msgbuf;	/* error message/message buffers (overlap) */
extern char file_name[NAME_MAX];

extern int cmd_mode;           /* set if in command-line mode */
extern int current_p;		/* index to current screen and pack */
extern int totalpasses;	/* number of passes in riffle */
extern int cmd_search_depth;	/* prevent looping of cmd search */
extern int draw_speed;		/* 0 = disp screen after all objects drawn */
extern int live_node[NUM_PACKS]; /* last circle plotted in 'projection' mode */
extern int live_face[NUM_PACKS]; /* last face plotted in 'projection' mode */
extern int init_shmem;         /* flag: shared memory initialized? */

extern double inc_factor,eucl_factor; /* default adjustment factors */

extern struct Edgelist *b_flip,*r_flip; /* ptrs to edges for Whitehead moves */

extern Mobius Mob;		/* currently specified Mobius transform */
extern int first_face_global;
extern int first_red_face_global;
extern int num_bdry_comp_global;
extern int length_global;
extern int i_bdry_thing_global;
extern char topology_global[100];
extern double area_global;
extern int v_count_global;
extern int f_count_global;
extern int e_count_global;
extern int V_count_global;
extern int F_count_global;
extern int E_count_global;
extern double inv_dist_global;
extern int max_ideg_global;
extern int min_ideg_global;
extern int max_bdeg_global;
extern int min_bdeg_global;
extern int last_index_global;
#endif
