#include "sa_types.h"
#include "sa_proto.h"

/* Read face indices for pack p, put in linked list. Return NULL
on error; also, endptr returns NULL or pointer to inappropriate 
char, if such is encountered. hits=0 ==> no expressions parsed.

Codes:
   a or a(f1 f2)    all faces or all from f1 to f2
   b or b(f1 f2)    all boundary faces (at least one bdry vert)
   B or B(f1 f2)    boundary faces of redchain, in order, f1 to f2
   r                redfaces, in order
   w                white faces in drawing order
   flist or Flist   from recorded face list
   m (or mc) [pn]   marked (or unmarked) faces [from pack n]
   nan              bad corner data.
   p (or pc)        plot_flag >0 (or <=0).
   {..}             set-builder notation
   <f..>            list of indices (stop with any improper index)
   z <x y>          faces containing point z
*/ 

struct Vertlist *Face_link_parse(struct p_data *p,char *dpoint,
				 char **endptr,int *hits,
				 struct Vertlist **Vlist,
				 struct Edgelist **Elist,
				 struct Vertlist **Flist,
				 CPRegion *region,struct Pathlist *pathlist,
				 int pathlength)
{
  int count,v1,v2,n,m,flag,i,cflg,localcount,okay,f1,f2,indx,j;
  int lastface,q,qcount=0,q_flag,cf_flag=0,wflag,face,hts,shts,*u_flag;
  char next[256],next2[256],*nextptr,*lastptr,phold[1],*eptr;
  struct Vertlist *facelist=NULL,*trace,*clobber,*tmp_ptr,*vtrace;
  struct Vertlist *vlptr=NULL,*cfptr=NULL,*vertlist=NULL;
  struct Vertlist *whole_list=NULL;
  struct RedList *ftrace;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;
  f_data *qfaces=NULL;

  nextptr=lastptr=dpoint;
  *endptr=NULL;
  *hits=count=0;
  if (*nextptr == '\0' || !grab_next(&nextptr,next)) return NULL;
  facelist=(struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
  trace=facelist;
  clobber=trace;
  do
    {
      lastptr += notspace(lastptr);
      if (*next=='a') /* all */
	{
	  lastptr++;
	  lastptr=lastptr+notspace(lastptr);
	  if (*lastptr!='(') /* run thru whole complex */
	    {v1=1;v2=p->facecount;}
	  else if (!paren_parse(&lastptr,&v1,&v2) || v2<v1)
	    /* malformed expression */
	    goto F_GETOUT;
	  if (v1<1) v1=1;
	  if (v2>p->facecount) v2=p->facecount;
	  if (lastptr>nextptr) nextptr=lastptr;
	  /* now should have valid v1 and v2 */
	  for (n=v1;n<=v2;n++)
	    {
	      trace->v=n;
	      trace->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      clobber=trace;
	      trace=trace->next;
	      count++;
	    }
	} /* finished with 'all' */
      else if (*next=='b') /* boundary */
	{
	  lastptr++;
	  lastptr=lastptr+notspace(lastptr);
	  if (*lastptr!='(') /* run thru whole complex */
	    {v1=1;v2=p->facecount;}
	  else if (!paren_parse(&lastptr,&v1,&v2) || v1>v2)
	    /* malformed expression */
	    goto F_GETOUT;
	  if (v1<1) v1=1;
	  if (v2>p->facecount) v2=p->facecount;
	  for (n=v1;n<=v2;n++)	
	    {
	      flag=0;
	      for (i=0;i<3;i++) 
		if (pK_ptr[p->faces[n].vert[i]].bdry_flag)
		  flag++;
	      if (flag)
		{
		  trace->v=n;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		}
	    }
	} /* finished with bdry */
       else if (*next=='B') /* chains along boundary */
	{
	  lastptr++;
	  lastptr=lastptr+notspace(lastptr);
	  if (*lastptr!='(') /* run thru redchain faces on bdry */
	    {
	      if ((ftrace=p->redfaces)) 
		{
		  wflag=0;
		  while (ftrace!=p->redfaces || !(wflag++))
		    {
		      flag=0;
		      for (i=0;i<3;i++) 
			if (pK_ptr[p->faces[ftrace->face].
					vert[i]].bdry_flag)
			  flag++;
		      if (flag) /* yes, face has a bdry vert */
			{
			  trace->v=ftrace->face;
			  trace->next=(struct Vertlist *)
			    calloc((size_t)1,sizeof(struct Vertlist));
			  clobber=trace;
			  trace=trace->next;
			  count++;
			}
		      ftrace=ftrace->next;
		    }
		}
	    }
	  else 
	    {
	      /* check for bad data: malformed or non-bdry faces. */
	      if (!paren_parse(&lastptr,&f1,&f2) || f1<1 || f2<1
		  || f1>p->facecount || f2>p->facecount)
		goto F_GETOUT;
	      flag=0;
	      for (i=0;i<3;i++) 
		if (pK_ptr[p->faces[f2].vert[i]].bdry_flag)
		  flag++;
	      if (!flag) goto F_GETOUT; /* f2 has no bdry vert */

	      /* now, follow contig chain of boundary faces starting
		 at f1 until you hit f2 or return to f1 */

	      indx=0;
	      while (!pK_ptr[p->faces[f1].vert[indx]].bdry_flag && indx<3)
		indx++;
	      if (indx==3) goto F_GETOUT; /* f1 has no bdry vert */
	      if (pK_ptr[p->faces[f1].vert[(indx+1)%3]].bdry_flag)
		indx=(indx+1)%3; /* want furthest bdry vert */

	      lastface=f1;
	      okay=1; /* flag whether to save the last face */
	      wflag=0;
	      localcount=0;
	      /* entering while, indx should always point to furthest
		 counterclockwise bdry vert */
	      while ((lastface!=f2 && lastface!=f1) || !(wflag++))
		{
		  localcount++; /* emergency exit */
		  trace->v=lastface;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		  v1=p->faces[lastface].vert[indx];
		  v2=p->faces[lastface].vert[(indx+1)%3];
		  if ((lastface=abutting_face(p,v1,v2))<0)
		    {
		      lastface=f2; /* get out of loop */
		      okay=0;
		    }
		  indx=0;
		  while (!pK_ptr[p->faces[lastface].vert[indx]].bdry_flag 
			 && indx<3)
		    indx++;
		  if (indx==3 || localcount++>2*(p->facecount))
		    {
		      lastface=f2; /* shouldn't happen, get out of loop */
		      okay=0;
		    }
		  else 
		    if (pK_ptr[p->faces[lastface].vert[(indx+1)%3]].bdry_flag)
		      indx=(indx+1)%3; /* want furthest bdry vert */
		} /* end of while */
	      if (okay) /* yes, things went well, save the lastface */
		{
		  trace->v=lastface;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		}
	    } /* end of else */
	} /* finished with bdry chains */
      else if (*next=='I')
	/* faces incident to faces or vertices */
	{
	  if (next[1]=='v'    /* to vertices */
	      && (vertlist=Node_link_parse(p,lastptr+2,endptr,&hts,Vlist,
					  Elist,Flist,region,pathlist,
					  pathlength)))
	    {
	      nextptr=*endptr;
	      tmp_ptr=vertlist;
	      u_flag=(int *)calloc((size_t)(p->nodecount+1),sizeof(int));
	      while (tmp_ptr)
		{
		  u_flag[tmp_ptr->v]=1;
		  tmp_ptr=tmp_ptr->next;
		}
	      vert_free(&vertlist);
	      for (i=1;i<=p->facecount;i++)
		if (u_flag[p->faces[i].vert[0]] 
		    || u_flag[p->faces[i].vert[1]] || 
		    u_flag[p->faces[i].vert[2]])
		  {
		    trace->v=i;
		    trace->next=(struct Vertlist *)
		      calloc((size_t)1,sizeof(struct Vertlist));
		    clobber=trace;
		    trace=trace->next;
		    count++;
		  }
	      free(u_flag);
	    }
	  /* fixup: write code for "If": faces incident to given faces */
	}
      else if (*next=='r' && p->redfaces)    
	/* red faces using double linked list */
	{
	  localcount=0;
	  lastptr++;
	  wflag=0;
	  ftrace=p->redfaces;
	  while ((ftrace!=p->redfaces || !(wflag++)) 
		 && localcount < 3*p->facecount)
	    {
	      trace->v=ftrace->face;
	      trace->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      clobber=trace;
	      trace=trace->next;
	      localcount++;
	      ftrace=ftrace->next;
	    }
	  count += localcount;
	}
      else if (*next=='r' && (face=p->first_red_face))
	/* red faces using f_data list (e.g., simply connected case) */
	{
	  localcount=0;
	  lastptr++;
	  wflag=0;
	  while (((face!=p->first_red_face) || !(wflag++)) 
		 && localcount < 3*p->facecount)
	    {
	      trace->v=face;
	      trace->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      clobber=trace;
	      trace=trace->next;
	      localcount++;
	      face=p->faces[face].next_red;
	    }
	  count += localcount;
	}	  
      else if (*next=='w' && p->rwb_flags) /* white faces in draw order, 
					      as possible.*/
	{
	  lastptr++;
	  lastptr=lastptr+notspace(lastptr);
	  if (*lastptr!='(') /* run thru whole complex */
	    {v1=1;v2=p->facecount;}
	  else if (!paren_parse(&lastptr,&v1,&v2) || v1>v2)
	    /* malformed expression */
	    goto F_GETOUT;
	  if (v1<1) v1=1;
	  if (v2>p->facecount) v2=p->facecount;
	  n=m=p->first_face;
	  do
	    {
	      if (n>=v1 && n<=v2 && p->rwb_flags[n]<0)
		{
		  trace->v=n;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		}
	      n=p->faces[n].next_face;
	    }
	  while (n!=m);
	}  
      else if (strncmp(next+1,"list",4)==0) /* use flist */
	{
	  if (((*next=='F' && (cfptr=(*Flist) )) 
	       || (*next=='f' && (cfptr=p->flist)))
	      && *(next+5)=='[')
	    {
	      if (*(next+6)=='r') /* rotate list */
		{
		  vlptr=cfptr;
		  while (vlptr && vlptr->next) vlptr=vlptr->next;
		  vlptr->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  vlptr->next->v=cfptr->v;
		}
	      if (*(next+6)=='n' || *(next+6)=='r')
		{
		  v1=consume_one_index(&cfptr);
		  if (*next=='F') (*Flist)=cfptr;
		  else p->flist=cfptr;
		}
	      else if (sscanf(next+5,"[%d]",&n) && n>=0)
		{
		  vlptr=cfptr;
		  while (n-- > 1 && vlptr->next)
		    vlptr=vlptr->next;
		  if (n==0 && vlptr) v1=vlptr->v;
		  else v1=0;
		}
	      else v1=0;
	      if (v1>0 && v1<=p->facecount ) 
		{
		  trace->v=v1;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		}	
	      vlptr=(struct Vertlist *)NULL;
	    }
	  else if (*next=='F') vlptr=(*Flist); /* adjoint Flist */
	  else if (*next=='f') vlptr=p->flist; /* adjoint pack flist */
	  while (vlptr && (n=vlptr->v)>0 && n<=p->facecount)
	    {
	      trace->v=n;
	      trace->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      clobber=trace;
	      trace=trace->next;
	      vlptr=vlptr->next;
	      count++;
	    }
	}
      else if (*next=='m') /* marked */
	{
	  if ((cflg=(next[1]=='c'))) lastptr++; 
				/* 1==>complement of marked */
	  if (next[1]=='p' || (cflg && next[2]=='p') )
				/* take marking from pack q */
	    {
	      lastptr++;
	      if (next[1]=='p') phold[0]=next[2];
	      else phold[0]=next[3];
	      q=atoi(phold);
	      if (q>=0 && packdata[q].status)
		{
		  lastptr++;
		  q_flag=1;
		  qfaces=packdata[q].faces;
		  qcount=packdata[q].facecount;
		}
	      else q_flag=0;
	    }				
	  else q_flag=0;				
	  lastptr++;
	  lastptr=lastptr+notspace(lastptr);
	  if (*lastptr!='(') /* run thru whole complex */
	    {v1=1;v2=p->facecount;}
	  else if (!paren_parse(&lastptr,&v1,&v2) || v1>v2)
	    /* malformed expression */
	    goto F_GETOUT;
	  if (v1<1) v1=1;
	  if (v2>p->facecount) v2=p->facecount;
	  if (lastptr>nextptr) nextptr=lastptr;
	  /* now should have valid v1 and v2 */
	  for (n=v1;n<=v2;n++)
	    if ( (!cflg && !q_flag && p->faces[n].mark) 
		 || (cflg && !q_flag && !p->faces[n].mark) 
		 || (!cflg && q_flag && n<=qcount
		     && qfaces[n].mark)
		 || (cflg && q_flag && n<=qcount
		     && !qfaces[n].mark) )
	      {
		trace->v=n;
		trace->next=(struct Vertlist *)
		  calloc((size_t)1,sizeof(struct Vertlist));
		clobber=trace;
		trace=trace->next;
		count++;
	      }
	} 
/* invalid corner data */
      else if (!strncmp(next,"nan",3))
	{
	  for (i=1;i<=p->facecount;i++)
	    for (j=0;j<3;j++)
	      {
		n=p->faces[i].vert[j];
		if (isnan((singledouble)pR_ptr[n].center.re)
		    || isnan((singledouble)pR_ptr[n].center.im))
		{
		  trace->v=i;
		  trace->next=(struct Vertlist *)
		    calloc((size_t)1,sizeof(struct Vertlist));
		  clobber=trace;
		  trace=trace->next;
		  count++;
		  j=3; /* exit little loop */
		}
	      }
	}
      else if (*next=='p') /* plot_flag >0? */
	{
	  wflag=0;
	  if (*(next+1)=='c') wflag=1;  /* NOT set */
	  for (n=1;n<=p->facecount;n++)
	    if ((!wflag && p->faces[n].plot_flag>0) 
		|| (wflag && p->faces[n].plot_flag<=0))
	      {
		trace->v=n;
		trace->next=(struct Vertlist *)
		  calloc((size_t)1,sizeof(struct Vertlist));
		clobber=trace;
		trace=trace->next;
		count++;
	      }
	}
/* random subset of given list */
      else if (next[0]=='q') {
	double prob=.5; /* default, bernoulli */
	lastptr++;
	if (!grab_next(&nextptr,next2)) { /* default to all, bernoulli */
	  strcpy(next2,"a");
	}
	else if (sscanf(next2,"%lf",&prob)) {
	  if (prob<=0 || prob>=1.0) 
	    prob=.5;
	  if (!grab_next(&nextptr,next2)) /* default to all */
	    strcpy(next2,"a");
	}
	if ((whole_list=Face_link_parse(p,next2,&eptr,&shts,Vlist,Elist,
					Flist,region,pathlist,pathlength))
	    && shts) {
	  vtrace=whole_list;
	  while (vtrace) {
	    /* randomly check next vert for inclusion */
	    if (gwrand()<=prob) {
	      trace->v=vtrace->v;
	      trace->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      clobber=trace;
	      trace=trace->next;
	      count++;
	    }
	    vtrace=vtrace->next;
	  }
	  vert_free(&whole_list);
	}
      }  
      else if (*next=='{')
	{
	  tmp_ptr=brace_parse(p,lastptr,&cf_flag,&lastptr,packdata);
	  if (tmp_ptr!=NULL) /* fit into growing list */
	    {
	      if (facelist==trace) facelist=tmp_ptr;
	      free(trace);
	      clobber=tmp_ptr;
	      count++;
	      while (clobber->next!=NULL)
		{
		  count++;
		  clobber=clobber->next;
		}
	      clobber->next=(struct Vertlist *)
		calloc((size_t)1,sizeof(struct Vertlist));
	      trace=clobber->next;
	    }
	}
/* containing complex number z; in sphere case, assume z = (a,b)
 * is apparent y,z-coords and so must be converted to (theta,phi)
 * and then via ss_view to real location. */
      else if (*next=='z') {
	struct Vertlist *tmplist=NULL,*tml;
	complex z;
	if (grab_next(&nextptr,next) && sscanf(next,"%lf",&z.re)
	    && grab_next(&nextptr,next) && sscanf(next,"%lf",&z.im)) {
	    int okay=1;
	    if (p->hes>0) {
	    	double a;
	    	if ((a=(z.re*z.re+z.im*z.im))>=1.0) // not on sphere
	    		okay=0;
	    	else 
	    		z=ss_view(p->screen,proj_vec_to_sph(sqrt(1.0-a),z.re,z.im),0,&i);
	    }
	    if (okay && (tmplist=tml=tri_search(p,z))) { /* found some */
	  		while (tml){
			    trace->v=tml->v;
	    		trace->next=(struct Vertlist *)
			      calloc((size_t)1,sizeof(struct Vertlist));
	    		clobber=trace;
			    trace=trace->next;
	    		tml=tml->next;
			    count++;
	  		}
		  vert_free(&tmplist);
	    }
	    }
      }

/* just ordinary face numbers */
      else if (sscanf(next,"%d",&v1) && v1>0 && v1<=p->facecount) 
	{
	  trace->v=v1;
	  trace->next=(struct Vertlist *)
	    calloc((size_t)1,sizeof(struct Vertlist));
	  clobber=trace;
	  trace=trace->next;
	  count++;
	}
      else goto F_GETOUT;   /* inappropriate entry */

      stripsp(nextptr);
      lastptr=nextptr;
    }
  while (*nextptr!='-' && grab_next(&nextptr,next));
  *endptr=nextptr;
 F_GETOUT: /* clean up and leave */
  *hits=count;
  if (trace!=clobber) 
    {
      clobber->next=NULL; /* clobber should be last valid data */
      free(trace);trace=NULL;
    }
  if (!count) /* no data; should have trace=clobber=vertlist */
    {
      free(trace);
      return NULL;
    }
  return facelist;
} /* Face_link_parse */

int grab_one_face(struct p_data *p,char **ptr)
     /* shortcut to grab first legal face ptr, advance ptr.
Return 0 on error or no data. */
{
  int f,hits;
  char next[256],*endptr;
  struct Vertlist *flist=NULL;

  if (!grab_next(ptr,next) 
      || !(flist=Face_link_parse(p,next,&endptr,&hits,&Vlist,&Elist,&Flist,
				 &region,pathlist,pathlength)))
    return 0;
  f=flist->v;
  vert_free(&flist);
  return f;
} /* grab_one_face */

struct Vertlist *rectify_flist(struct p_data *p,struct Vertlist *list)
/* remove any illegal indices from list */
{
	int v;
	struct Vertlist *trace,*clobber,*hold=NULL;
	
	trace=clobber=list;
	while (trace && ((v=trace->v)<1 || v>p->facecount)) {
		clobber=trace;
		trace=trace->next;
		free(clobber);
	}
	hold=trace;
	if (!hold) return NULL; // nothing legal is left
	
	// else have valid starting item
	trace=hold;
	while (trace->next) {
		if ((v=trace->next->v)<1 || v>p->facecount) {
			clobber=trace->next;
			trace->next=trace->next->next;
			free(clobber);
		}
		else trace=trace->next;
	}
	return hold;
} /* rectify_flist */
