#include "Interface.h"
#include "sa_head.h"
#include "sa_types.h"

int sa_main(int argc,char *argv[]);
JNIEnv* AC_global_env;
JNIEnv* C2J_global_env;
jclass AC_class_ID;
jclass C2J_class_ID;


// java call to execute this program; see CPNative.java.
JNIEXPORT void JNICALL Java_JNI_CPNative_init(JNIEnv* env, jclass jc) {
	sa_main(1,NULL);
}

/* ======== setting ID's for CPack calls to static routines in CPack2java */

jmethodID C2J_drawCircle_ID;
jmethodID C2J_drawFilledCircle_ID;
jmethodID C2J_drawColorCircle_ID;
jmethodID C2J_drawFilledColorCircle_ID;
jmethodID C2J_drawFace_ID;
jmethodID C2J_drawFilledFace_ID;
jmethodID C2J_drawColorFace_ID;
jmethodID C2J_drawFilledColorFace_ID;
jmethodID C2J_drawEdge_ID;
jmethodID C2J_drawColorEdge_ID;
jmethodID C2J_drawIndex_ID;
jmethodID C2J_clearCanvas_ID;
jmethodID C2J_setPackName_ID;
jmethodID C2J_setGeometry_ID;
jmethodID C2J_setLineThick_ID;
jmethodID C2J_resetPack_ID;
jmethodID C2J_setActive_ID;
jmethodID C2J_drawPath_ID;
jmethodID C2J_updateWorld_ID;
jmethodID C2J_setAxisMode_ID;
jmethodID C2J_showMsgBuf_ID;
jmethodID C2J_showErrorMsgBuf_ID;
jmethodID C2J_showScratchMsg_ID;
jmethodID C2J_showDebugMsg_ID;
jmethodID C2J_setMatrix_ID;
jmethodID C2J_setViewBox_ID;
jmethodID C2J_setFtnText_ID;
jmethodID C2J_returnFtnValue_ID;
jmethodID C2J_returnFtnReal_ID;
jmethodID C2J_returnFtnImag_ID;
jmethodID C2J_setFlagOptions_ID;
jmethodID C2J_loadScript_ID;
jmethodID C2J_openMap_ID;
jmethodID C2J_closeMap_ID;
jmethodID C2J_drawUnitCircle_ID;
jmethodID C2J_drawEquator_ID;
jmethodID C2J_runSpecialCmd_ID;
jmethodID C2J_setMobius_ID;
jmethodID C2J_flushSidePairs_ID;
jmethodID C2J_refresh_ID;
jmethodID C2J_callInfileRead_ID;
jmethodID C2J_scriptWriteFile_ID;

/**
 * These next two routines are called from java to set info once only 
 * for various CPack calls to static java routines. (e.g. call signatures, etc.)
 * They are declared in CPNative.java and java routines are in either
 * ActiveCanvas.java or CPack2java.java
 */
/* ====================================================================
 * Setting up for calls to routines in CPack2java.java
 ==================================================================== */


JNIEXPORT void JNICALL Java_JNI_CPNative_C2JinitIDs(JNIEnv* env, jclass jc) {
  /* save the environment */
  C2J_global_env=env;
  C2J_class_ID=C2J_global_env->FindClass("JNI/CPack2java");
	
  C2J_drawCircle_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawCircle","(IDDD)V");
  C2J_drawFilledCircle_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawFilledCircle","(IDDDIII)V");
  C2J_drawColorCircle_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawColorCircle","(IDDDIII)V");
  C2J_drawFilledColorCircle_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawFilledColorCircle","(IDDDIIIIII)V");
  
  C2J_drawFace_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawFace","(IDDDDDD)V");
  C2J_drawFilledFace_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawFilledFace","(IDDDDDDIII)V");
  C2J_drawColorFace_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawColorFace","(IDDDDDDIII)V");
  C2J_drawFilledColorFace_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawFilledColorFace","(IDDDDDDIIIIII)V");

  C2J_drawEdge_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawEdge","(IDDDD)V");
  C2J_drawColorEdge_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawColorEdge","(IDDDDIII)V");
  
  C2J_drawIndex_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawIndex","(IDDII)V");
  C2J_clearCanvas_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"clearCanvas","(I)V");
  C2J_setPackName_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setPackName","(ILjava/lang/String;)V");
  C2J_setGeometry_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setGeometry","(II)V");
  C2J_setLineThick_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setLineThick","(II)V");
  C2J_resetPack_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"resetPack","(I)V");
  C2J_setActive_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setActive","(I)V");
  C2J_drawPath_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawPath","(II)V");
  C2J_updateWorld_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"updateWorld","(I)V");
  C2J_setAxisMode_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setAxisMode","(II)V");
  C2J_showMsgBuf_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"showMsgBuf","(ILjava/lang/String;)V");
  C2J_showErrorMsgBuf_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"showErrorMsgBuf","(ILjava/lang/String;)V");
  C2J_showScratchMsg_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"showScratchMsg","(ILjava/lang/String;)V");
  C2J_showDebugMsg_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"showDebugMsg","(ILjava/lang/String;)V");
  C2J_setMatrix_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setMatrix","(IDDDDDDDDD)V");
  C2J_setViewBox_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setViewBox","(IDDDD)V");
  C2J_setFtnText_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setFtnText","(Ljava/lang/String;)I");
//  C2J_returnFtnValue_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"returnFtnValue","(DD)[D");
  C2J_returnFtnReal_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"returnFtnReal","(DD)D");
  C2J_returnFtnImag_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"returnFtnImag","(DD)D");
  C2J_setFlagOptions_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setFlagOptions","(III)V");
  C2J_loadScript_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"loadScript","(Ljava/lang/String;)V");
  C2J_openMap_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"openMap","(II)V");
  C2J_closeMap_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"closeMap","()V");
  C2J_drawUnitCircle_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawUnitCircle","(I)V");
  C2J_drawEquator_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"drawEquator","(IDD)V");
  C2J_runSpecialCmd_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"runSpecialCmd","(Ljava/lang/String;)V");
  C2J_setMobius_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"setMobius","(IILjava/lang/String;DDDDDDDDI)V");
  C2J_flushSidePairs_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"flushSidePairs","(I)V");
  C2J_refresh_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"ReFreshActive","()V");
  C2J_callInfileRead_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"callInfileRead","(Ljava/lang/String;I)I");
  C2J_scriptWriteFile_ID=C2J_global_env->GetStaticMethodID(C2J_class_ID,"scriptWriteFile","(IIILjava/lang/String;)I");
  
} /* C2JinitIDs */ 

/* ====================================================================
 * Setting up for calls to routines in ActiveCanvas.java
 ==================================================================== */

JNIEXPORT void JNICALL Java_JNI_CPNative_CPNinitIDs(JNIEnv* env, jclass jc) {
   /* save the environment */
   AC_global_env=env;
   AC_class_ID=AC_global_env->FindClass("canvasses/ActiveCanvas");
} /* CPNinitIDs */

/* ====================================================================
 * The following are the CPack subroutines making calls to CPack2java routines
 ==================================================================== */

void drawCircle(int packnum,double x, double y, double rad) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawCircle_ID,packnum,x,y,rad);
} /* drawCircle */

void drawFilledCircle(int packnum, double x, double y,double rad, int r, int g, int b) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawFilledCircle_ID,packnum,x,y,rad,r,g,b);
} /* drawFilledCircle */

void drawColorCircle(int packnum, double x, double y,	double rad, int r, int g, int b) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawColorCircle_ID,packnum,x,y,rad,r,g,b);
} /* drawColorCircle */

void drawFilledColorCircle(int packnum, double x, double y,double rad, int r, int g, int b, int br, int bg, int bb) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawFilledColorCircle_ID,packnum,x,y,rad,r,g,b,br,bg,bb);
} /* drawFilledColorCircle */
			
void drawFace(int packnum,double x1,double y1,double x2,double y2,double x3,double y3) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawFace_ID,packnum,x1,y1,x2,y2,x3,y3);
} /* drawFace */

void drawFilledFace(int packnum,double x1, double y1, double x2, double y2,double x3, double y3,int r,int g,int b) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawFilledFace_ID,packnum,x1,y1,x2,y2,x3,y3,r,g,b);
} /* drawFilledFace */

void drawColorFace(int packnum,double x1, double y1, double x2, double y2,double x3, double y3,int r,int g,int b) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawColorFace_ID,packnum,x1,y1,x2,y2,x3,y3,r,g,b);
} /* drawColorFace */

void drawFilledColorFace(int packnum,double x1, double y1, double x2, double y2,double x3, double y3,int r,int g,int b,int br,int bg,int bb) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawFilledColorFace_ID,packnum,x1,y1,x2,y2,x3,y3,r,g,b,br,bg,bb);
} /* drawFilledColor */
			
void drawEdge(int packnum,double x1, double y1, double x2, double y2) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawEdge_ID,packnum,x1,y1,x2,y2);
} /* drawEdge */

void drawColorEdge(int packnum,double x1, double y1, double x2, double y2,int r,int g,int b) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawColorEdge_ID,packnum,x1,y1,x2,y2,r,g,b);
} /* drawColorEdge */

void clearCanvas(int packnum) {
  C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_clearCanvas_ID,packnum);
} /* clearCanvas */

void setPackName(int packnum, char *name) {
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)name);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setPackName_ID,packnum,jstr);
} /* setPackName */

void setGeometry(int packnum, int hes) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setGeometry_ID,packnum,hes);
} /* setGeometry */

void setLineThick(int packnum,int thick) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setLineThick_ID,packnum,thick);
} /* setLineThick */

void resetPack(int packnum) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_resetPack_ID,packnum);
} /* resetPack */

void drawPath(int packnum,int length) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawPath_ID,packnum,length);
} /* drawPath */

void setActive(int act) { // only called from ch_active_pack
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setActive_ID,act);
} /* setActive */

void updateWorld(int packnum) { // tells java ask for info 'getRealBox')
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_updateWorld_ID,packnum);
} /* updateWorld */

void setAxisMode(int packnum,int flag) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setAxisMode_ID,packnum,flag);
} /* setAxisMode */

void drawIndex(int packnum,double x,double y,int n,int msg_flag) {
	// msg_flag bits: 1=on canvas, 2=in scratch window, default=3
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawIndex_ID,packnum,x,y,n,msg_flag);
} /* drawIndex */

void showMsgBuf(int msgflag,char *msg) {
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)msg);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_showMsgBuf_ID,msgflag,jstr);
} /* showMsgBuf */
	
void showErrorMsgBuf(int errflag,char *errmsg) {
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)errmsg);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_showErrorMsgBuf_ID,errflag,jstr);
} /* showErrorMsgBuf */
	
void showScratchMsg(int msgflag,char *msg) {
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)msg);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_showScratchMsg_ID,msgflag,jstr);
} /* showScratchMsg */

void showDebugMsg(int msgflag,char *debugmsg) {
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)debugmsg);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_showDebugMsg_ID,msgflag,jstr);
} /* showDebugMsg */

void setMatrix(struct p_data *p) {
	struct s_data *q=p->screen;
	int pnum=pack_num(p);
	double q00=q->disp_trans[0][0];
	double q01=q->disp_trans[0][1];
	double q02=q->disp_trans[0][2];
	double q10=q->disp_trans[1][0];
	double q11=q->disp_trans[1][1];
	double q12=q->disp_trans[1][2];
	double q20=q->disp_trans[2][0];
	double q21=q->disp_trans[2][1];
	double q22=q->disp_trans[2][2];
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setMatrix_ID,pnum,
	   q00,q01,q02,q10,q11,q12,q20,q21,q22);
} /* setMatrix */

void setViewBox(struct p_data *p) {
	struct s_data *q=p->screen;
	jint pnum=pack_num(p);
	jdouble lx=q->box.lx;
	jdouble ly=q->box.ly;
	jdouble rx=q->box.rx;
	jdouble ry=q->box.ry;
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setViewBox_ID,pnum,lx,ly,rx,ry);
} /* setViewBox */

int setFtnText(char *ftn_text) {
	jstring jstr; 
	jstr=C2J_global_env->NewStringUTF((char *)ftn_text);
	jint reslt=C2J_global_env->CallStaticIntMethod(C2J_class_ID,C2J_setFtnText_ID,jstr);
	return (int)reslt;
} /* setFtnText */

/**
 * Still not working right.
 */
complex returnFtnValue(complex z,int *errflag) {
	jdouble *ans;
	double re=z.re;
	double im=z.im;
	ans[3]=C2J_global_env->CallStaticDoubleMethod(C2J_class_ID,C2J_returnFtnValue_ID,re,im);
	if (ans[2]<0) *errflag=1;
	else *errflag=0;
	complex w;
	w.im=ans[0];
	w.re=ans[1];
	return w;
} /* returnFtnValue */

// tmp until I get around problems with arrays
complex gogetFtnValue(complex z) {
	complex ans;
	double re=z.re;
	double im=z.im;
	ans.re=(double)C2J_global_env->CallStaticDoubleMethod(C2J_class_ID,C2J_returnFtnReal_ID,re,im);
	ans.im=(double)C2J_global_env->CallStaticDoubleMethod(C2J_class_ID,C2J_returnFtnImag_ID,re,im);
	return ans;
} /* gogetFtnValue */

int setFlagOptions(int pnum,int cflag_opt,int fflag_opt) {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setFlagOptions_ID,pnum,cflag_opt,fflag_opt);
} /* setFlagOptions */

void loadScript(char *datastr) {
	jstring jstr; 
	jstr=C2J_global_env->NewStringUTF((char *)datastr);
	C2J_global_env->CallStaticIntMethod(C2J_class_ID,C2J_loadScript_ID,jstr);
} /* loadScript */

int openMap(int p,int q) {
	jint pnum=p;
	jint qnum=q;
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_openMap_ID,pnum,qnum);
	return 1;
} /* openMap */

int closeMap() {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_closeMap_ID);
	return 1;
} /* closeMap */
	
int drawUnitCircle(struct p_data *p) {
	if (p->hes!=0) return 1; // eucl only
	jint packnum=pack_num(p);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawUnitCircle_ID,packnum);
	return 1;
} /* drawUnitCircle */

int drawEquator(struct p_data *p,double re,double im) {
	if (p->hes<=0) return 1; // sph only
	jdouble Re=re;
	jdouble Im=im;
	jint packnum=pack_num(p);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_drawEquator_ID,packnum,Re,Im);
	return 1;
} /* drawEquator */

/** Called when CPack encounters '[..]' or 'infile_' commands.
 */
int runSpecialCmd(char *datastr) {
	jstring jstr; 
	jstr=C2J_global_env->NewStringUTF((char *)datastr);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_runSpecialCmd_ID,jstr);
	return 1;
} /* runSpecialCmd */

/**
 * There are three MobTypes:
 * 0 = general; available (perhaps) to all, goes in mobiusHandler.toolBar.
 * 1 = sidepairing; for multi-connected cases, goes in sidePairHandler.toolBar.
 * 2 = generic; for this packing only, goes in utilityHandler.toolBar.
 */
int setMobius(struct p_data *p,int MobType,char *name,Mobius mob) {
//sprintf(msgbuf,"enter setMobius");
//debugmsg();
	jint packnum=pack_num(p);
	jint mob_type=MobType;
	jstring jstr;
	jstr=C2J_global_env->NewStringUTF((char *)name);
	jdouble are=mob.a.re;
	jdouble aim=mob.a.im;	
	jdouble bre=mob.b.re;
	jdouble bim=mob.b.im;	
	jdouble cre=mob.c.re;
	jdouble cim=mob.c.im;	
	jdouble dre=mob.d.re;
	jdouble dim=mob.d.im;
	jint flip = mob.flip; // orientation: flip=1 means orientation reversing
						  // (normally flip=0.)
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_setMobius_ID,packnum,mob_type,jstr,
		are,aim,bre,bim,cre,cim,dre,dim,flip);
	return 1;
} /* setMobius */

int flushSidePairs(struct p_data *p) {
	jint packnum=pack_num(p);
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_flushSidePairs_ID,packnum);
	return 1;
} /* flushSidePairs */

int refresh() {
	C2J_global_env->CallStaticVoidMethod(C2J_class_ID,C2J_refresh_ID);
	return 1;
} /* refresh */

/**
 * This is telling Java to load a file from the script; Java
 * converts the filename to the temp name and calls back to
 * CPack to tell it to read.
 * CAUTION: Currently one can only pass filename; in future may
 * want options as well.
 */
int callInfileRead(char *filename,int mode) {
	jstring jfile;
	jint md=mode;
	jfile=C2J_global_env->NewStringUTF((char *)filename);
	jint result=C2J_global_env->CallStaticIntMethod(C2J_class_ID,
		C2J_callInfileRead_ID,jfile,md);
	return (int)result;
}

/**
 * Tell Java to call for writing a file to the script: Java adjusts 
 * filename and calls CPack to do the writing.
 */
int scriptWriteFile(int act,int append_flag,int mode,char *filename) {
	jint jact=act;
	jint jappend=append_flag;
	jint md=mode;
	jstring jfile;
	jfile=C2J_global_env->NewStringUTF((char *)filename);
	jint result=C2J_global_env->CallStaticIntMethod(C2J_class_ID,
		C2J_scriptWriteFile_ID,jact,jappend,md,jfile);
	return (int)result;
}

/* ====================================================================
 * Calls from java. These are declared in CPNative.java. 
 ==================================================================== */
 
 // This is a key routine: it handles java requests to execute cmds for pack pnum
JNIEXPORT jint JNICALL Java_JNI_CPNative_Execute(JNIEnv* env, jclass jc, jint pnum,jstring str) {
	int ok=0;
	int thePack=pnum;
	if (thePack<0 || thePack>=NUM_PACKS) thePack=current_p;
	AC_global_env=env;
	jboolean isCopy;
	const char* utf_ptr = env->GetStringUTFChars(str,&isCopy);
//	AC_global_env->CallStaticVoidMethod(AC_class_ID,AC_setBusy_ID);
	
	// send command through CPack parsing/execution
//sprintf(msgbuf,"got thePack = %d",thePack);
//debugmsg();	
	ok=sort_cmd(utf_ptr,&thePack);
	
//	AC_global_env->CallStaticVoidMethod(AC_class_ID,AC_setBusy_ID);
	env->ReleaseStringUTFChars(str,utf_ptr);
	return ok;
} /* execute */

// This is the same except it applies to the active pack (current_p).
JNIEXPORT jint JNICALL Java_JNI_CPNative_execute(JNIEnv* env, jclass jc, jstring str) {
	int ok=0;
	AC_global_env=env;
	jboolean isCopy;
	const char* utf_ptr = env->GetStringUTFChars(str,&isCopy);
//	AC_global_env->CallStaticVoidMethod(AC_class_ID,AC_setBusy_ID);
	
	// send command through CPack parsing/execution
	ok=sort_cmd(utf_ptr,&current_p);
	
//	AC_global_env->CallStaticVoidMethod(AC_class_ID,AC_setBusy_ID);
	env->ReleaseStringUTFChars(str,utf_ptr);
	return ok;
} /* execute */

// Note: when starting, have to using showMsgBuf, showErrorMsgBuf,
//   and showScratchMsg until routines are initialized. This routine is
//   called by activeCanvas to say it's okay to send messages.
//   javaMsgHook is CPack_global variable set to zero in sa_main.
//   When set nonzero, allows 'msg()', 'emsg()', 'smsg()' calls
//   in sa_XML.
JNIEXPORT void JNICALL Java_JNI_CPNative_msgHook(JNIEnv* env,jclass jc,jint flag) {
	javaMsgHook=flag;
	return;
}

// for receiving closed path sent by java
JNIEXPORT jint JNICALL Java_JNI_CPNative_takeThisPath(JNIEnv* env,jclass jc,jint size,jdoubleArray pts) {
	jint j;
	int count=0;
	struct Pathlist *ptrace;
	
	if (size<=3) {
		sprintf(msgbuf,"path size of %d is too small.",size);
		emsg();
		return -size;
	}
	jdouble newpts[size*2];
	env->GetDoubleArrayRegion(pts,(jsize)0,(jsize)(2*size),newpts);
	path_free(&pathlist);
	pathlength=0;
    if ( (pathlist=(struct Pathlist *)calloc(1,sizeof(struct Pathlist))) !=NULL ) {
        pathlist->x=newpts[0];pathlist->y=newpts[1];
	}
    else return -count;
	ptrace=(pathlist);
	for (j=1;j<size && (ptrace->next=(struct Pathlist *)calloc(1,sizeof (struct Pathlist)))!=NULL;j++) {
		ptrace=ptrace->next;
		ptrace->x=newpts[2*j];ptrace->y=newpts[2*j+1];
		count++;	
    }
    if (count<3)  {
		sprintf(msgbuf,"path: only got %d points.",count);
		emsg();
    	path_free(&pathlist); 
    	pathlist=NULL;
    	pathlength=0;
		return -count;
    }
    /* make sure path closes up */
	ptrace=pathlist;
	while (ptrace->next!=NULL) ptrace=ptrace->next;
	if (ptrace->x!=pathlist->x || ptrace->y!=pathlist->y) {
		ptrace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
        ptrace->next->x=pathlist->x;
        ptrace->next->y=pathlist->y;
        count++;
    } 
	pathlength=count;
    return count;
}

// Java asks CPack to send the RealBox data
JNIEXPORT void JNICALL Java_JNI_CPNative_getRealBox(JNIEnv* env, jclass jc,jint packnum) {
	setViewBox(&packdata[packnum]);
} /* getRealBox */

/**
 * CPack maintains the matrix giving the current view
 * for spherical packings; it always sends info to graph
 * already adjusted, so java always draws things as
 * though the sphere is in standard position. 
 * But, one can adjust the view by dragging the mouse,
 * so java needs to keep updating the matrix. This call
 * sends a new matrix to java.
 */

/**
 *  This is called by java after it gets the size in an drawPath call 
 *  from CPack. The size should be the global constant pathlength, and the
 *  path should be in the global linked list pathlist.
 *  (See also takeThisPath, used when path is created in java.)
 */
JNIEXPORT jdoubleArray JNICALL Java_JNI_CPNative_getPathToDraw(JNIEnv* env, jclass jc,jint length) {
	jdoubleArray realPath=env->NewDoubleArray(2*length);
	double newpath[2*length];
	struct Pathlist *ptrace=pathlist; // this is a global variable in CPack
	int j=0;
	while (ptrace && j<2*length) {
		newpath[j]=ptrace->x;
		newpath[j+1]=ptrace->y;
		j+=2;
		ptrace=ptrace->next;
	}
	env->SetDoubleArrayRegion(realPath,0,2*length,newpath);
	return realPath;
}
	
/*
 * I think these global data are set with a 'pdata' call (print_inq_data).
 */
JNIEXPORT jint JNICALL Java_JNI_CPNative_getMaxIDeg(JNIEnv* env, jclass jc) {
	return max_ideg_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getMinIDeg(JNIEnv* env, jclass jc) {
	return min_ideg_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getMaxBDeg(JNIEnv* env, jclass jc) {
	return max_bdeg_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getMinBDeg(JNIEnv* env, jclass jc) {
	return min_bdeg_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getFirstFace(JNIEnv* env, jclass jc) {
	return first_face_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getFirstRedFace(JNIEnv* env, jclass jc) {
	return first_red_face_global;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getNBdryComponents(JNIEnv* env, jclass jc) {
	return num_bdry_comp_global;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getArea(JNIEnv* env, jclass jc) {
	/*if (packdata[current_p].hes)
	{
	  sprintf(msgbuf,"p%d is not euclidean.",pack_num(&packdata[current_p]));
	  emsg();
	  return 0;
	}
      double dum=0.0;
	  for (int i=1;i<=packdata[current_p].facecount;i++)
		  dum += e_area(packdata[current_p].packR_ptr[packdata[current_p].faces[i].vert[0]].rad,
		      packdata[current_p].packR_ptr[packdata[current_p].faces[i].vert[1]].rad,
		      packdata[current_p].packR_ptr[packdata[current_p].faces[i].vert[2]].rad);
      sprintf(msgbuf," eucl area=%f; ",(singledouble)dum);
	  return dum;*/
	return area_global;
} /* getArea */

JNIEXPORT jint JNICALL Java_JNI_CPNative_getBdryLength(JNIEnv* env, jclass jc) {
	return length_global;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getInvDistance(JNIEnv* env, jclass jc) {
	return inv_dist_global;
}

JNIEXPORT jstring JNICALL Java_JNI_CPNative_getTopology(JNIEnv* env, jclass jc) {
	jstring jstr;
	jstr=env->NewStringUTF(topology_global);
	return jstr;
} /* getTopology */

/*
 * whereas these data are stored with in packdata
 */

JNIEXPORT jint JNICALL Java_JNI_CPNative_getNodeCount(JNIEnv* env, jclass jc) {
	return packdata[current_p].nodecount;
}

JNIEXPORT long JNICALL Java_JNI_CPNative_getGeometry(JNIEnv* env, jclass jc) {
	return packdata[current_p].hes;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getAlpha(JNIEnv* env, jclass jc) {
	return packdata[current_p].alpha;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getGamma(JNIEnv* env, jclass jc) {
	return packdata[current_p].gamma;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getFaceCount(JNIEnv* env, jclass jc) {
	return packdata[current_p].facecount;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getGenus(JNIEnv* env, jclass jc) {
	return packdata[current_p].genus;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getEuler(JNIEnv* env, jclass jc) {
	return packdata[current_p].euler;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getIntrinsicGeometry(JNIEnv* env, jclass jc) {
	return packdata[current_p].intrinsic_geom;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getNInteriorComponents(JNIEnv* env, jclass jc) {
	return packdata[current_p].num_int_comp;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getNInteriorNodes(JNIEnv* env, jclass jc) {
	return packdata[current_p].intnode;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getDisplayFlag(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->cflag_opt;
}

JNIEXPORT jint JNICALL Java_JNI_CPNative_getFacesDisplayFlag(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->fflag_opt;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getXMin(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->box.lx;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getYMin(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->box.ly;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getXMax(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->box.rx;
}

JNIEXPORT jdouble JNICALL Java_JNI_CPNative_getYMax(JNIEnv* env, jclass jc) {
	return packdata[current_p].screen->box.ry;
}

/**
 * get count of vlist/elist/flist (idx 0,1,2) or (if flag true) Vlist/Elist/Flist 
 *    idx 3 for list of 'boundary starts'.
 */
JNIEXPORT jint JNICALL Java_JNI_CPNative_getListCount(JNIEnv* env, jclass jc,jint pnum,jint idx,jint flag) {
	if (pnum<0 || pnum>NUM_PACKS || !packdata[pnum].status || idx<0 || idx>3) 
	  return 0;
	switch (idx) {
		case '0': // vlist
		if (flag) return count_call(&packdata[pnum],"-v Vlist");
		return count_call(&packdata[pnum],"-v vlist");
		case '1': // elist
		if (flag) return count_call(&packdata[pnum],"-e Elist");
		return count_call(&packdata[pnum],"-e elist");
		case '2': // elist
		if (flag) return count_call(&packdata[pnum],"-f Flist");
		return count_call(&packdata[pnum],"-f flist");
		case '3': // boundary starts
		return packdata[pnum].num_bdry_comp;
	}
	return 0;
}

/**
 * get vlist/elist/flist (idx 0,1,2) or (if flag true) Vlist/Elist/Flist 
 *   3 returns list of boundary starts
 */
JNIEXPORT jintArray JNICALL Java_JNI_CPNative_getList(JNIEnv* env, jclass jc, jint pnum,jint idx,jint flag) {
	int i,counter=0;
	Vertlist* vl;
	Edgelist* ptr;
    jintArray ret;
    jint *arr;
    
	if (pnum<0 || pnum>=NUM_PACKS || !packdata[pnum].status || idx<0 || idx>3) {
		ret=env->NewIntArray(0);
	    arr=(jint *)calloc(1,sizeof(jint));
		env->SetIntArrayRegion(ret,0,0,arr);
		return ret;
	}
	
	if (idx==0) { // verts/edges/faces/bdry starts
		if(flag==0)
			vl=packdata[pnum].vlist;
		else 
			vl=Vlist;
		Vertlist* next=vl;
		if(!next) {
			ret=env->NewIntArray(0);
			env->SetIntArrayRegion(ret,0,0,arr);
			return ret;
		}
		while(next->next) {
			counter++;
			next=next->next;
		}
		counter++;
		arr=(jint*)calloc(counter,sizeof(jint));
		ret=env->NewIntArray(counter);
		counter=0;
		next=vl;
		while(next->next) {
			arr[counter++]=next->v;
			next=next->next;
		}
		arr[counter++]=next->v;
		env->SetIntArrayRegion(ret,0,counter,arr);
		goto OUTAHERE;
	}
	else if (idx==1)  { // edges 
		if(flag==0)
			ptr=packdata[pnum].elist;
		else
			ptr=Elist;
		Edgelist* next=ptr;
		if(!next) {
			ret=env->NewIntArray(0);
			env->SetIntArrayRegion(ret,0,0,arr);
			return ret;
		}
		while(next->next) {
			counter++;
			next=next->next;
		}
		counter++;
		arr=(jint*)calloc(2*counter,sizeof(jint));
		ret=env->NewIntArray(2*counter);
		counter=0;
		next=ptr;
		while(next->next) {
			arr[counter++]=next->v;
			arr[counter++]=next->w;
			next=next->next;
		}
		arr[counter++]=next->v;
		arr[counter++]=next->w;
		env->SetIntArrayRegion(ret,0,counter,arr);
		goto OUTAHERE;
	  }
	  else if (idx==2) { // faces
		if(!flag)
			vl=packdata[pnum].flist;
		else 
			vl=Flist;
		Vertlist* next=vl;
		if(!next) {
			ret=env->NewIntArray(0);
			env->SetIntArrayRegion(ret,0,0,arr);
			return ret;
		}
		while(next->next) {
			counter++;
			next=next->next;
		}
		counter++;
		arr=(jint*)calloc(counter,sizeof(jint));
		ret=env->NewIntArray(counter);
		counter=0;
		next=vl;
		while(next->next) {
			arr[counter++]=next->v;
			next=next->next;
		}
		arr[counter++]=next->v;
		env->SetIntArrayRegion(ret,0,counter,arr);
		goto OUTAHERE;
	  }
	  else if (idx==3) { // bdry starts
		counter=packdata[pnum].num_bdry_comp;
		if (counter==0) {
			ret=env->NewIntArray(0);
			env->SetIntArrayRegion(ret,0,0,arr);
			return ret;
		}
		arr=(jint*)calloc(counter,sizeof(jint));
		ret=env->NewIntArray(counter);
		for (i=0;i<counter;i++) {
			arr[i]=(jint)packdata[pnum].bdry_starts[i+1];
		}
		env->SetIntArrayRegion(ret,0,counter,arr);
		goto OUTAHERE;
	  }

	// fell through?
	ret=env->NewIntArray(0);
    arr=(jint *)calloc(1,sizeof(jint));
	env->SetIntArrayRegion(ret,0,0,arr);
	
   OUTAHERE:
	free(arr);
	return ret;
}

JNIEXPORT jdoubleArray JNICALL Java_JNI_CPNative_getMobius(JNIEnv* env, jclass jc) {
	jdoubleArray ret=env->NewDoubleArray(8);
	double arr[8];
	complex cc=Mob.a;
	arr[0]=cc.re;
	arr[1]=cc.im;
	cc=Mob.b;
	arr[2]=cc.re;
	arr[3]=cc.im;
	cc=Mob.c;
	arr[4]=cc.re;
	arr[5]=cc.im;
	cc=Mob.d;
	arr[6]=cc.re;
	arr[7]=cc.im;

	env->SetDoubleArrayRegion(ret,0,8,arr);
	return ret;
} /* getMobius */

JNIEXPORT jboolean JNICALL Java_JNI_CPNative_showAxis(JNIEnv* env, jclass jc) {
	if(packdata[current_p].screen->coord_flag)
		return 1;
	else
		return 0;
} /* showAxis */

/**
 * The units in lineSlider (ScreenPanel) of Java are roughly in pixels (at about 272
 * pixels per unit on 2.2 unit wide default screen). We therefore take width/(2.2*272)
 * as meaning of a pixel for CPack. jpixels ranges from 0 to 12; have to add epsilon.)
 */
JNIEXPORT void JNICALL Java_JNI_CPNative_setCLineThick(JNIEnv* env, 
jclass jc,jint packnum,jint jpixels,jint jflag) {
	// jflag off = don't report back to java
	set_line_thick(&packdata[packnum],(int)jpixels,(int)jflag); 
}

/**
 * Java requests that side pairing maps be updated for the mobiusPanel 
 */
JNIEXPORT jint JNICALL Java_JNI_CPNative_sendmeSidePairs(JNIEnv* env,
jclass jc,jint packnum) {
	return updateSidePairs(&packdata[packnum]);
}

/**
 * Java wants info on circle at given (x,y) location. Return 0 in first
 * entry on error. 
 * Returned array is in order (indx, x, y, radius). indx=0 is failure.
 */
JNIEXPORT jdoubleArray JNICALL Java_JNI_CPNative_getCircleIndex(JNIEnv* env,
jclass jc,jint packnum,jdouble x,jdouble y) {
	jdoubleArray ret=env->NewDoubleArray(4);
	double arr[4];
	if (packnum<0 || packnum>=NUM_PACKS) {
		arr[0]=0.0;
		env->SetDoubleArrayRegion(ret,0,4,arr);
		return ret;
	}
	struct p_data *p=&packdata[packnum];
	
	int indx,dum;
	double a;
	struct Vertlist *list=NULL;
	complex z;
	
	z.re=x;
	z.im=y;

  	if (p->hes>0) { // convert to real sph point 
  		if ((a=(z.re*z.re+z.im*z.im))>=1.0) { // not in sphere 
			arr[0]=0.0;
			env->SetDoubleArrayRegion(ret,0,4,arr);
			return ret;
  		}
    	z=ss_view(p->screen,proj_vec_to_sph(sqrt(1.0-a),z.re,z.im),0,&dum); 
	}
	if (!(list=cir_search(p,z))) {
		arr[0]=0.0;
		env->SetDoubleArrayRegion(ret,0,4,arr);
		return ret;
	}
	indx=list->v;
	vert_free(&list);
	z=p->packR_ptr[indx].center;
	arr[0]=indx;
	arr[1]=z.re;
	arr[2]=z.im;
	arr[3]=p->packR_ptr[indx].rad;
	env->SetDoubleArrayRegion(ret,0,4,arr);
	return ret;
}
