#include "sa_types.h"
#include "sa_proto.h"

/* ================ save adjacency/transition matrix ============= */
/* Save adjacency matrix of packing (<= 5000 verts) in form for
   matlab to read; the adjacency matrix has a 1 in (i,j) spot if
   i and j are neighboring vertices, else a 0. 

   If flag=1, divide every row by the sum of its entries and set
   the rows for boundary vertices to zero: this yields the 
   "transition" matrix for the "simple" (ie. equal probability) 
   random walk on the edge-graph of the packing with absorbing
   boundary. Also add an ordered list of the boundary vertices to
   the matlab file.

   filepointer fp must be open, return 0 on error.
*/

int adjacency(FILE *fp,struct p_data *p,int flag)
{
  int i,j,start,nv,nodes;
  struct K_data *pK_ptr=p->packK_ptr;

  if ((nodes=p->nodecount)>10000) 
    {
      sprintf(msgbuf,"Packing has over 10000 vertices; too large "
	      "for 'adjacency' call.");
      emsg();
      return 0;
    }
  if (flag){
    fprintf(fp,"%% transition matrix for the simple random "
	    "walk on a circle packing graph; %d vertices.\n\n"
	    "echo on\n"
	    "T=zeros(%d,%d);\n",nodes,nodes,nodes);
    for (j=1;j<=nodes;j++) {
      if (pK_ptr[j].bdry_flag) 
	fprintf(fp,"T(%d,%d)=1;\n",j,j); /* absorbing bdry vertices */
      else {
	for (i=1;i<=nodes;i++)
	  if (nghb(p,i,j)>=0) 
	    fprintf(fp,"T(%d,%d)=1/%d;\n",i,j,pK_ptr[j].num);
      }
    }
    if (p->num_bdry_comp) { /* nonempty boundary? */
      fprintf(fp,"Bdry=[");
      for (i=1;i<=p->num_bdry_comp;i++) {
	  start=p->bdry_starts[i];
	  fprintf(fp,"%d ",start);
	  nv=start;
	  while ((nv=pK_ptr[nv].flower[0])!=start) {
	    fprintf(fp,"%d ",nv);
	  }
      }
      fprintf(fp,"];\n\n");
      fprintf(fp,"pause %% press any key to get exit distribution "
	      "function of exit probabilities.\n");
      fprintf(fp,"dist=zeros(%d);dist(%d)=1\n",nodes,p->alpha);
      fprintf(fp,"equib=T^10000*dist\n");
      fprintf(fp,"Acum(1)=equib(Bdry(1));\n");
      fprintf(fp,"for i=2:size(Bdry,2),\n");
      fprintf(fp,"Acum(i)=Acum(i-1)+equib(Bdry(i));\n");
      fprintf(fp,"end\nAcum\n");
    }
  }
  else {
    fprintf(fp,"%% adjacency matrix for a circle packing graph; "
	    "%d vertices\n\n"
	    "%s=zeros(%d,%d);\n",nodes,"",nodes,nodes);
    for (i=1;i<=nodes;i++)
      for (j=1;j<=nodes;j++)
	if (nghb(p,i,j)>=0) 
	  fprintf(fp,"A(%d,%d)=1;\n",i,j);
  }
  fprintf(fp,"%% end\n");
  return 1;
} /* adjacency */



