#include "sa_types.h"
#include "sa_proto.h"

/* apply Mobius Mob (flip_flag==1), or inverse (flip_flag==-1) to
specified list of circles. '-s' means this is a side-pairing map
and sets sp_flag.

In multiply connected cases, have to consider redchain and
side-pairing adjustments, also. However, do this iff "a" (all
vertices) was specified. Set rflg in this case and adjust redchain
circles; if sp_flag is off, also recompute the side-pairing maps.
(Note, if given mobius is a side-pairing map, then adjustment 
will simply permute the others, hence sp_flag may be turned off.)
*/

int apply_Mobius(struct p_data *p,char *datastr,int flip_flag,Mobius Mob)
{
  int v,hits,wflag=0,rflg=0,count=0,n,debug=0,sp_flag=0;
  double R;
  complex Z;
  char *endptr,*tptr,*nextptr;
  struct Vertlist *vertlist,*trace;
  struct RedList *rtrace;
  struct R_data *pR_ptr;
  
  nextptr=datastr;
  // check for -s flag (side pair)
  stripsp(nextptr);
  if (strlen(nextptr)>=2 && nextptr[0]=='-' && nextptr[1]=='s') {
  	sp_flag=1;
  	nextptr+=2;
  }
  stripsp(nextptr);
  if ((n=strlen(nextptr))>0) {
  	tptr=(char *)calloc((size_t)(strlen(nextptr)+1),sizeof(char));
  	strcpy(tptr,nextptr);
  }
  else {
  	tptr=(char *)calloc((size_t)2,sizeof(char));
    tptr[0]='a';tptr[1]='\0';
    rflg=1;
  }
  if (*tptr=='a' && *(tptr+1)!='(') rflg=1;
  if ( !(vertlist=Node_link_parse(p,tptr,&endptr,&hits,
		  &Vlist,&Elist,&Flist,&region,pathlist,pathlength)))
    {free(tptr);return 0;}
  pR_ptr=p->packR_ptr;
  trace=vertlist;
  do 
    {
      v=trace->v;
      count += mobius_of_circle(Mob,p->hes,pR_ptr[v].center,pR_ptr[v].rad,
	       &Z,&R,flip_flag);
      if (p->hes==0 && R<0) R=fabs(R);
      if (debug && (isnan(Z.re) || isnan(Z.im) || isnan(R))) 
	{
	  sprintf(msgbuf,"bad data %d",v);
	  emsg();
	}
      pR_ptr[v].center=Z;
      pR_ptr[v].rad=R;
	
      trace=trace->next;
    } while (trace!=NULL);
  vert_free(&vertlist);

  /* fix centers in redlist, too */
  if (rflg && (rtrace=p->redfaces)) 
    while (rtrace!=p->redfaces || !(wflag++))
      {
	v=p->faces[rtrace->face].vert[rtrace->v_flag];
	mobius_of_circle(Mob,p->hes,rtrace->center,rtrace->rad,
	       &Z,&R,flip_flag);
	if (p->hes==0 && R<0) R=fabs(R);
	rtrace->center=Z;
	rtrace->rad=R;
	rtrace=rtrace->next;
      }

  /* fix side-pairing mobius transforms also */
  if (p->hes<=0 && rflg && !sp_flag)
    update_pair_mob(p);
  /* fixup: side-pairing transforms not yet implemented for sph */
  free(tptr);
  return count;
} /* apply_Mobius */








