#include "sa_types.h"
#include "sa_proto.h"

/* Generate new combinatorics in pack p which are the barycentric
subdivision of the original faces. I.e., each edge gets new vertex, 
each face gets a hex barycenter. The original vertices and degrees
are unchanged, the top vertex indices are those for face barycenters.

For now, most pack information is lost; may try to fixup in the
future. For example, should make stab at new radii, centers, maybe
save overlaps, branching, etc., can interpolate new xyz locations 
if there is xyz data. 

Return 0 on error; p might be corrupt.*/

/* idea: (1) hex refine (2) add barycenters to the new 'middle' faces
   (3) "flip" each of the 3 edges around each new barycenters. */

int bary_refine(struct p_data *p)
{
  int orig_count,min_new,max_new,count=0,f,v,j0,j1,j2;
  struct K_data *pK_ptr=p->packK_ptr;

  if (!p->status) goto ABORT;
  orig_count=p->nodecount;
  if (!hex_refine(p)) goto ABORT;
  min_new=orig_count+1;
  max_new=p->nodecount;
  
  /* find the 'middle' faces --- those with all vertices new -- add barys. */
  
  for (f=1;f<=p->facecount;f++) 
    if (p->faces[f].vert[0]>=min_new && p->faces[f].vert[0]<=max_new
	&& p->faces[f].vert[1]>=min_new && p->faces[f].vert[1]<=max_new
	&& p->faces[f].vert[2]>=min_new && p->faces[f].vert[2]<=max_new) { 
      add_barycenter(p,f);
      count++;
    }

  if (!count) goto ABORT;

  /* the barycenters have the higher vertex indices */

  for (v=max_new+1;v<=p->nodecount;v++) {
    j0=pK_ptr[v].flower[0];
    j1=pK_ptr[v].flower[1];
    j2=pK_ptr[v].flower[2];
    if (!flip_edge(p,j0,j1,2) 
	|| !flip_edge(p,j1,j2,2) 
	|| !flip_edge(p,j2,j0,2))
      goto ABORT;
  }

  complex_count(p,0);
  facedraworder(p,0);
  sprintf(msgbuf,"bary_refine: %d faces have been barycentrically refine.",
	  count);
  msg();
  return count;
    
 ABORT:
  sprintf(msgbuf,"bary_refine: aborted due to problems. "
	  "Pack may be corrupt.");
  emsg();
  return 0;
} /* bary_refine */ 
