#include "sa_types.h"
#include "sa_proto.h"

extern int **face_org;

int catalog_faces(struct p_data *p)
     /* Build array of ptrs, one for each vertex v. Point to list 
of indices of faces v is in. In list for v, 0 entry is num of faces v 
has, followed by face 1, face 2, .... Eg. the face we call face 1 in 
the list contains v, flower[0], and flower[1]. Note: face_org[0]
contains the nodecount and can be used when freeing the space. */
{
  int v,j,f,w,n,*face_ptr;
  struct K_data *pK_ptr;

  if (face_org)  /* Note: this normally shouldn't happen, face_org 
		    should be cleared before this routine */
    {
      destroy_catalog();
      sprintf(msgbuf,"Warning: ran 'catalog_faces' with face_org "
	      "already set.");
      emsg();
    }
  face_org=(int **)calloc((size_t)(p->nodecount+1),sizeof(int *));

  face_org[0]=(int *)calloc((size_t)1,sizeof(int));
  *(face_org[0])=p->nodecount; /* record nodecount */

  pK_ptr=p->packK_ptr;
  for (v=1;v<=p->nodecount;v++)
    {
      face_org[v]=(int *)
	calloc((size_t)(pK_ptr[v].num+2),sizeof(int));
      *(face_org[v])=pK_ptr[v].num;
    }
  if (!p->faces) alloc_faces_space(p);
  for (f=1;f<=p->facecount;f++)
    for (j=0;j<3;j++)
      {
	v=p->faces[f].vert[j];
	face_ptr=face_org[v];	
	w=p->faces[f].vert[(j+1)%3];
	if ((n=nghb(p,v,w))<0) 
	  {
	    sprintf(msgbuf,"Combinatoric error in 'catalog_faces'.");
	    emsg();
	    destroy_catalog();
	    return 0;
	  }
	*(face_ptr+(n+1))=f;
      }
  return 1;		
} /* catalog_faces */

int destroy_catalog()
     /* face_org is global, set only in 'catalog_faces'; free this memory. */
{
  int v,n;

  if (face_org)
    {
      n=*(face_org[0]); /* contains count of pointers allocated */
      for (v=0;v<=n;v++)
	if (face_org[v]) free(face_org[v]);
      free(face_org);
      face_org=NULL;
    }
  return 1;
} /* destroy_catalog */
