#include "sa_types.h"
#include "sa_proto.h"

/* Fill 'count' with number of edges in packing which are more than 
'factor' or '1/factor' times what they are supposed to be based on current 
radii and inv dist. Return 0 if none found. Otherwise, return end
vertex (the one, say, of largest degree) of the edge having worst
error. Keep count of 'nan' occurances in center or radii.

fixup: currently only for euclidean.

*/

typedef struct
{
  int v;           /* which vertex? */
  int out;         /* invalid cent/rad or inf hyp radius */
  double maxlr;    /* max of log(length/ideal_length)) 
		      when length>ideal_length. */
  double minlr;    /* max of abs(log(length/ideal_length))
		      when length<ideal_length. */
}  Qual_Verts;

int order_code;
void qsort_qual(Qual_Verts *qual_verts,int left,int right);
int ck_qual_swap(Qual_Verts *qual_verts,int left,int right);
void swap_qual(Qual_Verts *qual_verts,int i,int j);
int ck_qual_swap(Qual_Verts *qual_verts,int left,int right);

int diag_layout_quality(struct p_data *p,double *factor,
			int *count,int *err_count,int opt_flag)
{
  int vert=0,v,w,j;
  double r1,r2,length,largest,ideal_length,a,b,ovlap,lfact;
  double maxval=0.0,newval;
  char filename[NAME_MAX];
  complex z1,z2;
  struct R_data *pR_ptr=p->packR_ptr;
  struct K_data *pK_ptr=p->packK_ptr;
  FILE *fperr=NULL;
  Qual_Verts *qual_verts=NULL;

  order_code=opt_flag;
  if (!p->status || (*factor)<(1.0+okerr)) {
    sprintf(msgbuf,"diag_layout_quality: usage, x > (1+okerr).");
    emsg();
    return -1;
  }
  *count=*err_count=0;
  lfact=log(*factor);
  
  /* create storage for data */
  qual_verts=(Qual_Verts *)calloc((p->nodecount+1),sizeof(Qual_Verts));

  /* compute the data; skip verts noted for invalid data. */
  for (v=1;v<=p->nodecount;v++)
    if(!qual_verts[v].out) {
      qual_verts[v].v=v;
      r1=pR_ptr[v].rad;
      z1=pR_ptr[v].center;
      if (isnan(r1) || isnan(z1.re) || isnan(z1.im)) {
	qual_verts[v].out=1;
	(*err_count)++;
      }
      else if (r1<=0) qual_verts[v].out=1; /* skip inf hyp case */
      else {  
	for (j=0;j<pK_ptr[v].num+pK_ptr[v].bdry_flag;j++) 
	  if ((w=pK_ptr[v].flower[j])>v && !qual_verts[w].out) {
	    r2=pR_ptr[w].rad;
	    z2=pR_ptr[w].center;
	    if (isnan(z2.re) || isnan(z2.im) || isnan(r2)) {
	      qual_verts[w].v=w;
	      qual_verts[w].out=1;
	      (*err_count)++;
	    }
	    else if (r2<=0) qual_verts[w].out=1;/* skip inf hyp case */
	    else { 
	      if (p->overlap_status) ovlap=pK_ptr[v].overlaps[j];
	      else ovlap=1.0;
	      if (p->hes<0) {
		length=h_dist(z1,z2,0);
		ideal_length=h_invdist_length(r1,r2,ovlap);
	      }
	      else if (p->hes>0) {
		length=s_dist(z1,z2);
		ideal_length=s_invdist_length(r1,r2,ovlap);
	      }
	      else {
		length=cAbs(csub(z1,z2));
		ideal_length=invdist_length(r1,r2,ovlap);
	      }
	      if (ideal_length>0 && length>ideal_length) {
		newval=log(length/ideal_length);
		qual_verts[v].maxlr =
		  (newval>qual_verts[v].maxlr) ? newval : qual_verts[v].maxlr;
		qual_verts[2].maxlr =
		  (newval>qual_verts[w].maxlr) ? newval : qual_verts[w].maxlr;
	      }
	      else if (length>0 && ideal_length>length) {
		newval=-log(length/ideal_length);
		qual_verts[v].maxlr =
		  (newval>qual_verts[v].maxlr) ? newval : qual_verts[v].maxlr;
		qual_verts[w].maxlr =
		  (newval>qual_verts[w].maxlr) ? newval : qual_verts[w].maxlr;
	      }
	    }
	  }
      }
    }

  /* now report the results */

  if (opt_flag==4) {  /* want number of vertices with length/ideal_length
			 exceeding factor */
    for (v=1;v<=p->nodecount;v++)
      if (qual_verts[v].maxlr>lfact || qual_verts[v].minlr>lfact) {
	(*count)++;
	if (qual_verts[v].maxlr>maxval) {
	  maxval=qual_verts[v].maxlr;
	  vert=v;
	}
	if (qual_verts[v].minlr>maxval) {
	  maxval=qual_verts[v].minlr;
	  vert=v;
	}
      }
    if (*count) {
      free(qual_verts);
      return vert;
    }
    free(qual_verts);
    return 0;
  }
  
  /* other options involve sorting */
  qsort_qual(qual_verts,1,p->nodecount);
  sprintf(filename,"diag_vlist.txt");
  if (!(fperr=fopen(filename,"w"))) {
    sprintf(msgbuf,"diag: couldn't open file diag_vlist.txt.");
    emsg();
    free(qual_verts);
    return 0;
  }
  fprintf(fperr,"set_vlist ");
  for (v=1;v<=p->nodecount && v<=50;v++)
    fprintf(fperr,"%d ",qual_verts[v].v);
  fclose(fperr);
  sprintf(msgbuf,"diag: top 50 vlist stored; set vlist "
	  "with 'fexec diag_vlist.txt'.");
  msg();
  /* pass ratio back in argument 'factor'*/
  if (qual_verts[1].maxlr>qual_verts[1].minlr) 
    *factor=exp(qual_verts[1].maxlr);
  else *factor=exp((-1.0)*qual_verts[1].minlr);
  v=qual_verts[1].v;
  free(qual_verts);
  return (v);
} /* diag_layout_quality */

void qsort_qual(Qual_Verts *qual_verts,int left,int right)
  /* copied qsort from K&R. */
{
  int i,last;

  if (left>=right) return;
  swap_qual(qual_verts,left,(left+right)/2);
  last=left;
  for (i=left+1;i<=right;i++) {
    if (ck_qual_swap(qual_verts,i,left))
      swap_qual(qual_verts,++last,i);
  }
  swap_qual(qual_verts,left,last);
  qsort_qual(qual_verts,left,last-1);
  qsort_qual(qual_verts,last+1,right);
} /* qsort_qual */

int ck_qual_swap(Qual_Verts *qual_verts,int left,int right)
   /* return 1 if right wins, else 0 */
{
  double mxrt,mxlt;

  if (left==right) return 0;
  if (order_code==1) { /* largest length/ideal_length */
    if (qual_verts[right].maxlr>qual_verts[left].maxlr) 
      return 0;
    return 1;
  }
  if (order_code==2) { /* smallest length/ideal_length */
    if (qual_verts[right].minlr>qual_verts[left].minlr) 
      return 0;
    return 1;
  }
  if (order_code==3) { /* largest of either */
    mxrt=(qual_verts[right].maxlr>qual_verts[right].minlr) ?
      qual_verts[right].maxlr : qual_verts[right].minlr;
    mxlt=(qual_verts[left].maxlr>qual_verts[left].minlr) ?
      qual_verts[left].maxlr : qual_verts[left].minlr;
    if (mxrt>mxlt) 
      return 0;
    return 1;
  }
  return 1;
} /* ck_qual_swap */

void swap_qual(Qual_Verts *qual_verts,int i,int j)
{
  Qual_Verts *tmp=NULL;

  tmp=(Qual_Verts *)calloc(1,sizeof(Qual_Verts));
  *tmp=qual_verts[i];
  qual_verts[i]=qual_verts[j];
  qual_verts[j]=*tmp;
  free(tmp);
} /* swap_qual */

