/* face_movie.cpp            September 2004, Ken Stephenson
Standalone routine associated with parallel implementation of 'toy_mpi';
idea is to watch emergence of a face shape during packing iterations.
The *.rs input file contains a sequence of RADII data sets.

    face_movie {packname}.p {radiifile}.rs

Output is a script to run and show a *.jpg movie.
*/

#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include "cp_types.h"
#include "cp_proto.h"
#include "cp_head.h"

char *msgbuf;
char *emsgbuf;

int main(int argc, char **argv)
{
  int v,j,readkey,count=0;
  char filename[BUFSIZE],outname[BUFSIZE],scname[BUFSIZE];
  char command[128],buf[256];
  FILE *fp=NULL,*psfp=NULL,*scfp=NULL;
  struct p_data *p;
  struct R_data *pR_ptr;

  set_CPack_globals();
  sc_print_size=5;

  if (argc<3) {
    printf("Usage: ViewCP {packname.p}.\n");
    exit(2);
  }

  /* read a packing */
  if (!strcpy(filename,argv[argc-2])
      || !(p=init_packing())
      || !(fp=fopen(filename,"r"))
      || !(readkey=readpack(fp,p))) {
    if (fp) fclose(fp);
    printf("  Failed to get the packing %s.\n",filename);
    exit(3);
  } 
  complex_count(p,0);
  facedraworder(p,0);
  fclose(fp);

  pR_ptr=p->packR_ptr;

  /* open the file of radii */
  if (!strcpy(filename,argv[argc-1]) || !(fp=fopen(filename,"r"))) {
    if(fp) fclose(fp);
    printf("  Failed to get the file %s of radii lists.\n",filename);
    exit(3);
  }  

  /* set up the a script file that runs an animation */

  sprintf(scname,"%s_animate",filename);
  if (!(scfp=fopen(scname,"w"))) {
    printf("  Failed to open the animate script file %s\n",scname);
    exit(3);
  }  
  fprintf(scfp,"animate \\\n  -delay 25\\\n");

  /* iterate through the sets of stored radii */

  while( (fscanf(fp,"%127s",command)!=EOF))
    if (!strcmp(command,"RADII:"))  {
      for (v=1;v<=p->nodecount;v++)
	if (CPgetd(fp, &pR_ptr[v].rad)<0) { /* error; set rest to
					       default radii */
	  sprintf(msgbuf, "format error in radius for %d; "
		  "rest set to default.",v);
	  emsg();
	  for (j=v;j<=p->nodecount;j++) pR_ptr[j].rad=.5;
	  v=p->nodecount; /* kick out of loop */
	}

      fix_call(p,"");
      for (v=1;v<=p->nodecount;v++) p->packK_ptr[v].plot_flag=1;
      set_screen(p,"-a");
      
      sprintf(outname,"%s_%d.ps",filename,count);
      strcpy(buf,"-o -c -e b -xjpg");
      print_call(&psfp,p,outname,buf,
		 "your-name-here","message","lpr");
      fprintf(scfp,"%s_%d.jpg \\\n",filename,count); /* put in animate file */
      count++;
    } /* end of while */
    
  /* close up and report */

  fprintf(scfp," \n");
  fclose(scfp);
  sprintf(buf,"chmod +x %s_animate",filename);
  system(buf);

  sprintf(msgbuf,"face_movie has created %d %s_\?.ps files.\n"
	  "Execute %s_animate to see an animation of the *.jpg files.\n",
	  count,filename,filename);
  exit(1);
} /* face_movie */

