#include "sa_types.h"
#include "sa_proto.h"

/* get_xyz command is for retrieving and storing various pack 
information obtained from 3D information on the vertices.
The packing must have xyz data stored in p->xyzpoint. */

int get_xyz_data(struct p_data *p,char *datastr)
{
  int count=0,v,first,nxt_v,pre_v,i,j,wflag=0;
  double rad,mindist,maxdist,dist;
  char *nextpoint,next[BUFSIZE];
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;

  if (!p->status || !p->xyzpoint) {
    sprintf(msgbuf,"get_xyz: there is no 3D data stored for this "
	    "packing.");
    emsg();
    return 0;
  }
  nextpoint=datastr;
  stripsp(nextpoint);
  while (grab_next(&nextpoint,next) && next[0]=='-') { /* options */

    switch (next[1]) {

    case 'b': { /* set boundary */
      if (p->hes) { /* not euclidean */
	sprintf(msgbuf,"get_xyz: -b option: pack must be euclidean.");
	emsg();
      }
      else {
	count=0;
	for (i=1;i<=p->num_bdry_comp;i++) { /* go around all components */
	  first=v=p->bdry_starts[i];
	  wflag=0;
	  while ((nxt_v=pK_ptr[v].flower[0])!=pK_ptr[first].flower[0] 
	    || !(wflag++)) { 
	    pre_v=pK_ptr[v].flower[pK_ptr[v].num];
	    rad=(xyz_dist(p->xyzpoint[v],p->xyzpoint[nxt_v])
		 + xyz_dist(p->xyzpoint[v],p->xyzpoint[pre_v]))/2.0;
	    if (rad>0) {
	      pR_ptr[v].rad=rad;
	      count++;
	    }
	    v=nxt_v;
	  } 
	} /* end of for loop */
	return count;
      }
      break;
    } /* end of b */

    case 'l': { /* compute longest/shortest edges in 3D */
      mindist=xyz_dist(p->xyzpoint[1],p->xyzpoint[pK_ptr[1].flower[0]]);
      maxdist = mindist;
      mindist = (1.0>mindist) ? 1.0 : mindist;
      for (i=1;i<=p->nodecount;i++) {
	for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++) {
	  if (j>i) {
	    dist=xyz_dist(p->xyzpoint[i],p->xyzpoint[pK_ptr[i].flower[j]]);
	    maxdist = (dist > maxdist) ? dist : maxdist;
	    mindist = (dist < mindist) ? dist : mindist;
	  }
	}
      }
      sprintf(msgbuf,"xyz edgelength: min = %f, max = %f, "
	      "ratio = %f.",(singledouble)mindist,
	      (singledouble)maxdist,(singledouble)(maxdist/mindist));
      msg();
    } /* end of l */


    } /* end of switch */
  } /* end of while */
  return 0;
} /* get_xyz_data */

