#include "sa_types.h"
#include "sa_proto.h"

/* Given vectors of x,y data, write a postscript file of the graph;
   if show is set, then popup a ghostview image. Both x and y
   should have N values. */

int graph_popup(double *xdata,double *ydata,int N,char *datastr,int show)
{
  int n;
  double minx,maxx,miny,maxy,diffx,diffy;
  char filename[256];
  FILE *fp=NULL,*sysfp=NULL;
  
  /* determine length of x and y data; should be the same */
  if (N<2 || !sscanf(datastr,"%256s",filename)
      || !(fp=fopen(filename,"w"))) {
    return 0;
  }

  /* find max/min of x and of y */
  minx=miny=1000000;
  maxx=maxy=-1000000;
  for (n=0;n<N;n++) {
    if (xdata[n]< minx) minx=xdata[n];
    if (ydata[n]< miny) miny=ydata[n];
    if (xdata[n]> maxx) maxx=xdata[n];
    if (ydata[n]> maxy) maxy=ydata[n];
  }
  diffx=maxx-minx;
  diffy=maxy-miny;

  /* initiate PostScript file */

  /* Print the header for the file. */
  fprintf(fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
  fprintf(fp, "%%%%BoundingBox: 0 40 480 470\n");   
  fprintf(fp,"  72 72 scale %% inches\n");
  fprintf(fp,"/Times-Roman findfont 0.035 scalefont setfont\n");
  fprintf(fp,"5 5 scale\n.15 .2 translate\nnewpath\n");
  fprintf(fp,"-0.05 -.035 moveto\n");
  fprintf(fp,"((%.3f,%.3f)) show\n",minx,miny);
  fprintf(fp,".925 1.035 moveto\n((%.3f,%.3f)) show",maxx,maxy);

  /* box around graph */
  fprintf(fp,"\nnewpath\n");
  fprintf(fp,"0 0 moveto\n");
  fprintf(fp,"1 0 lineto\n1 1 lineto\n0 1 lineto\n");
  fprintf(fp,"closepath\ngsave .0025 setlinewidth stroke grestore\nclip\n\n");
  fprintf(fp,".005 setlinewidth\n");
  
  /* graph itself */
  fprintf(fp,"\nnewpath\n");
  fprintf(fp,"%f %f moveto\n",(xdata[0]-minx)/diffx,(ydata[0]-miny)/diffy);
  for (n=1;n<N;n++)
    fprintf(fp,"%f %f lineto\n",
	    (xdata[n]-minx)/diffx,(ydata[n]-miny)/diffy);
  fprintf(fp, "stroke\nshowpage\n");

  fclose(fp);
  sprintf(msgbuf,"Stored plot in %s",filename);
  msg();
  if (show) {
    sysfp=popen("/bin/sh","w");
    fprintf(sysfp," ghostview %s &\n",filename);
    pclose(sysfp);
  }
  return 1;
} /* graph_popup */ 



