#include "sa_types.h"
#include "sa_proto.h"

int holonomy_trace(struct p_data *p,char *datastr,Mobius *mob)
{
  int hits,v,w,append=0,print_flag=0,digs=1;
  complex tr2,z1,z2,w1,w2,denom;
  char *nextpoint,*endptr,*name_ptr=NULL,filename[NAME_MAX];
  struct Vertlist *facelist=NULL,*ftrace;
  FILE *fp=NULL;

  if (p->hes>0) {
    sprintf(msgbuf,"trace^2: not yet available in spherical setting.");
    emsg();
    return 0;
  }
  if (strchr(datastr,'>') 
      && (name_ptr=ptr_to_name_at_end(datastr,&append))
      && grab_next(&name_ptr,filename) ) {
    print_flag=1;
  }
  nextpoint=datastr;
  stripsp(nextpoint);
  if (!p->status || strlen(nextpoint)<2) return 0;
  if (!(facelist=Face_link_parse(p,nextpoint,&endptr,&hits,
		 &Vlist,&Elist,&Flist,&region,pathlist,pathlength))) 
    return 0;
  z1=p->packR_ptr[p->faces[(w=facelist->v)].vert[0]].center;
  z2=p->packR_ptr[p->faces[facelist->v].vert[1]].center;
  ftrace=facelist;
  while (ftrace->next) ftrace=ftrace->next;
  v=ftrace->v;
  layout_facelist(p,&facelist,1,0);
  w1=p->packR_ptr[p->faces[v].vert[0]].center;
  w2=p->packR_ptr[p->faces[v].vert[1]].center;
  vert_free(&facelist);
  *mob=ident_mob(); /* initialize transformation */
  if (p->hes<0) *mob=disc_match_mob(z1,z2,w1,w2);
  else {
    denom=csub(z1,z2);
    if (cAbs(denom)<.00000000001) return 0;
    (*mob).a=cdiv(csub(w1,w2),denom);
    (*mob).b=csub(w1,cmult((*mob).a,z1));
  }
  tr2=tsqr_mob(*mob);
  sprintf(msgbuf,"trace squared is (%lf,%lf), Mobius is: \n"
	  "a = (%lf,%lf),   b = (%lf,%lf),\n"
	  "c = (%lf,%lf),   d = (%lf,%lf),\n",
	  (singledouble)tr2.re,(singledouble)tr2.im,
	  (singledouble)(*mob).a.re,(singledouble)(*mob).a.im,
	  (singledouble)(*mob).b.re,(singledouble)(*mob).b.im,
	  (singledouble)(*mob).c.re,(singledouble)(*mob).c.im,
	  (singledouble)(*mob).d.re,(singledouble)(*mob).d.im);
  msg();
  if (print_flag) { /* print to file also */
    if (append) fp=fopen(filename,"a");
    else fp=fopen(filename,"w");
    if (!fp) {
      sprintf(msgbuf,"holonomy_trace: failed to open file %s.",filename);
      emsg();
      return 0;
    }
    while ((pow(10.,(double)digs))*toler < 0.1  /* need num digits */
	   && digs<MAX_ACCUR) digs++;
    /* print trace_squared */
    fprintf(fp,"trace_squared:\n  ");
    CPputd(fp,tr2.re,digs,'e');
    fprintf(fp," + i*(");
    CPputd(fp,tr2.im,digs,'e');
    fprintf(fp,") \nMobius:\n  ");
    /* print mobius */
    CPputd(fp,(*mob).a.re,digs,'e');
    fprintf(fp," + i*(");
    CPputd(fp,(*mob).a.im,digs,'e');
    fprintf(fp,")\n  ");
    CPputd(fp,(*mob).b.re,digs,'e');
    fprintf(fp," + i*(");
    CPputd(fp,(*mob).b.im,digs,'e');
    fprintf(fp,")\n  ");
    CPputd(fp,(*mob).c.re,digs,'e');
    fprintf(fp," + i*(");
    CPputd(fp,(*mob).c.im,digs,'e');
    fprintf(fp,")\n  ");
    CPputd(fp,(*mob).d.re,digs,'e');
    fprintf(fp," + i*(");
    CPputd(fp,(*mob).d.im,digs,'e');
    fprintf(fp,")\n\n");

    fclose(fp);
  }
  return 1;
} /* holonomy_trace */
  
