//#include "cp_head.h" FA
#include "sa_head.h"
#include <time.h>

int init_remote(char *datastr,int *rp_num)
  /* initiate a remote program. Return 0 on failure. rp_num is the
     index of this remote process in the 'remote' structure. */
{
  int i=0;
  char command[32];
  
  *rp_num=-1;
  while (i< NUM_PROC && my_remote[i].pid) i++;
  if (i==NUM_PROC) {
    sprintf(msgbuf,"Can't start remote process; "
	    "all %d slots occupied.",NUM_PROC);
    msg();
    return 0;
  }
  if (!grab_next(&datastr,command)) return 0;
  stripsp(command);
  if (strcmp(command,"-n")==0) { /* attach moniker to process */
    if (!grab_next(&datastr,command)) return 0;
    strncpy(my_remote[i].moniker,command,31);
    my_remote[i].moniker[31]='\0';
    if (!grab_next(&datastr,command)) return 0;
  }
  else /* use default moniker, command name. */
    strcpy(my_remote[i].moniker,command);
  if ( !open_service(&my_remote[i],command, datastr) ) {
    sprintf(msgbuf,"Remote process %s failed to start.",command);
    emsg();
    return 0;
  }
  strcpy(my_remote[i].program,command);
  sprintf(msgbuf,"Remote program '%s' started: moniker=%s, pid=%d.",
	  my_remote[i].program,my_remote[i].moniker,my_remote[i].pid);
  msg();
  *rp_num=i;
  return 1;
} /* init_remote */

int open_service(struct remote_control *pc,char *command,char *datastr)
{
  int pid,i;
  char *arg[13],*ptr;
  
  if (pipe(pc->pipeto) 
      || (pipe(pc->pipefr) && (close(pc->pipeto[0]),close(pc->pipeto[1]),1))){
    sprintf(msgbuf,"Pipe opening error.  ");
    emsg();
    return 0;
  } 
  /* parse datastr into argument pointers (max 12). */
  ptr=datastr;
  for(i=0;i<12 && grab_next(&ptr,buf);i++) {
    arg[i]=ptr-strlen(buf);
    if (*ptr!='\0') {*ptr='\0';ptr++;}
  }
  arg[i]=(char *)NULL;
#ifndef WIN32
/* ------------- fork ------------------------- */
  
  if ((pid = fork())) { /* --- in parent --- */
    if (pid == -1) {
      sprintf(msgbuf,"Failed to fork for process.  ");
      emsg();
      close(pc->pipeto[0]);close(pc->pipeto[1]);
      close(pc->pipefr[0]);close(pc->pipefr[1]);
      return 0;
    }
    pc->pid=pid;
    (void)close(pc->pipeto[0]);
    (void)close(pc->pipefr[1]);
    /* need file pointers for XView code */
    pc->fp_to_remote=fdopen(pc->pipeto[1], "w");
    pc->fp_from_remote=fdopen(pc->pipefr[0], "r");
    /* need IOBUF's for standalone code */
    pc->client_to=bdopen(pc->pipeto[1],"w");
    pc->client_from=bdopen(pc->pipeto[0],"r");
    pc->op_code=1;
  }
#else
if(false)
;
#endif
  else { /* --- in child --- */
    (void)close(pc->pipeto[1]);
    (void)close(pc->pipefr[0]);
    (void)dup2(pc->pipeto[0], 0);
    (void)dup2(pc->pipefr[1], 1);
    /*	(void)dup2(pc->pipefr[1], 2);  */
    /* in child, close out all notifier stuff. */
#ifndef WIN32	// FA getdtablesize is Unix call - 
				// returns the maximum number of files a process can have open
    for (i=getdtablesize();i>2;i--) 
#else
	for (i=10;i>2;i--)
#endif
      (void)close(i);
    for (i=0;i<NSIG;i++)
      (void)signal(i,SIG_DFL);
    
    execlp(command, command,
	   arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],
	   arg[7],arg[8],arg[9],arg[10],arg[11]);
    _exit(1);
  }
  return 1;
} /* open_service */
