#include "sa_types.h"
#include "sa_proto.h"

     /* call a routine to mend a localized flaw in a packing layout.
Two options: 
  - specify region with this call using "D v {n} (or c {x y}) r {r}"
    format.
  - Use vertices already marked in some earlier command. Just need to
    identify a reference vert inside the marked region for 'cookie' call.
    It can be specified with "v {n}", otherwise, we do a little
    search.
In either case, also have the 'l' option to specify the number of 
generations to be blended.
In sphere case we have to project to the plane; we use the identified
vert -- it its in the lower hemisphere, we move it and all the marked
verts. Thus, if marked region is extensive, this could cause an error.
     */
     /* fixup: -- make this automatic */
  
int mend_call(struct p_data *p,char *datastr,int totalpasses)
{
  int laps=1,vert=0,depth,max_depth,v,hits,dum,riffle_count=0;
  int *newold,*gen_list,mid_vert,i,j,done=3,qcount;
  double rad=.1;
  complex ctr;
  struct Edgelist *etrace,*eclobber;
  struct Vertlist *vertlist=NULL,*trace;
  char *nextptr,*ptr,*endptr,copydata[16];
  struct p_data *mendpack=NULL,*spack=NULL;
  Mobius mob;

  nextptr=datastr;
  if ((ptr=after_char(nextptr,"v")))
    /* read our reference vert */
    {
      if (!(vert=grab_one_vert(p,&ptr)))
	done = 16; /* some error */
      else 
	{
	  done -= 1;
	  ctr=p->packR_ptr[vert].center;
	}
    }
  if ((ptr=after_char(nextptr,"l"))) /* num. of gens to overlap. */
    {
      if (!sscanf(ptr,"%d",&laps) || laps<1)
	done=16; /* error */
    }
  /* Note: 1-bit of 'done' is for center (v OR c option), 2-bit is
     for radius (in 'D ..' option case), 16-bit reflects error. */

  if ((ptr=after_char(nextptr,"D"))) /* we're going to send string
					to Node_link_parse; have to
					read center first */
    {
      qcount=3;
      nextptr=datastr;
      while (qcount && done>0 && done<16)
	{
	  qcount--;
	  if ((done & 1) && (ptr=after_char(nextptr,"c"))) 
	    /* give x,y coords for center */
	    {
	      if (sscanf(ptr,"%lf %lf",&ctr.re,&ctr.im)!=2)
		done=16; /* some error */
	      else done -= 1;
	    }
	  if ((ptr=after_char(nextptr,"r"))) /* radius */
	    {
	      if (!sscanf(ptr,"%lf",&rad) || rad<m_toler
		  || (p->hes>0 && rad>M_PI_2))
		done=16; /* error */
	      else done -= 2;
	    }
	} /* end of while */
      if (done) /* didn't get all needed data or had error. */
	{
	  sprintf(msgbuf,
		  "usage mend: mend [D option] v {n} l {k}.");
	  emsg();
	  return 0;
	}

      /* now get the verts inside the specified disc */
      if (!(vertlist=Node_link_parse(p,datastr,&endptr,&hits,
		     &Vlist,&Elist,&Flist,&region,pathlist,pathlength)))
	{
	  sprintf(msgbuf,"usage mend: mend -D v {n} (or c {x y}) r {x} -l {k}."
		  );
	  emsg();
	  return 0;
	}
    }
  else 
    {
      vertlist=Node_link_parse(p,"m",&endptr,&hits,
	       &Vlist,&Elist,&Flist,&region,pathlist,pathlength);
      if (vertlist && (!vert || !p->packK_ptr[vert].mark)) 
	/* use already marked verts; need to find a seed interior 
	   to marked verts for use in the 'cookie' routine. */
	for (i=1;(!vert && i<=p->nodecount);i++)
	  if (!p->packK_ptr[i].bdry_flag)
	    {
	      j=0;
	      while (!vert && j<=p->packK_ptr[i].num
		     && p->packK_ptr[j].mark)
		j++;
	      if (j>p->packK_ptr[i].num) /* all neighbors are marked */
		vert=i;
	    }
      if (!vertlist || !vert) 
     /* didn't find seed or didn't find any marked vertices. */
	{
	  sprintf(msgbuf,"mend error: didn't find necessary marked"
		  " verts or seed vert.");
	  emsg();
	  vert_free(&vertlist);
	  return 0;
	}
    }

/* Summary: should have everything needed at this point. */	  
      
/* allocate pack for mend region fragment */
  if ((mendpack=(struct p_data *)
       calloc((size_t)1,sizeof(struct p_data)))==NULL
      || !copy_pack(p,mendpack))
    /* fixup: need to allocate s_data mendpack->screen for some
       calls that need it */
    {
      sprintf(msgbuf,"mend: memory or copy problems for mend packing.");
      emsg();
      vert_free(&vertlist);
      if (mendpack) free_p_data(&mendpack);
      return 0;
    }
  if (mendpack->xyzpoint) 
    {free(mendpack->xyzpoint);mendpack->xyzpoint=NULL;}
     
/* set up poison and cookie out mend fragment */

  for (i=1;i<=mendpack->nodecount;i++)
    mendpack->packK_ptr[i].util_flag=-1; /* set all to poison */
  trace=vertlist;
  while (trace)
    {
      mendpack->packK_ptr[trace->v].util_flag=0; /* non-poison */
      trace=trace->next;
    }
  vert_free(&vertlist);

  sprintf(copydata,"-v %d",vert);
  if (!cookie_cutter(mendpack,copydata,&newold) || !newold)
    {
      sprintf(msgbuf,"mend: error in cookie'ing out mend region.");
      emsg();
      if(newold) free(newold);
      free_p_data(&mendpack);
      return 0;
    }

/* set vertex_map of p */
  mendpack->vertex_map=etrace=eclobber=
    (struct Edgelist *)calloc((size_t)1,sizeof(struct Edgelist));
  for (i=1;i<=mendpack->nodecount;i++)
    {
      etrace->v=i;
      etrace->w=newold[i];
      etrace->next=
	(struct Edgelist *)calloc((size_t)1,sizeof(struct Edgelist));
      eclobber=etrace;
      etrace=etrace->next;
    }
  if (eclobber==etrace) /* none found; must be error */
    {
      free(mendpack->vertex_map);
      mendpack->vertex_map=NULL;
      sprintf(msgbuf,"mend: no vertex_map was generated.");
      emsg();
      free_p_data(&mendpack);
      return 0;
    }
  else
    {
      free(etrace);
      eclobber->next=NULL;
    }
  free(newold);

/* spherical case different: move center into northern hemisphere,
   project mend fragment to eucl plane, make copy to be repacked,
   blend that with original fragment, project back, move results into
   correct hemisphere, manually feed new centers/radii into p. */

  if (mendpack->hes>0) 
    {
      mob=ident_mob();
      if (ctr.im>M_PI_2) /* apply mobius for z --> 1/z to pack. */
	{
	  mob.a.re=mob.d.re=0.0;
	  mob.b.re=mob.c.re=1.0;
	  apply_Mobius(mendpack,"a",1,mob);
	}
      geom_to_e_pack_convert(mendpack);

      /* allocate spack for copy of mend fragment */
      if ((spack=(struct p_data *)
	   calloc((size_t)1,sizeof(struct p_data)))==NULL
	  || !copy_pack(mendpack,spack)
	  || !mendpack->vertex_map)
	{
	  sprintf(msgbuf,"mend: memory or copy problems for mend pack copy.");
	  emsg();
	  if (spack) free_p_data(&spack);
	  free_p_data(&mendpack);
	  return 0;
	}
      edge_free(&(mendpack->vertex_map)); /* keep link to p in spack's
					     vertex_map. Set mendpack's
					     vertex_map to identity. */
      mendpack->vertex_map=etrace=eclobber=
	(struct Edgelist *)calloc((size_t)1,sizeof(struct Edgelist));
      for (i=1;i<=mendpack->nodecount;i++)
	{
	  etrace->v=i;
	  etrace->w=i;
	  etrace->next=
	    (struct Edgelist *)calloc((size_t)1,sizeof(struct Edgelist));
	  eclobber=etrace;
	  etrace=etrace->next;
	}
      free(etrace);
      eclobber->next=NULL;
    } 

/* now pack the mend fragment; it should have the bdry radii of p
   (or of spack in sph case). */

  if (mendpack->hes<0) { /* hyperbolic */
    if (p->overlap_status) riffle_count=h_riffle(p,totalpasses);
    else riffle_count=h_riffle_O3(p,totalpasses); /* h_riffle_O3, no 
						     overlaps yet. */
  }
  else if (!mendpack->hes) { /* euclidean */
    if (p->overlap_status) riffle_count=e_riffle(p,totalpasses);
    else riffle_count=e_riffle_O3(p,totalpasses);
  }
  if (riffle_count) comp_pack_centers(mendpack,0,0,2,okerr);

  /* find a bdry vert v, mark all bdry's util_flags to be used
     in generation counting. */
  v=0;
  for (i=1;i<=mendpack->nodecount;i++)
    {
      if (mendpack->packK_ptr[i].bdry_flag) 
	{
	  v=i;
	  mendpack->packK_ptr[i].util_flag=1;
	}
      else mendpack->packK_ptr[i].util_flag=0;
    }
  
  /* label generations of mendpack, find depth, vert deep inside, etc */

  if (!(gen_list=label_generations(mendpack,500,&mid_vert,&dum)))
    {
      sprintf(msgbuf,"mend: error in counting generations of mend region.");
      emsg();
      if (spack) free_p_data(&spack);
      free_p_data(&mendpack);
      return 0;
    } 

  mendpack->alpha=mid_vert; /* this should be deepest vertex */
  max_depth=gen_list[mid_vert];
  free(gen_list);gen_list=NULL;
  depth=(int)(max_depth/4);
  if (depth<1) depth=1;

/* Now call blend */

  if ((p->hes>0 && !blend(spack,mendpack,v,depth))
      || ((p->hes<=0 && !blend(p,mendpack,v,depth))))
    {
      sprintf(msgbuf,"mend: didn't blend correctly.");
      emsg();
      if (spack) free_p_data(&spack);
      free_p_data(&mendpack);
      return 0;
    }
  if (!spack) /* eucl or hyp cases */
    {
      sprintf(msgbuf,"mend successful:"
	      "mend patch size %d; blend depth %d; max_depth %d; "
	      "riffle count %d.",
	      mendpack->nodecount,depth,max_depth,riffle_count);
      msg();
      free_p_data(&mendpack);
      return 1;
    }

/* handle the sph case; spack has the new stuff in it. */

  geom_to_s_pack_convert(spack);
  apply_Mobius(spack,"a",1,mob); /* in case we had switched hemispheres */
  etrace=spack->vertex_map;
  while (etrace)
    {
      p->packR_ptr[etrace->w].rad=spack->packR_ptr[etrace->v].rad;
      p->packR_ptr[etrace->w].center=spack->packR_ptr[etrace->v].center;
      etrace=etrace->next;
    }
  sprintf(msgbuf,"mend successful:"
	  "mend patch size %d; blend depth %d; max_depth %d).",
	  mendpack->nodecount,depth,max_depth);
  msg();
  free_p_data(&spack);
  free_p_data(&mendpack);
  return 1;
} /* mend_call */
