#include "sa_types.h"
#include "sa_proto.h"

/* Alternative face drawing orders for simply connected complexes.
The aim is to avoid some layout problems by, e.g., avoiding use of
4-degree vertices and/or vertices having extremely small radii.
 
Marked vertices are the ones to be avoided: that is, lay out
all others before fitting these in as possible. We will keep a
linked list of vertices whose faces we should search through
for new vert to place at a given stage, adding to the end of the
list as we progress; when that's used up, we try some that were 
skipped, creating a new list to search through. Cycle through this 
process until all circles are eventually placed.

We will return a whole new "faces" data structure and first_face;
it is the responsibility of the calling routine to use the new
stuff, throw out the redfaces, empty p->rwb_flags, allocate
edge_pair, set first_face, manage the new f_data, etc. 
We empty and use 'util_flag'. */

f_data  *new_face_order(struct p_data *p,int flag,int *first_face)
     /* create new f_data */
{
  int i,j,alpha,num,vert,face,v1,v2,f,lastface,indx,v,w,fstop;
  int nbr=0,hits=0,vertcount=0,org_flag=0,*forg;
  struct Vertlist *priortrace,*endtrace,*trace,*tracker=NULL;
  f_data *newfaces=NULL;
  struct K_data *pK_ptr=p->packK_ptr;

  if ((p->euler!=1 && p->euler!=2) || p->genus!=0 || !p->faces)
    return 0; /* not simply connected */
  /* create "newfaces" space, copy 'vert' data. */
  newfaces=(f_data *)calloc((size_t)(p->facecount+1),sizeof(f_data));
  for (j=1;j<=p->facecount;j++) 
    {
      newfaces[j].vert[0]=p->faces[j].vert[0];
      newfaces[j].vert[1]=p->faces[j].vert[1];
      newfaces[j].vert[2]=p->faces[j].vert[2];
    }
    
  /* initialize */
  if (!face_org)
    {
      catalog_faces(p);
      org_flag=1;
    }
  tracker=trace=endtrace=priortrace=
    (struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
  for (i=1;i<=p->nodecount;i++) pK_ptr[i].util_flag=pK_ptr[i].plot_flag=0;

  /* start with alpha (regardless of its situation) and contiguous
     neighbors (unmarked, if possible). */
  alpha=p->alpha;
  v=pK_ptr[alpha].flower[0];
  w=pK_ptr[alpha].flower[1];
  if (pK_ptr[v].mark || pK_ptr[w].mark)
    {
      for (i=0;(i<pK_ptr[alpha].num && !nbr);i++)
	if (!pK_ptr[(v=pK_ptr[alpha].flower[i])].mark
	    && !pK_ptr[(w=pK_ptr[alpha].flower[i+1])].mark)
	  nbr=i;
    }
  pK_ptr[alpha].plot_flag=pK_ptr[v].plot_flag=pK_ptr[w].plot_flag=1;
  for (i=0;i<=pK_ptr[alpha].num;i++)
    pK_ptr[pK_ptr[alpha].flower[i]].util_flag++;
  for (i=0;i<=pK_ptr[v].num;i++)
    pK_ptr[pK_ptr[v].flower[i]].util_flag++;
  for (i=0;i<=pK_ptr[w].num;i++)
    pK_ptr[pK_ptr[w].flower[i]].util_flag++;
  vertcount=3;
  endtrace->v=alpha;
  endtrace=endtrace->next=(Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
  endtrace->v=v;
  endtrace=endtrace->next=(Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
  endtrace->v=w;
  *first_face=lastface=*(face_org[alpha]+nbr+1);
  newfaces[lastface].index_flag=find_index(p,lastface,alpha);
  for (i=1;i<=p->facecount;i++) newfaces[i].next_face=lastface;


  /* Repeatedly go through 'tracker' list of vertices as long as
     list isn't empty and hits indicates we're still adding new
     faces. At each vert in list, add new faces as possible to the 
     drawing order, keeping track in util_flag of count. Remove 
     vert from list if all its faces are in drawing order. Set 
     f_util for faces in drawing order. */

  hits=1;
  while (tracker && hits) 
    {
      hits=0;
      priortrace=trace=tracker;

      /* Go through list of verts. Add verts on end as
	 they're first encountered, remove those we're finished
	 with. */
      while (trace)
	{
	  vert=trace->v;
	  /* already hit all neighbors of this vert? */
	  if (pK_ptr[vert].util_flag==pK_ptr[vert].num)
	    {
	      if (trace==tracker) /* at beginning of list */
		{
		  priortrace=trace=tracker->next;
		  free(tracker);
		  tracker=trace;
		}
	      else 
		{
		  priortrace->next=trace->next;
		  free(trace);
		  trace=priortrace->next;
		}
	    }
	  else /* process this vert */
	    {
	      num=pK_ptr[vert].num;
	      forg=face_org[vert];
	      /* go through the faces */
	      fstop=1;
	      while (fstop && pK_ptr[vert].util_flag!=num)
		{
		  fstop=0;
		  for (f=1;(f<=num && pK_ptr[vert].util_flag!=num);f++) 
		    {
		      face=*(forg+f);
		      indx=find_index(p,face,vert);
		      v1=newfaces[face].vert[(indx+1)%3];
		      v2=newfaces[face].vert[(indx+2)%3];
		      /* can we place either v1 or v2? */
		      if (pK_ptr[v1].plot_flag>0 
			  && pK_ptr[v2].plot_flag<=0
			  && !pK_ptr[v2].mark)
			{
			  newfaces[lastface].next_face=face;
			  lastface=face;
			  newfaces[face].index_flag=indx;
			  pK_ptr[v2].plot_flag=1;
			  for (i=0;i<=pK_ptr[v2].num;i++)
			    pK_ptr[pK_ptr[v2].flower[i]].util_flag++;
			  hits++;
			  fstop++;
			  vertcount++;
			  endtrace=endtrace->next=
			    (Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
			  endtrace->v=v2;
			}
		      else if (pK_ptr[v2].plot_flag>0 
			       && pK_ptr[v1].plot_flag<=0
			       && !pK_ptr[v1].mark)
			{
			  newfaces[lastface].next_face=face;
			  lastface=face;
			  newfaces[face].index_flag=(indx+2)%3;
			  pK_ptr[v1].plot_flag=1;
			  for (i=0;i<=pK_ptr[v1].num;i++)
			    pK_ptr[pK_ptr[v1].flower[i]].util_flag++;
			  hits++;
			  fstop++;
			  vertcount++;
			  endtrace=endtrace->next=
			    (Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
			  endtrace->v=v1;
			}
		    } /* end of for f */
		} /* end of face while */
	    }
	  if (trace) 
	    {
	      priortrace=trace;
	      trace=trace->next;
	    }
	} /* end of trace while */
      if (!hits) /* no fresh vertices found; go through again until 
		    one new vertex is plotted. */
	{
	  hits=0;
	  priortrace=trace=tracker;

	  /* Go through list of verts, but only until first hit. Add 
	     verts on end as they're first encountered, remove those 
	     we're finished with. */
	  while (trace && !hits)
	    {
	      vert=trace->v;
	      /* already hit all neighbors of this vert? */
	      if (pK_ptr[vert].util_flag==pK_ptr[vert].num)
		{
		  if (trace==tracker) /* at beginning of list */
		    {
		      priortrace=trace=tracker->next;
		      free(tracker);
		      tracker=trace;
		    }
		  else 
		    {
		      priortrace->next=trace->next;
		      free(trace);
		      trace=priortrace->next;
		    }
		}
	      else /* process this vert */
		{
		  num=pK_ptr[vert].num;
		  forg=face_org[vert];
		  /* go through the faces of vert */
		  fstop=1;
		  while (fstop && pK_ptr[vert].util_flag!=num)
		    {
		      fstop=0;
		      for (f=1;(f<=num && pK_ptr[vert].util_flag!=num);f++) 
			{
			  face=*(forg+f);
			  indx=find_index(p,face,vert);
			  v1=newfaces[face].vert[(indx+1)%3];
			  v2=newfaces[face].vert[(indx+2)%3];
			  /* can we place either v1 or v2? */
			  if (pK_ptr[v1].plot_flag>0 
			      && pK_ptr[v2].plot_flag<=0)
			    {
			      newfaces[lastface].next_face=face;
			      lastface=face;
			      newfaces[face].index_flag=indx;
			      pK_ptr[v2].plot_flag=1;
			      for (i=0;i<=pK_ptr[v2].num;i++)
				pK_ptr[pK_ptr[v2].flower[i]].util_flag++;
			      vertcount++;
			      endtrace=endtrace->next=
				(Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
			      endtrace->v=v2;
			      hits++;
			      fstop++;
			    }
			  else 
			    if (pK_ptr[v2].plot_flag>0 
				&& pK_ptr[v1].plot_flag<=0)
			    {
			      newfaces[lastface].next_face=face;
			      lastface=face;
			      newfaces[face].index_flag=(indx+2)%3;
			      pK_ptr[v1].plot_flag=1;
			      for (i=0;i<=pK_ptr[v1].num;i++)
				pK_ptr[pK_ptr[v1].flower[i]].util_flag++;
			      vertcount++;
			      endtrace=endtrace->next=
				(Vertlist*)calloc((size_t)1,sizeof(struct Vertlist));
			      endtrace->v=v1;
			      hits++;
			      fstop++;
			    }
			} /* end of for f */
		    } /* end of face while */
		}
	      if (trace)
		{
		  priortrace=trace;
		  trace=trace->next;
		}
	    } /* end of trace while */
	}
    } /* end of tracker while */
    
  /* clean up and leave */
  if (tracker) vert_free(&tracker);
  if (vertcount!=p->nodecount)
    {
      sprintf(msgbuf,"new_face_order: only placed %d of %d vertices.",
	      vertcount,p->nodecount);
      emsg();
    }
  if (org_flag) destroy_catalog();
  if (vertcount) return newfaces;
  free(newfaces);
  return NULL;
} /* new_face_order */
	      

		      
		      

		  




