#include "sa_types.h"
#include "sa_proto.h"

/* compute centers of face, 'indx' circle is at origin, next vertex
in standard orienta */

int place_face(struct p_data *p,int face,int indx)
{
  int a,k,n;
  double x,x1,x2,r,r2,ovlp,erad,s,ss1,ss2,s1,s2;
  double x12, x1p2, s12;
  struct R_data *pR_ptr=p->packR_ptr;

  if (face>p->facecount || face<1 || indx <0 || indx >2) return 0;
  a=p->faces[face].vert[indx];
  k=p->faces[face].vert[(indx+1) % 3];
  n=p->faces[face].vert[(indx+2) % 3];
  if (p->overlap_status)
    ovlp=p->packK_ptr[a].overlaps[nghb(p,a,k)];
  else ovlp=1.0;
  if (p->hes<0) { /* hyp case */
    x1=pR_ptr[a].rad;
    s1=x_to_s_rad(x1);
    x2=pR_ptr[k].rad;
    s2=x_to_s_rad(x2);
    if (s1<=0) {
      x1 = 0.99;
      s1=x_to_s_rad(x1);
      pR_ptr[a].rad=x1;
      strcpy(msgbuf,"Circle at origin had "
	     "infinite radius; radius reset.");
      emsg();
    }
    pR_ptr[a].center.re=0; pR_ptr[a].center.im=0;
    if (s2<=0) { /* if next one is infinite radius */
      pR_ptr[k].center.re=1;pR_ptr[k].center.im=0;
      erad=x1/((1+s1)*(1+s1));
      pR_ptr[k].rad=(-1)*(1-erad*erad)/(2.0+2.0*erad*ovlp);
    }
    else { 
      x12 = x1*x2;
      x1p2 = x1+x2;
      s12 = s1*s2;
      x = (x1p2-x12)/(s12*(1+s12)) - (2*x1p2 - (1+ovlp)*x12)/(4*s12);
      s= x + sqrt(x*(x+2));
      pR_ptr[k].center.re= s/(s+2);
      pR_ptr[k].center.im=0.0;
    }
  }
  else if (p->hes>0) { /* sphere case */
    /* alpha at north pole */
    pR_ptr[a].center.re=pR_ptr[a].center.im=0.0; 
    /* next out pos x-axis */
    pR_ptr[k].center.re=0.0;
    pR_ptr[k].center.im=pR_ptr[a].rad+pR_ptr[k].rad;
  }
  else { /* eucl case */
    /* alpha at origin */
    r=pR_ptr[a].rad;
    pR_ptr[a].center.re=0;pR_ptr[a].center.im=0; 
    /* next on x-axis */
    r2=pR_ptr[k].rad;
    pR_ptr[k].center.re=sqrt(r*r+r2*r2+2*r*r2*ovlp);
    pR_ptr[k].center.im=0; 
  }
  return (fancy_comp_center(p,n,nghb(p,n,a),0,1,0,1,1,toler));
} /* place_face */
