#include "sa_types.h"
#include "sa_proto.h"

     /* Project hyp max packing to sphere with normalization options. 
This is combination of 'geom_to_s', 'add_ideal', and 'NSpole' in one.
Use eucl intermediary for numerical reasons. Alpha vert goes to north 
pole, added ideal S to south. 
Options (in order of precedence): 
   -e v   gives vert to go at 1 on equator.
   -m     compute a vertex furthest in generations from alpha and S
          and place it at 1 on equator.
   -t x   dilation factor; sets ratio of spherical radii, alpha/S.
default is "-t 1" (equal radii at N and S).
Note: if packing is hyperbolic, but bdry circles are not horocycles,
then the spherical packing will not be coherent.
     */

int proj_max_to_s(struct p_data *p,char *datastr)
{
  int eflag=0,v=0,w,j,dum;
  double lam,cabs,rad;
  singledouble ratio=1.0;
  complex factor;
  Mobius Mob;
  struct Vertlist *vlist=NULL;
  char *nextpoint;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;

  if (!p->status || p->hes >= 0 || p->num_bdry_comp!=1
      || !geom_to_e_pack_convert(p)) {
    sprintf(msgbuf,"proj_to_s: invalid packing, check conditions.");
    emsg();
    return 0;
  }
  pK_ptr=p->packK_ptr;
  stripsp(data);
  nextpoint=data;
  /* find vertex to be put at 1 on equator */
  while (!eflag && grab_next(&nextpoint,next)){
    if (next[0]=='-'){
      if (next[1]=='e'){ /* designated vert (takes precedence) */
	eflag=1;
	ratio=0.0;
	if (!(v=grab_one_vert(p,&nextpoint)) || v==p->alpha
	    || p->packK_ptr[v].bdry_flag) {
	  sprintf(msgbuf,"proj_to_s: invalid -e data.");
	  emsg();
	  return 0;
	}
      }
      else if (next[1]=='m'){ /* find combinatorial midpoint: mark
				 bdry and alpha and find a vert which
				 is 'halfway' in between. */
	for (j=1;j<=p->nodecount;j++){
	  if (pK_ptr[j].bdry_flag) pK_ptr[j].util_flag=1;
	  else pK_ptr[j].util_flag=0;
	}
	pK_ptr[p->alpha].util_flag=1;
	int *list=label_generations(p,-1,&v,&dum);
	free(list);
	if (dum<3 || !v) v=0; /* failure to find vert? */
	ratio=0.0;
      }
      else if (!v && next[1]=='t'){ /* arrange right ratio of radii */
	if (!grab_next(&nextpoint,next) || !sscanf(next,"%lf",&ratio) 
	    || ratio<=0)
	  ratio=1; /* default */
      }
    }
  } /* end of 'while' for reading flags */
  
  if (v || ratio>0.0) geom_to_e_pack_convert(p);
  pR_ptr=p->packR_ptr;
  if (v && (cabs=cAbs(pR_ptr[v].center))>m_toler) { /* have vertex */
    rad=pR_ptr[v].rad;
    lam=1.0/sqrt((cabs+rad)*(cabs-rad));
    factor=cExp(lam,(-1.0)*Arg(pR_ptr[v].center));
    ratio=0.0;
  }
  else if (ratio>0.0){ /* have ratio instead */
    lam=1.0/sqrt(pR_ptr[p->alpha].rad);
    factor=cExp(lam,0.0);
    v=0;
  }
  else { /* error? */
    sprintf(msgbuf,"proj_to_s: failed to get valid normalization.");
    emsg();
    return 0;
  }

  for (j=1;j<=p->nodecount;j++) { /* scale, rotate */
    pR_ptr[j].center=cmult(pR_ptr[j].center,factor);
    pR_ptr[j].rad *= lam;
    }
  geom_to_s_pack_convert(p); /* project to sphere */
  w=p->bdry_starts[1];
  p->screen->unitcircle=0;
  vlist=(struct Vertlist *)
    calloc((size_t)1,sizeof(struct Vertlist));
  vlist->v=w;
  add_ideal(p,vlist);
  vert_free(&vlist);
  p->packR_ptr[p->nodecount].center.re=0.0;
  p->packR_ptr[p->nodecount].center.im=M_PI;
  p->packR_ptr[p->nodecount].rad=asin(2.0*lam/(1.0+lam*lam));
  if (ratio>0.0 && fabs(ratio-1.0)>m_toler) {
    /* adjust for specified ratio when on sphere */
    factor.re=factor.im=0.0;dum=0;
    NS_mobius(p->packR_ptr[p->alpha].center,
	      p->packR_ptr[p->nodecount].center,
	      factor,p->packR_ptr[p->alpha].rad,
	      p->packR_ptr[p->nodecount].rad,0.0,
	      ratio,&dum,&Mob);
    if (!dum) apply_Mobius(p,"a",1,Mob);
  }
  return 1;
} /* proj_max_to_s */
