#include "sa_types.h"
#include "sa_proto.h"

/* Routines to find circles/triangles containing particular point;
routines to make lists bassed on -x flag and xy-coords. */

struct Vertlist *cir_search(struct p_data *p,complex z)
     /* list plotted circles containing canvas point z. 
      * For sphere, z is already a real (not apparent) point
      * in spherical coords. */
{
  int i;
  complex pt;
  struct Vertlist *trace=NULL,*nodelist=NULL;

  if (p->hes>0) { 
    for (i=1;i<=p->nodecount;i++) {
      if ( (s_dist(z,p->packR_ptr[i].center)<p->packR_ptr[i].rad)
	  && p->packK_ptr[i].plot_flag>0) {
	if (!nodelist) { // first catch?
	  nodelist=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  nodelist->v=i;
	  trace=nodelist;
	}
	else {
	  trace->next=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  trace=trace->next;
	  trace->v=i;
	}
      }
    }
  }
  else {
    for (i=1;i<=p->nodecount;i++) {
      if (pt_in_cir(p,i,z) && p->packK_ptr[i].plot_flag>0) {
	if (!nodelist) {
	  nodelist=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  nodelist->v=i;
	  trace=nodelist;
	}
	else {
	  trace->next=(struct Vertlist *)
	    calloc(1,sizeof(struct Vertlist));
	  trace=trace->next;
	  trace->v=i;
	}
      }
    }
  }
  return nodelist;
} /* cir_search */

int pt_in_cir(struct p_data *p,int n,complex z)
     /* is z in eucl or hyp circle? */
{
  int front;
  double a,u,v,rad;
  complex w,ectr,pt;
  
  if (p->hes>0) {
  	if (s_dist(z,p->packR_ptr[n].center)<p->packR_ptr[n].rad) return 1;
  	return 0;
  }
  if (p->hes<0)
    h_to_e_data(p->packR_ptr[n].center,
		p->packR_ptr[n].rad, &ectr,&rad);
  else if (p->hes==0)
    {
      rad=p->packR_ptr[n].rad;
      ectr=p->packR_ptr[n].center;
    }
  w=csub(z,ectr);
  if (cAbs(w)<rad) return 1;
  return 0;
} /* pt_in_cir */

struct Vertlist *tri_search(struct p_data *p,complex z)
     /* list plotted triangles under canvas pt z.
      * For sphere, z is already a real (not apparent) point
      * in spherical coords. */
{
  int j;
  struct Vertlist *trace=NULL,*nodelist=NULL;

  for (j=1;j<=p->facecount;j++) {
	if (pt_in_tri(p,j,z) && p->faces[j].plot_flag>0) {
	  if (!nodelist) {
		nodelist=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
	    nodelist->v=j;
	    trace=nodelist;
	  }
	  else {
	    trace->next=(struct Vertlist *)calloc(1,sizeof(struct Vertlist));
	    trace=trace->next;
	    trace->v=j;
	  }
	}
  }
  return nodelist;
} /* tri_search */

int pt_in_tri(struct p_data *p,int n,complex z)
     /* complex num in triangle? */
{
  int k,rflag=0,iflag=0,front;
  complex ctr[3],pt;
  double det,dum,a;

  if (p->hes>0) return ( pt_in_sph_tri(z,p->packR_ptr[p->faces[n].vert[0]].center,
	p->packR_ptr[p->faces[n].vert[1]].center,p->packR_ptr[p->faces[n].vert[2]].center));

  for (k=0;k<=2;k++)
    {
      if (p->hes<0) 
	h_to_e_data(p->packR_ptr
		    [p->faces[n].vert[k]].center,
		    p->packR_ptr[p->faces[n].vert[k]].rad,
		    &ctr[k],&dum);
      else ctr[k]=p->packR_ptr
	     [p->faces[n].vert[k]].center;
      rflag += (z.re<=ctr[k].re);
      iflag += (z.im<=ctr[k].im);
    }
  if ( rflag < 3 && rflag > 0 && iflag < 3 && iflag > 0 ) 
    /* possible? */
    {
      if ((ctr[0].re*(ctr[1].im-ctr[2].im)
	   -ctr[1].re*(ctr[0].im-ctr[2].im)
	   +ctr[2].re*(ctr[0].im-ctr[1].im))>0) det=1.0;
      else det=-1.0;
      if  /* check for convex combination */
	(  (det*row_col(z.re,z.im,ctr[1].re,ctr[2].re,ctr[1].im,ctr[2].im))>0
	   && (det*row_col(z.re,z.im,ctr[2].re,ctr[0].re,
			   ctr[2].im,ctr[0].im))>0
	   && (det*row_col(z.re,z.im,ctr[0].re,ctr[1].re,
			   ctr[0].im,ctr[1].im))>0 )
	return 1;
    }
  return 0;
} /* pt_in_tri */

double row_col(double x,double y,double xa,double xb,double ya,double yb)
{ return (x*(ya-yb)-y*(xa-xb)+xa*yb-xb*ya);}

