#include "sa_types.h"
#include "sa_proto.h"

/* Read, and store xyz data from an open file or from a buffer. 
'datastr' is passed for future use. Data is in form: 
------------------
POINTS: {n} float
x1 y1 z1
x2 y2 z2
...
xn yn zn
----------------- 
or in the OFF format
-----------------
OFF
n . .
x1 y1 z1
x2 y2 z2
...
xn yn zn
----------------- 

(NOTE: Assume initial index is 1.) Return NULL on error. */

struct xyzPoint *read_xyz(FILE *fp,char *datastr,int *count)
{
  int ptcount=0,j;
  struct xyzPoint *xyz_list=NULL;
  char keyword[NAME_MAX],c,newln[BUFSIZE];

  *count=0;
  if (!fp) return NULL; 
  while ((fscanf(fp,"%127s",keyword)!=EOF))
    {
      if (!xyz_list 
	  && (!(strncmp(keyword,"POINTS:",7)) || (strstr(keyword,"OFF"))) 
	  && fscanf(fp,"%d",&ptcount) && ptcount>0 
	  && (xyz_list=(struct xyzPoint *)calloc((size_t)(ptcount+1),
						sizeof(struct xyzPoint))))
	/* Okay, found first instance of POINT list, ready to read */
	{
	  while ((c=(char)fgetc(fp)) && c!='\n'); /* eject rest of line */
	  for (j=1;j<=ptcount;j++)
	    {
	      if (!fgets(newln,BUFSIZE,fp) /* pick off first 3 doubles 
					       of next line */
		  || sscanf(newln,"%lf %lf %lf",
			 &xyz_list[j].x,&xyz_list[j].y,&xyz_list[j].z)!=3)
		{
		  sprintf(msgbuf,"read_xyz: didn't get 3D coords for %d.",
			  j);
		  emsg();
		  free(xyz_list);
		  *count=0;
		  return NULL;
		}
	    }
	  *count=ptcount;
	  return (xyz_list);
	}
    } /* end of while */
  sprintf(msgbuf,"read_xyz: error in finding or reading xyz points.");
  emsg();
  return NULL;
} /* read_xyz */

struct xyzPoint *read_xyz_buf(char *buf,char *datastr,int *count)
{
  int ptcount=0,j,m;
  double val[3];
  char *buf_ptr=NULL,next[BUFSIZE];
  struct xyzPoint *xyz_list=NULL;

  *count=0;
  if (!(buf_ptr=strstr(buf,"POINTS:"))) /* key word not found */
    {
      sprintf(msgbuf,"read_xyz_buf: didn't find keyword 'POINTS:'.");
      emsg();
      return NULL;
    }
  buf_ptr += 6; /* skip over it */
  if (grab_next(&buf_ptr,next) 
      && sscanf(next,"%d",&ptcount) && ptcount>0 
      && (xyz_list=
	  (struct xyzPoint *)calloc((size_t)(ptcount+1),
				    sizeof(struct xyzPoint))))
    /* Okay, found POINT list; flush rest of line. */
    {
      while (*buf_ptr!='\n' && *buf_ptr!='\0')
	buf_ptr++; 
      if (*buf_ptr=='\n') buf_ptr++; 
      if (*buf_ptr=='\0') return NULL;
      
      /* ready to read */
      for (j=1;j<=ptcount;j++)
	{
	  for (m=0;m<3;m++)
	    if (!grab_next(&buf_ptr,next) 
		|| !sscanf(next,"%lf", &val[m]))
	      {
		sprintf(msgbuf,"color_qc: didn't get "
			"all %d 3D points.",ptcount);
		emsg();
		free(xyz_list);
		return NULL;
	      }
	  xyz_list[j].x=val[0];xyz_list[j].y=val[1];xyz_list[j].z=val[2];
	}
      *count=ptcount;
      return xyz_list;
    }
  return NULL;
} /* read_xyz_buf */

	      
