#include "sa_types.h"
#include "sa_proto.h"

/* Read path data from an open file. */

int readpath(FILE *fp,struct Pathlist **pathlist,int *pathlength)
{
  int count=0;
  double xx,yy;
  char command[50];
  struct Pathlist *trace;
  
  if (!fp) return 0;
  while ( (fscanf(fp,"%s",command)!=EOF) && (strcmp(command,"PATH")!=0) 
          && (strcmp(command,"END")!=0) );
  if (strcmp(command,"END")==0 || fscanf(fp,"%lf %lf",&xx,&yy)!=2) 
    return 0; /* no points */
  *pathlength=0;
  if ( ((*pathlist)=(struct Pathlist *)calloc(1,sizeof(struct Pathlist)))
       !=NULL ) 
    {
      (*pathlist)->x=xx;(*pathlist)->y=yy;
      count++;
    }
  else return 0; 
  trace=(*pathlist);
  while 
    ( fscanf(fp,"%lf %lf",&xx,&yy)==2 && count<2000
      && ((trace->next=(struct Pathlist *)calloc(1,sizeof
                                                 (struct Pathlist)))!=NULL ) )
    {
      trace=trace->next;
      trace->x=xx;trace->y=yy;
      count++;
    }
  if (count<3)  /* didn't get enough points */
    {path_free(pathlist); *pathlist=NULL;*pathlength=0;return 0;} 
  trace=(*pathlist);
  while (trace->next!=NULL) trace=trace->next;
  if (trace->x!=(*pathlist)->x || trace->y!=(*pathlist)->y)
    {
      trace->next=(struct Pathlist *)calloc
        (1,sizeof(struct Pathlist));
      trace->next->x=(*pathlist)->x;
      trace->next->y=(*pathlist)->y;
      count++;
    } /* make sure path closes up */
  *pathlength=count;
  return count;
} /* readpath */

int points_to_path(double *points,int size,struct Pathlist **pathlist,int *pathlength)
/* Given double vector 'points', alternating real/imaginary values of points
 * forming a path, this stores the path. This is mainly for getting
 * paths from java front end. 
 */
 {
 	int path_length=0;
 	struct Pathlist *path_list=NULL,*trace;
 	int count=0;
 	
 	if (size<4 || !(path_list=(struct Pathlist *)calloc(1,sizeof(struct Pathlist))))
		return 0;
  	trace=path_list;
	while(count<size && (trace->next=(struct Pathlist *)calloc(1,sizeof
		(struct Pathlist))) )
    {
      trace=trace->next;
      trace->x=points[count];trace->y=points[count+1];
      count++;
    }
    // nothing was put in first spot
	trace=path_list->next;
	free(path_list);
	path_list=trace;
	
	while (trace->next!=NULL) trace=trace->next;
	if (trace->x!=path_list->x || trace->y!=path_list->y) {
      trace->next=(struct Pathlist *)calloc(1,sizeof(struct Pathlist));
      trace->next->x=path_list->x;
      trace->next->y=path_list->y;
      count++;
    } /* make sure path closes up */
	path_length=count;
	// things seem okay, store in return variables
	path_free(pathlist); // free old stuff
	*pathlist=path_list;
	*pathlength=path_length;
	return count;
} /* points_to_path */

// We would open the ifstream and pass it to this function as follows:
//
// ifstream infile("filename");
// readpath(infile, pathlist, &pathlength);
//
// See the updated writepath for more thorough documentation of some
// of the syntax used below.
//
// The syntax for input is similar to the output syntax, except
// we use the >> operator instead of <<.

/*
int readpath(ifstream &infile,struct Pathlist **pathlist,int *pathlength)
{
  int count=0;
  double xx,yy;
  char command[50];
  struct Pathlist *trace;
  
  if (infile.eof() || infile.fail()) return 0;
  while(!infile.eof())  {
      infile >> command;  // Extracting a string from a stream gets one
                          // word at a time.
      if(infile.fail())  break;  // We can't detect input failure
                                 // until the >> operator returns.
      if(strcmp(command, "END") == 0)  break;
      if(strcmp(command, "PATH") == 0)  break;
  }
  infile >> xx >> yy;
  if (strcmp(command,"END")==0 || infile.fail() || infile.eof())
    return 0; // no points

  *pathlength=0;
  if ( ((*pathlist)=(struct Pathlist *)calloc(1,sizeof(struct Pathlist)))
       !=NULL ) 
    {
      (*pathlist)->x=xx;(*pathlist)->y=yy;
      count++;
    }
  else return 0; 
  trace=(*pathlist);
  while 
    (!(infile.eof()) &&  count<2000
      && ((trace->next=(struct Pathlist *)calloc(1,sizeof
                                                 (struct Pathlist)))!=NULL ) )
    {
      infile >> xx >> yy;
      if(infile.fail())  break;
      trace=trace->next;
      trace->x=xx;trace->y=yy;
      count++;
    }
  if (count<3)  // didn't get enough points
    {path_free(pathlist); *pathlist=NULL;*pathlength=0;return 0;} 
  trace=(*pathlist);
  while (trace->next!=NULL) trace=trace->next;
  if (trace->x!=(*pathlist)->x || trace->y!=(*pathlist)->y)
    {
      trace->next=(struct Pathlist *)calloc
        (1,sizeof(struct Pathlist));
      trace->next->x=(*pathlist)->x;
      trace->next->y=(*pathlist)->y;
      count++;
    } // make sure path closes up
  *pathlength=count;
  return count;
} // readpath

// Other useful input functions not shown above:
//
// char ch;
// int len = 256;
// char buf[len];
//
// infile.get(ch);            // Like ch = getc(infile);
// infile.read(buf, len);     // Like fread()
// infile.getline(buf, len);  // Like fgets()
*/
