#include "sa_types.h"
#include "sa_proto.h"
#include "Interface.h"

extern IOBUF *disp_fp;

/* Drawing flag scheme: flags indicate drawing instructions
   for objects -- circles/lines/faces. Bits set as follows:
      1  ->  draw object?
      2  ->  fill? (4 and 16 imply 2, also)
      4  ->  off = foreground, on = background   --- don't think this used now
             (applies to interior, not border, overriden by bit 16)
      8  ->  border color? (off = foreground, on = recorded color)
      16 ->  interior color? (default set by bit 4, on --> recorded color)
      32 ->  display label?

Eg.  flag=3: filled object, in foreground
     flag=9: open object, border in (recorded) color
             (for 'edge', this gives colored edge)
     flag=19: filled in color, border in foreground
     flag=27: filled, border and interior in color
     flag=15: filled with background, border in color
     flag=32: label only (face or circle)

Normally, flag for each type of object; often, when passed on to
subroutine, you may need to pass color code too:
Eg. (cflag, ecol, col) for circle flag, border color code, int color 
code. */

int sa_draw_circle(struct p_data *p,int n,
		   int cflag,int ecol,int col,int show)
/* send display commands to GUI; draw circle. plot_flag>0 indicates 
circles 'reliably' placed (see, e.g. fix -l). */
{
  int front,pnum=pack_num(p);

  if (n<1 || n > p->nodecount || p->packK_ptr[n].plot_flag<=0) return 0;
  if (p->hes>0) /* adjust for sphere view */
    return (sa_draw_circle_data(pnum,p->packR_ptr[n].rad,
	    ss_view(p->screen,p->packR_ptr[n].center,1,&front),
	    cflag,ecol,col,n,show));
  return (sa_draw_circle_data(pnum,p->packR_ptr[n].rad,
			      p->packR_ptr[n].center,cflag,ecol,col,n,show));
} /* sa_draw_circle */
	
int sa_draw_circle_data(int pnum,double r,complex ctr,
			 int cflag,int ecol,int col,int label,int show)
{
  int ir=-1,ig,ib;
  int br=-1,bg,bb;
  char buf[64];
  
       if (cflag==32) { /* label only */
	 		drawIndex(pnum,ctr.re,ctr.im,label,1);
	 		return 1;
	   }
	   if (!(cflag & 1)) // 1 flag not set, don't draw
	     goto LABL;
	
	   if (!(cflag & 26)) { // just regular circle
	   	drawCircle(pnum,ctr.re,ctr.im,r);
	   	goto LABL;
	   }
	   if (cflag & 22)  { // 2, 4, or 16 filled
	   	ir=red[col];ig=green[col];ib=blue[col];
	   }
	   if (cflag & 8) { // col bdry
	   	br=red[ecol];bg=green[ecol];bb=blue[ecol];
	   }
	   if (ir>=0) { // fill
	   	if (br>=0) // fill and bdry color
	   	  drawFilledColorCircle(pnum,ctr.re,ctr.im,r,ir,ig,ib,br,bg,bb);
	   	else drawFilledCircle(pnum,ctr.re,ctr.im,r,ir,ig,ib);
	   	goto LABL;
	   }
	   if (br>=0) // bdry color
	   	  drawColorCircle(pnum,ctr.re,ctr.im,r,br,bg,bb);
	LABL:
		if (cflag & 32) // label, too
		  drawIndex(pnum,ctr.re,ctr.im,label,1);
		return 1;
} /* sa_draw_circle_data */
