#include "sa_types.h"
#include "sa_proto.h"
#include "Interface.h"

extern IOBUF *disp_fp;

/* routines for drawing faces, edges, complex */

/* Drawing flag scheme: flags indicate drawing instructions
   for objects -- circles/lines/faces. Bits set as follows:
      1  ->  draw object?
      2  ->  fill? (4 and 16 imply 2, also)
      4  ->  off = foreground, on = background -- not used any longer
             (applies to interior, not border, overriden by bit 16)
      8  ->  border color? (off = foreground, on = recorded color)
      16 ->  interior color? (default set by bit 4, on --> recorded color)
      32 ->  display label?

Eg.  flag=3: filled object, in foreground
     flag=9: open object, border in (recorded) color
             (for 'edge', this gives colored edge)
     flag=19: filled in color, border in foreground
     flag=27: filled, border and interior in color
     flag=15: filled with background, border in color
     flag=32: label only (face or circle)

Normally, flag for each type of object; often, when passed on to
subroutine, you may need to pass color code too:
Eg. (cflag, ecol, ccol) for circle flag, border color code, int color 
code.

Note: see 'bc_offset' which may effect border color indices.
*/

int sa_layout_facelist(struct p_data *p,int fflag,struct Vertlist **facelist,
		    int fix,int l_face,int show)
/* lay out faces from facelist; display as you go along if 'show' is set. 
Standard flag for colors. fix=1 means to locate faces successively 
as drawn (changing stored centers). facelist is cleared. 
l_face="live face" may be given; */
{
	int ecol=FG_COLOR,fcol=FG_COLOR,n,v1,v2,v3,vv=0;
	double o1,o2,o3;
	struct Vertlist *trace;
	struct K_data *pK_ptr;
	struct R_data *pR_ptr;

	pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
	if ((*facelist)==NULL) return 0;
	if (fflag & 4) fcol=BG_COLOR;
	trace=*facelist;
		/* find first */
	if (l_face>0 && l_face<=p->facecount) vv=l_face;
	do {
			/* if fix, draw successive ones as contig. */
	  if (fix)
	    {
	      if ( (n=nghb_tri(p,vv,trace->v))<0 )
		n=p->faces[trace->v].index_flag;
	      /* if not contig, use index_flag */
	      v1=p->faces[trace->v].vert[n];
	      v2=p->faces[trace->v].vert[(n+1)%3];
	      v3=p->faces[trace->v].vert[(n+2)%3];
	      
	      if (p->overlap_status) /* oj for edge opposite vj */
		{
		  o1=pK_ptr[v2].overlaps[nghb(p,v2,v3)];
		  o2=pK_ptr[v3].overlaps[nghb(p,v3,v1)];
		  o3=pK_ptr[v1].overlaps[nghb(p,v1,v2)];
		}
	      else o1=o2=o3=1.0;
	      any_compcenter(p,
		pR_ptr[v1].center,pR_ptr[v2].center,&pR_ptr[v3].center,
		pR_ptr[v1].rad,pR_ptr[v2].rad,&pR_ptr[v3].rad,o1,o2,o3);
			/* compute and store new center */
	    }
	  
	  while (trace!=NULL && trace->v==vv) trace=trace->next;
					/* skip repeats */
	  if (trace==NULL) break;
	  vv=trace->v;
	  if ((fflag & 1) && (vv>0) && (vv <= p->facecount))
	    {
	      if (fflag & 16) fcol=p->faces[vv].color;
	      if (fflag & 8)
		{
		  ecol=p->faces[vv].color;
		  if (ecol>=bc_min && ecol<=bc_max) ecol+=bc_offset;
		}
	      sa_draw_face(p,vv,fflag,ecol,fcol,show);
	    }
	  trace=trace->next;
	} while (trace!=NULL);
	vert_free(facelist);
	return vv;
} /* sa_layout_facelist */

int sa_draw_edge_points(int pnum, complex v_cent,complex w_cent,int col)
{
	drawEdge(pnum,v_cent.re,v_cent.im,w_cent.re,w_cent.im);
  if (col!=FG_COLOR)
	drawColorEdge(pnum,v_cent.re,v_cent.im,w_cent.re,w_cent.im,red[col],green[col],blue[col]);
  return 1;
} /* sa_draw_edge_points */

int sa_draw_edge(struct p_data *p,int v1,int v2,int col)
{
  int front;
  complex end1,end2;

  if (!p->status 
      || p->packK_ptr[v1].plot_flag<=0
      || p->packK_ptr[v2].plot_flag<=0) return 0;
  if (p->hes>0) /* sphere */
    {
      end1=ss_view(p->screen,p->packR_ptr[v1].center,1,&front);
      end2=ss_view(p->screen,p->packR_ptr[v2].center,1,&front);
    }
  else 
    {
      end1=p->packR_ptr[v1].center;
      end2=p->packR_ptr[v2].center;
    }
  return (sa_draw_edge_points(pack_num(p),end1,end2,col));
} /* sa_draw_edge */

int sa_draw_bdry_seg(struct p_data *p,int n,int eflag,int ecol)
/* draw boundary segment, given starting face and index of begin vert.
Use eflag bits for options: 4=label; 8=in color; 16=circles also. */
{
  int indx=0,w_indx,colr,pnum=pack_num(p);
  complex v_cent,w_cent;
  struct RedList *trace,*spot;

  if (!(trace=p->edge_pair[n].edge)) return 0;
  indx=(p->edge_pair[n].edge_indx+2)%3;
  /* recall: edge_indx points to END of edge; here 'indx' keeps track
     of index of edge (ie. its BEGINNING) so we can find next red edge. */
  if (trace->next->face==trace->prev->face
      && trace->v_flag!=p->edge_pair[n].edge_indx
      && (spot=back_thru_blue(p,trace))) 
    /* blue face, second edge, get data from upstream face.
       spot==NULL would indicate comb. error. */
    w_cent=spot->center;
  else w_cent=trace->center;
  colr=FG_COLOR;
  if (eflag & 8) colr=ecol;
  if (eflag & 16) /* draw/label circles forming the edge? */
    {
      w_indx=p->faces[trace->face].vert[trace->v_flag];
      sa_draw_circle_data(pnum,p->packR_ptr[w_indx].rad,
			   w_cent,41,colr,BG_COLOR,w_indx,1);
    }
  do
    {
      v_cent=w_cent;
      if (trace->next->face==trace->prev->face
	  && trace->v_flag==(indx+1)%3) 
	/* blue face, use vert at end of second edge */
	{
	  indx=(indx+1) %3;
	  if ((spot=back_thru_blue(p,trace)))
	  w_cent=spot->center;
	}
      else 
	{
	  trace=next_red_edge(p,1,trace,indx,&indx);
	  w_cent=trace->center;
	}
      sa_draw_edge_points(pnum,v_cent,w_cent,colr);
      if (eflag & 16) 
	{
	  w_indx=p->faces[trace->face].vert[trace->v_flag];
	  sa_draw_circle_data(pnum,p->packR_ptr[w_indx].rad,
			       w_cent,41,colr,BG_COLOR,w_indx,1);
	}
    } while (!(trace->corner_flag[(indx+1)%3] & 3));
  if (eflag & 4) sa_draw_bdry_seg_num(p,n);
  return 1;
} /* sa_draw_bdry_seg */

int sa_draw_bdry_seg_num(struct p_data *p,int n)
/* put segment number at first circle */
{
  char buf[64];
  complex ctr;
  struct RedList *trace;

  if (!(trace=p->edge_pair[n].edge)) return 0;
  if (trace->next->face==trace->prev->face
      && trace->v_flag!=p->edge_pair[n].edge_indx) 
    /* blue face, second edge, get data from prev face. */
    ctr=trace->prev->center;
  else ctr=trace->center;
  sprintf(buf,"%d",n);
  drawIndex(pack_num(p),ctr.re,ctr.im,n,1);
  return 1;
} /* sa_draw_bdry_seg_num */
	
int sa_draw_face(struct p_data *p,int face,int fflag,
	      int ecol,int fcol,int show)
     /* send display commands to GUI; draw face. */
{
  int i,front,pnum=pack_num(p);
  char buf[64];
  complex ctr[4];
  struct R_data *pR_ptr=p->packR_ptr;

  if (face<1 || face > p->facecount || p->faces[face].plot_flag<=0)
    return 0;
  if (p->hes>0) { /* adjust for sphere view */
    for(i=0;i<3;i++)	ctr[i]=ss_view(p->screen,
      p->packR_ptr[p->faces[face].vert[i]].center,1,&front);
  }
  else {
    for(i=0;i<3;i++) 
      ctr[i]=p->packR_ptr[p->faces[face].vert[i]].center;
  }
  // need label location 
  if (fflag & 32) ctr[3]=face_center(p,face);
//  if (p->hes>0) ctr[3]=ss_view(p->screen,ctr[3],1,&front);
  return (sa_draw_face_data(pnum,ctr,face,fflag,ecol,fcol,show));
} /* sa_draw_face */

int sa_draw_face_data(int pnum,complex ctr[],int face_index,
		      int fflag,int ecol,int fcol,int show)
{
	int ir=-1,ig,ib;
	int br=-1,bg,bb;
  	if (fflag==32) { /* label only */
		drawIndex(pnum,ctr[3].re,ctr[3].im,face_index,1);
	 	return 1;
	}
	if (!(fflag & 1)) // 1 bit not set
	  goto LABL;	
	if (!(fflag & 26)) { // just regular open face
	  	drawFace(pnum,ctr[0].re,ctr[0].im,ctr[1].re,ctr[1].im,ctr[2].re,ctr[2].im);
	   	goto LABL;
    }
	if (fflag & 22)  { // 2, 4, or 16 filled
	   	ir=red[fcol];ig=green[fcol];ib=blue[fcol];
	}
	if (fflag & 8) { // col bdry
		br=red[ecol];bg=green[ecol];bb=blue[ecol];
	}
	if (ir>=0) { // fill
		if (br>=0) // fill and bdry color
	  		drawFilledColorFace(pnum,ctr[0].re,ctr[0].im,ctr[1].re,ctr[1].im,ctr[2].re,ctr[2].im,ir,ig,ib,br,bg,bb);
	   	else drawFilledFace(pnum,ctr[0].re,ctr[0].im,ctr[1].re,ctr[1].im,ctr[2].re,ctr[2].im,ir,ig,ib);
	   	goto LABL;
	}
	if (br>=0) // bdry color
		drawColorFace(pnum,ctr[0].re,ctr[0].im,ctr[1].re,ctr[1].im,ctr[2].re,ctr[2].im,br,bg,bb);
  LABL:
	if (fflag & 32) { // label, too
		drawIndex(pnum,ctr[3].re,ctr[3].im,face_index,1);
	}
	return 1;
} /* sa_draw_face_data */

int sa_draw_path(struct p_data *p,struct Pathlist *pathlist,int length) 
     /* send re/im coords to form path; NOT assumed closed --- first
	entry must be repeated at end to define closed path. */
{
  struct Pathlist *trace;

  if (!pathlist || length<=0) return 0;
  bprintf(disp_fp,"<path count=\"%d\"> ",length);
  trace=pathlist;
  while (trace)
    {
      bprintf(disp_fp," %lf %lf ",(singledouble)(trace->x),
	      (singledouble)(trace->y));
      trace=trace->next;
    }
  bprintf(disp_fp,"</path>\n");
  return 1;
} /* sa_draw_path */

	

