/* sa_io.c = input and output routines for standalone CirclePack */

#include "sa_head.h"
#include "Interface.h"

#define SHMODE 0666
#define MAXCMD 1000

extern char *itoA(int value, char*  str, int radix);
extern char *strrev(char *str);
static char *shmBlock=NULL;
static int *lockFlg=NULL;
static int *newFlag=NULL;
static int *vertNum=NULL;
static char *cmdStr=NULL;
FILE *sysfp;

int readcall(struct p_data *p,char *filename)
/* Open file ptr; assumed that overwrite has already been approved 
if p is nonempty; filename has complete path included. */
{
  int i;
  char packname[NAME_MAX];

  if ((strlen(filename)>NAME_MAX) 
  	|| !(fp = fopen(filename, "r"))) goto FAILED;
  i=strlen(filename);
  while (i>0 && filename[i-1]!='/') i--;
  strcpy(packname,filename+i); /* cut off any directory names */
  if (!call_read_pack(fp,p,packname) ) goto FAILED;
  new_pack_setup(p);
  fclose(fp);
  return 1;

 FAILED:
  sprintf(msgbuf,"Read of %s failed; check for .gz or "
	  "check name and path.",filename);
  emsg();
  if (fp) fclose(fp);
  return 0;
} /* readcall */

int call_read_pack(FILE *fp,struct p_data *p,char *packname)
/* assume fp open, call read_pack, update pack info. Return from 
read_pack has bits encoding what info was read:
Use this to know what to update.
    1: 0001     basic combinatorics (new pack)
    2: 0002     aims
    3: 0004     inv_dist
    4: 0010     radii
    5: 0020     centers
    6: 0040     angle sums
    7: 0100     vertex_map 
    8: 0200     lists of verts/faces/edges 
    9: 0400     colors (circle or face) 
    10: 01000   vertex plot_flags
    11: 02000   xyz points */
{
  int i,val,pnum=pack_num(p);
  double halfheight,realheight;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;
	
  if (!(val=readpack(fp,p)))
    {
      sprintf(msgbuf,"Read for pack %d has failed.",pnum);
      emsg();
      return 0;
    }
  pK_ptr=p->packK_ptr;
  pR_ptr=p->packR_ptr;

  /* depending on bit coding in val, update pack info */

  if (val & 0001) /* combinatoric info (ie, it's a new pack)
		   note: complex_count was done during reading. */
    {
      sprintf(msgbuf,"Read into p%d succeeded; %d circles.",
	      pnum,p->nodecount); 
      msg();
      choose_alpha(p);choose_beta(p);choose_gamma(p);
      facedraworder(p,0);
      p->active_node=p->alpha;
      if (!(val & 0010)) /* no radii read */
	for (i=1;i<=p->nodecount;i++) pR_ptr[i].rad=0.5;
      /* radii read but not centers */
      if ((val & 0010) && !(val & 0020)) 
	comp_pack_centers(p,0,0,2,okerr);
      if (!(val & 0040)) /* no angle sums read */
	fillcurves(p);
      if (p->xyzpoint && !(val & 02000)) /* old xyz data */
	{free(p->xyzpoint);p->xyzpoint=NULL;}

      /* fix up pack name and window label */
      if ((val & 0001) && p->file_name[0]=='\0')
	/* no pack name read, use call */
	strcpy(p->file_name,packname);
      set_pack_label(p);

      /* reset screen data */
      p->screen->cflag_opt=1;
      p->screen->fflag_opt=0;
      sa_send_flag_opts(p);
      p->screen->box=std_real_box;
      realheight=(p->screen->box.ry-p->screen->box.ly);
      halfheight=realheight*p->screen->pix_box.rx/p->screen->pix_box.ry/2.0;
      p->screen->box.lx=(-1)*halfheight;
      p->screen->box.rx=halfheight;
      updateWorld(pack_num(p));
      if (realheight<=0.0 
      || !(p->screen->pix_per_unit
	   =(int)((double)(p->screen->pix_box.ry-p->screen->pix_box.ly)
		  /realheight)))
	  p->screen->pix_per_unit=(int)CANV_SIZE/2; /* default */

      /* must fix so it fits current canvas pixel size */
      live_node[pnum]=live_face[pnum]=0;
      canvas_mode_defaults(p);
    }
  else /* partial data only, not a new pack */
    {
      sprintf(msgbuf,"Read of data for p%d succeeded.",pnum); 
      msg();
    }
  return 1;
} /* call_read_pack */

int writecall(struct p_data *p,char *filename,int act)
/* calls routine for writing file from pack p. Assumed that 
overwrite has already been approved if filename	already exists. 
filename includes full path.*/
{
	if (p->status && open_and_write(p,filename,act))
	 {
		write_consolidate(p,filename);
		return 1;
	 }
	sprintf(msgbuf,"Write from %d has failed.",pack_num(p));
	emsg();
	return 0;
} /* writecall */

int write_consolidate(struct p_data *p,char *filename)
/* Update after successful write */
{
	int n,pnum=pack_num(p);
	char *name;

	sprintf(buf,"\nSENT pack %d to disc as %s.  ",pnum,filename);
	n=strlen(filename);
	while (filename[n-1]!='/' && n>0) n--;
	name=filename+n;  /* cut off any directory names */
	strcpy(p->file_name,name); /* update file_name */
	sprintf(buf,"%s",p->file_name);
	if (p->hes<0) strcat(buf," (hyp)\0");
	else if (p->hes==0) strcat(buf," (eucl)\0");
	else strcat(buf," (sph)\0");
	set_pack_label(p);
	return 1;
} /* write_consolidate */

int open_and_write(struct p_data *p,char *filename,int act)
/* open input file, call write, return 1 if successful */
{
	int i;

	if ( (strlen(filename)>NAME_MAX) 
		|| !(fp = fopen(filename, "w"))) return 0;
	i=writepack(fp,p,act,0);
	if(fp) fclose(fp);
	return i;
} /* open_and_write */

int read_path_file(char *filename)
/* read a closed path from a file; full pathname. */
{
  int count;

  if ((fp=fopen(filename,"r"))!=NULL 
      && (count=readpath(fp,&pathlist,&pathlength))) {
    fclose(fp);
    sprintf(msgbuf,"path %s was read.",filename);
    msg();
    return count;
  }
  sprintf(msgbuf,"Failed to read path %s.",filename);
  emsg();
  return 0;
} /* read_path_file */

int write_path_file(char *filename)
/* write a closed path to a file; full pathname. */
{
	int count;

	if (pathlist!=NULL && strlen(filename)<=NAME_MAX)
	 {
		 std::ofstream myFile(filename); // FA
	   count=new_writepath(myFile,pathlist);
	   myFile.close();
	   return count;
	 }
	else return 0;
} /* write_path_file */

int copy_packcall(char *datastr)
/* copy p1 into p2; */
{
	int p1,p2,width,height;
	double midpt,halfheight,realheight;
#ifndef WIN32  /* fixup: This is sort of ugly, should use a wrapper here */
               /* Need to find out what purpose xpm, xid serve, whether */
               /* they are necessary for SA. */
    Pixmap xpm_hold;
	XID xid_hold;
#endif

	if ( sscanf(datastr,"%d %d",&p1,&p2) ==2 && p1>=0 && p2!=p1
		&& p1<NUM_PACKS && p2>=0 && p2<NUM_PACKS &&
		packdata[p1].status && !packdata[p2].locks
		&& copy_pack(&packdata[p1],&packdata[p2]) )
	 {
	 		/* save pointers */
	        packdata[p2].screen=&screendata[p2];
#ifndef WIN32
	 	xid_hold=screendata[p2].xid;
	 	xpm_hold=screendata[p2].xpm;		
#endif
		screendata[p2]=screendata[p1];
#ifndef WIN32
		screendata[p2].xid=xid_hold;
		screendata[p2].xpm=xpm_hold;
#endif
			/* move screen parameters */
		width=height=CANV_SIZE;
		midpt=(screendata[p2].box.lx+screendata[p2].box.rx)/2.0;
		realheight=(screendata[p2].box.ry-screendata[p2].box.ly);
		halfheight=realheight*width/height/2.0;
		screendata[p2].box.lx=midpt-halfheight;
		screendata[p2].box.rx=midpt+halfheight;
		if (realheight<=0.0 
		    || !(screendata[p2].pix_per_unit
			 =(int)((double)(height)/realheight)))
		  screendata[p2].pix_per_unit=(int)CANV_SIZE/2; /* default*/
		updateWorld(p2);
		set_pack_label(&packdata[p2]);

		/* send various canvas info to front end */
		send_sph_view(&packdata[p2]);
		sa_send_geom(&packdata[p2]);
		return 1;
	 }
	return 0;
} /* copy_packcall */

int seed_start_call(struct p_data *p,char *datastr)
/* bookkeeping */
{
  int draw_flag=1,n=6,pnum=pack_num(p);
  double halfheight,realheight;

  stripsp(datastr);
  if (!strncmp(datastr,"-q",2)) /* -q flag for quiet, don't draw */
    {
      draw_flag=0;
      datastr[0]=datastr[1]=' ';
      stripsp(datastr);
    }
  if (!sscanf(datastr,"%d",&n) || n<3 || n>MAX_PETALS) n=6;
  if (!seed_pack(p,-1,n)) return 0; /* default geom is hyp */

/* set screen stuff */
  p->screen->box=std_real_box;
  realheight=(p->screen->box.ry-p->screen->box.ly);
  halfheight=realheight*p->screen->pix_box.rx/p->screen->pix_box.ry/2.0;
  p->screen->box.lx=(-1)*halfheight;
  p->screen->box.rx=halfheight;
  if (realheight<=0.0 
      || !(p->screen->pix_per_unit
	   =(int)((double)(p->screen->pix_box.ry-p->screen->pix_box.ly)
		  /realheight)))
    p->screen->pix_per_unit=(int)CANV_SIZE/2; /* default */
  p->screen->cflag_opt=1;
  p->screen->fflag_opt=0;
  sa_send_flag_opts(p);
  p->screen->unitcircle=1;
  live_node[pnum]=live_face[pnum]=0;
  sprintf(buf,"new seed %d",n);
  strcpy(p->file_name,buf);
  sprintf(buf,"no header");
  strcpy(p->fhead,buf);
  sprintf(buf,"Seed_pack (hyp)");
//  set_pack_label(p);
  canvas_mode_defaults(p);
  
  // calls to java in sa_XML
  new_pack_setup(p);

/* repack/fix/draw */
  sprintf(buf,"repack -p%d",pnum);
  handle_cmd(buf,&current_p);
  sprintf(buf,"fix -p%d -c",pnum);
  handle_cmd(buf,&current_p);
  if (draw_flag)
    {
      sprintf(buf,"disp -p%d -w -c",pnum);
      handle_cmd(buf,&current_p);
    }
  return 1;
} /* seed_start_call */

int empty_pack(struct p_data *p)
/* reset pack, default size =5000. */
{
	int pnum=pack_num(p);

	alloc_pack_space(p,5000,0); // set up for new data
	p->screen->box.lx=p->screen->box.ly=-1.1;
  	p->screen->box.rx=p->screen->box.ry=1.1;
	canvas_mode_defaults(p); // inform Java
	new_pack_setup(p);
	return 1;
} /* empty_pack */

int confirm_continue(char *message)
/* alert package for confirming continuation in script */
{
  /* fixup: need GUI mechanism? */
  return 1;
} /* confirm_continue */

int print_inq_data(struct p_data *p,int f,char *datastr)
     /* f is code for desired action; see sa_cmd. */
{
  int i,j,k,v,count=0,n1,n2,vv[3],v1,v2,v3,v4,length=0;
  int num,pnum=pack_num(p),hits,incompat_flag,*genlist;
  int max_ideg,min_ideg,max_bdeg,min_bdeg,digits=1;
  double dum,b_dist,nc,dumf,duml,tf,tl,tc;
  double b_vert_dist,datum,CPrad1,CPrad2,inv_d,act_inv_d,diff;
  double over_err,ic_err,e1,e2,e3,e4,e5;
  /* double accum_ic_err=0.0,ic_mix_max=0.0,ic_large_max=0.0,
     ic_small_max=0.0, worst_as_err=0.0,accum_as_err=0.0;*/
  complex ctr,e_cent,z;
  Mobius mob;
  char buff[256],buff2[256],*endptr,format_buf[256];
  struct s_data *scr;
  struct Vertlist *vertlist=NULL,*trace;
  struct Edgelist *edgelist=NULL,*track;
  struct p_data *p1,*p2;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;
  
  pK_ptr=p->packK_ptr;pR_ptr=p->packR_ptr;
  while ((pow(10.,(double)digits))*toler < 0.1 
	 && digits<MAX_ACCUR) digits++; /* for output accuracy */
  if (f==21) /* program parameters */
    {
      sprintf(msgbuf,"Param: Packing: accuracy=%e  "
	      "(okerr=%e);\n  Packing algorithm cycles=%d;  "
	      "iterations=%d; \n  Misc: post_size=%d; brush=%d;"
	      "act pack=%d; act node=%d",
	      (singledouble)toler,(singledouble)okerr,
	      totalpasses,iterates,sc_print_size,
	      line_thick,pnum,p->active_node);
      msg();
    }
  if (!p->status)
    {
      sprintf(msgbuf,"Pack %d is empty.",pnum);
      msg();
      return 0;
    }
  if (f==34) /* compute area */
    {
      if (p->hes)
	{
	  sprintf(msgbuf,"p%d is not euclidean.",pack_num(p));
	  emsg();
	  return 0;
	}
      dum=0.0;
      for (i=1;i<=p->facecount;i++)
	dum += e_area(pR_ptr[p->faces[i].vert[0]].rad,
		      pR_ptr[p->faces[i].vert[1]].rad,
		      pR_ptr[p->faces[i].vert[2]].rad);
      sprintf(msgbuf," eucl area=%f; ",(singledouble)dum);
      msg();
      return 1;
    }
  scr=&screendata[pnum];
  switch (f)
    {
case 1: /* pack data (pdata) */
 {
	/* Some global stuff set, but Java gets most info via JNI calls;
	 * see Interface.cpp
	 */
   if (pnum<0 || pnum>=NUM_PACKS) pnum=current_p;
   if (!packdata[pnum].status) return 1;
   p1=&packdata[pnum];

   max_bdeg=0;
   for (v=1;v<=p1->nodecount;v++)
     if (p->packK_ptr[v].bdry_flag && (p->packK_ptr[v].num+1)>max_bdeg)
       max_bdeg=p->packK_ptr[v].num+1;

   min_bdeg=max_bdeg;
   for (v=1;v<=p1->nodecount;v++)
     if (p->packK_ptr[v].bdry_flag && (p->packK_ptr[v].num+1)<min_bdeg)
       min_bdeg=p->packK_ptr[v].num+1;

   min_ideg=max_ideg=p->packK_ptr[p->alpha].num;
   for (v=1;v<=p1->nodecount;v++)
     if (!p->packK_ptr[v].bdry_flag)
       {
	 i=p->packK_ptr[v].num;
	 if (i>max_ideg) max_ideg=i;
	 if (i<min_ideg) min_ideg=i;
       }

   max_ideg_global=max_ideg;
   min_ideg_global=min_ideg;
   max_bdeg_global=max_bdeg;
   min_bdeg_global=min_bdeg;
   for (i=1;p1->edge_pair[i].edge;i++)
     length=i;
   length_global = length;

   /* what's the topology? (Not all cases get a result here) */
   topology_global[0]='\0';
   if (!(p1->genus)) /* genus zero */
     {
       if (!p1->num_bdry_comp) /* sphere */
	 strcat(topology_global,"sphere");
       else if (p1->num_bdry_comp==1) /* disc */
	 strcat(topology_global,"disc");
       else if (p1->num_bdry_comp==2) /* annulus */
	 strcat(topology_global,"annulus");
	   else /* n-connected, planar */ {
		   char buffer[64];
		   itoA(p1->num_bdry_comp,buffer,10);
		strcat(topology_global,buffer);
		strcat(topology_global,"-connected");
	   }
     }
   else if (p1->genus>0 && !(p1->num_bdry_comp)) { /* n-torus */
     char buffer[64];
	 itoA(p1->genus,buffer,10);
	 strcat(topology_global,buffer);
	 strcat(topology_global,"-torus");
   }
   first_face_global = p1->first_face;
   first_red_face_global = p1->first_red_face;
   num_bdry_comp_global = p1->num_bdry_comp;

/* Got to check the area computation; can be done in Interface.cpp instead. */  
/*
   if (p->hes==0) 
     {
       dum=0.0;
       for (i=1;i<=p->facecount;i++)
	 dum += e_area(pR_ptr[p->faces[i].vert[0]].rad,
		       pR_ptr[p->faces[i].vert[1]].rad,
		       pR_ptr[p->faces[i].vert[2]].rad);
       sprintf(msgbuf+strlen(msgbuf),"%f ",(singledouble)dum);
     }
   else // hyperbolic/spherical area = (1/-1)* #faces-sum(angle sums) 
     {
       dum=0.0;
       for (i=1;i<=p->nodecount;i++) dum += pR_ptr[i].curv;
       if (p->hes<0)
	 area_global=(p->facecount)*M_PI-dum;
       else
	 area_global=dum-(p->facecount)*M_PI;
     }
*/

   inv_dist_global = p1->overlap_status;
   return 1;
 }
case 2: /* vertex data */
  {
    if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) != NULL)
      {
	trace=vertlist;
	do 
	  {
	    v=trace->v;
	    sprintf(msgbuf,"\nData on vert# %d, p%d: rad=% .5e, "
		    "ang sum=% .5ePi, aim=% .5ePi, bdry?=%d, "
		    "#star=%d, mark=%d, plot_flag? %d, color? %d.",
	      v,pnum,(singledouble)radius(p,v),
	      (singledouble)(p->packR_ptr[v].curv/(singledouble)M_PI),
	      (singledouble)(p->packR_ptr[v].aim/(singledouble)M_PI),
	      p->packK_ptr[v].bdry_flag,p->packK_ptr[v].num,
	      p->packK_ptr[v].mark,p->packK_ptr[v].plot_flag,
	      p->packK_ptr[v].color);
	    msg();
	    count++;
	    trace=trace->next;
	  } while (trace!=NULL && count < 5);
	vert_free(&vertlist); /* free any excess */
	return count;
      }
    else v=p->active_node;
    sprintf(msgbuf,"\nData on vert# %d, p%d: rad=% .5e, "
	    "ang sum=% .5ePi, aim=% .5ePi, bdry?=%d, "
	    "#star=%d, mark=%d, plot_flag? %d, color? %d.",
      v,pnum,(singledouble)radius(p,v),
      (singledouble)(p->packR_ptr[v].curv/(singledouble)M_PI),
      (singledouble)(p->packR_ptr[v].aim/(singledouble)M_PI),
      p->packK_ptr[v].bdry_flag,p->packK_ptr[v].num,
      p->packK_ptr[v].mark,p->packK_ptr[v].plot_flag,
      p->packK_ptr[v].color);
    msg();
    return 1;
  }
case 7: /* flower */
case 8: /* selected radii */
case 9: /* selected angle sums */
case 10: /* selected aims */
case 15: /* selected model curvatures */
 {
   if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) )
     {
       if (f==7) sprintf(msgbuf,"Flowers:");
       else if (f==8) sprintf(msgbuf,"Radii:");
       else if (f==9) sprintf(msgbuf,"Angle sums/Pi:");
       else if (f==10) sprintf(msgbuf,"Aims/Pi:");
       else if (f==15) sprintf(msgbuf,"Kappa:");
       trace=vertlist;
       do 
	 {
	   if (f==8) datum=radius(p,trace->v);
	   else if (f==9) 
	     datum=p->packR_ptr[trace->v].curv/(singledouble)M_PI;
	   else if (f==10) 
	     datum=p->packR_ptr[trace->v].aim/(singledouble)M_PI;
	   else if (f==15)
	     {
	       kappa_calc(p,trace->v,&datum);
	       datum=datum*(fabs(datum));
	     }
	   else if (f==7)
	     {
	       sprintf(buff,"\np%d,v%d,deg=%d,flower: ",
		       pnum,trace->v,
		       p->packK_ptr[trace->v].num);
	       strcat(msgbuf,buff);
	       for (i=0;i<=p->packK_ptr[trace->v].num;i++)
		 {
		   sprintf(buff,"%d  ",
			   p->packK_ptr[trace->v].flower[i]);
		   strcat(msgbuf,buff);
		 }
	       *buff='\0';
	     }
	   if (f!=7) 
	     sprintf(buff,"%d % .5e; ",trace->v,(singledouble)datum);
	   strcat(msgbuf,buff);
	   count++;
	   trace=trace->next;
	 } while (trace!=NULL && count < 5);
       vert_free(&vertlist); /* free any excess */
       msg();
     }
   return count;
 }
case 16: /* antipodal point */
  {
    endptr=datastr;
    if (!(v=grab_one_vert(p,&endptr))) v=p->alpha;
    for (i=1;i<=p->nodecount;i++) p->packK_ptr[i].util_flag=0;
    p->packK_ptr[v].util_flag=1;
    if (!(genlist=label_generations(p,-1,&j,&i)))
      return 0;
    free(genlist);
    sprintf(msgbuf,"Antipodal (max comb distance) to %d is "
	       "vertex %d.",v,j);
    msg();
    return 1;
  }
case 11: /* data on selected faces */
  {
    if ( (vertlist=face_link_parse(p,datastr,&endptr,&hits)) != NULL)
      {
	trace=vertlist;
	do
	  {
	    v=trace->v;
	    incompat_flag=(int)0.0;
	    for (j=0;j<3;j++) vv[j]=p->faces[v].vert[j];
	    sprintf(msgbuf,"face %d data, p%d: verts (%d,%d,%d), "
		    "places circle %d, \n  color? %d, plot_flag? %d.  ",
	      v,pnum,vv[0],vv[1],vv[2], 
		    vv[(p->faces[v].index_flag+2)%3],
		    p->faces[v].color,p->faces[v].plot_flag);
	    msg();
	    sprintf(msgbuf,"  Rad: (%f,%f,%f).\n  Angles: (%f,%f,%f)*Pi.\n",
		    (singledouble)radius(p,vv[0]),
		    (singledouble)radius(p,vv[1]),
		    (singledouble)radius(p,vv[2]),
		    (singledouble)comp_single_angle(p,vv[0],
				  vv[1],vv[2],&incompat_flag)/M_PI,
		    (singledouble)comp_single_angle(p,
				  vv[1],vv[2],vv[0],&incompat_flag)/M_PI,
		    (singledouble)comp_single_angle(p,
				  vv[2],vv[0],vv[1],&incompat_flag)/M_PI);
	    msg();
	    if (incompat_flag)
	      {
		sprintf(msgbuf," NOTE: radii/overlaps incompatible.\n");
		emsg();
	      }
	    sprintf(msgbuf,"  Overlaps (0->1->2->0): assigned=");
	    if (p->overlap_status)
	      sprintf(buff,"%f,%f,%f\n   errors "
		      "(assigned-actual)=%f,%f,%f.\n",
		    (singledouble)pK_ptr[vv[0]].overlaps[nghb(p,vv[0],vv[1])],
		    (singledouble)pK_ptr[vv[1]].overlaps[nghb(p,vv[1],vv[2])],
		    (singledouble)pK_ptr[vv[2]].overlaps[nghb(p,vv[2],vv[0])],
		    (singledouble)invdist_err(p,vv[0],vv[1]),
		    (singledouble)invdist_err(p,vv[1],vv[2]),
		    (singledouble)invdist_err(p,vv[2],vv[0]));
	    else sprintf(buff,"tangency\n   errors "
			 "(assigned-actual)=%f,%f,%f.\n",
		    (singledouble)invdist_err(p,vv[0],vv[1]),
		    (singledouble)invdist_err(p,vv[1],vv[2]),
		    (singledouble)invdist_err(p,vv[2],vv[0]));
	    strcat(msgbuf,buff);
	    msg();
	    count++;
	    trace=trace->next;
	  } while (trace!=NULL && count < 5);
	vert_free(&vertlist); /* free any excess */
      }
    return count;
  }
case 12: /* overlaps */
 {
   sprintf(msgbuf,"Overlaps, p%d:",pnum);
   if (!p->overlap_status)
     {
       strcat(msgbuf," All set to 0.0, (default=tangency case).");
       msg();
       return count;
     }
   if ((edgelist=node_pair_link(p,datastr,&endptr,&hits))!=NULL)
     {
       track=edgelist;
       do {
	 inv_d=pK_ptr[track->v].
	   overlaps[nghb(p,track->v,track->w)];
	 if (inv_d>=0 && inv_d<=1) /* want overlap */
	   {
	     act_inv_d=comp_inv_dist(p,track->v,track->w);
	     if (act_inv_d<=1)
	       sprintf(buff," (%d,%d): aim=cos(% .5e Pi); "
		       "actual=cos(% .5e Pi).\n",
		       track->v,track->w,
		       (singledouble)acos(inv_d)/(singledouble)M_PI,
		       (singledouble)acos(act_inv_d)/(singledouble)M_PI);
	     else
	       sprintf(buff," (%d,%d): aim=cos(% .5e Pi); "
		       "actual=*% .5e.\n",
		       track->v,track->w,
		       (singledouble)(acos(inv_d)/(singledouble)M_PI),
		       (singledouble)act_inv_d);
	     strcat(msgbuf,buff);
	     count++;
	   }
	 else if (inv_d>1) /* want separated */
	   {
	     act_inv_d=comp_inv_dist(p,track->v,track->w);
	     if (act_inv_d>1)
	       sprintf(buff," (%d,%d): aim=*% .5e; "
		       "actual=*% .5e.\n",
		       track->v,track->w,(singledouble)inv_d,
		       (singledouble)act_inv_d);
	     else
	       sprintf(buff," (%d,%d): aim=*% .5e; "
		       "actual=cos(% .5e Pi).\n",
		       track->v,track->w,(singledouble) inv_d,
		       (singledouble)(acos(act_inv_d)/(singledouble)M_PI));
	     strcat(msgbuf,buff);
	     count++;
	   }
	 else /* assigned overlaps exceeding Pi/2 */
	   {
	     act_inv_d=comp_inv_dist(p,track->v,track->w);
	     if (act_inv_d>1)
	       sprintf(buff," (%d,%d): aim=cos(% .5e Pi) "
		       "(NOTE: > Pi/2); actual=*% .5e.\n",
		       track->v,track->w,
		       (singledouble)(acos(inv_d)/(singledouble)M_PI),
		       (singledouble)act_inv_d);
	     else
	       sprintf(buff," (%d,%d): aim=cos(% .5e Pi) "
		       "(NOTE: > Pi/2); actual=cos(% .5e Pi).\n",
		       track->v,track->w,
		       (singledouble)acos(inv_d)/(singledouble)M_PI,
		       (singledouble)acos(act_inv_d)/(singledouble)M_PI);
	     
	     strcat(msgbuf,buff);
	     count++;
	   }
	 track=track->next;
       } while (track!=NULL);
       edge_free(&edgelist);
     }
   msg();
   return count;
 } 
case 14: /* altn_rad alternative radii: eucl, hyp. */
 {
   if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) != NULL)
     {
       if (p->hes<0) sprintf(msgbuf,"Radii, p%d (v  eucl, hyp):",
			     pnum);
       else if (p->hes==0) sprintf(msgbuf,"Eucl:");
       else sprintf(msgbuf,"Sph:");
       trace=vertlist;
       do {
	 if (p->hes<0)		   
	   h_to_e_data(p->packR_ptr[trace->v].center,
		       p->packR_ptr[trace->v].rad,&e_cent,&datum);
	 else datum=p->packR_ptr[trace->v].rad;
	 sprintf(buff," %d  % .5e, ",
		 trace->v,(singledouble)datum);
	 strcat(msgbuf,buff);
	 if (p->hes<0)
	   {
	     datum=p->packR_ptr[trace->v].rad;
	     if (datum<=0) sprintf(buff,"inf ");
	     else 
	       {
		 datum=(-log(datum));
		 sprintf(buff,"% .5e ",(singledouble)datum);
	       }
	     strcat(msgbuf,buff);
	   }
	 count++;
	 trace=trace->next;
       } while (trace!=NULL && count < 5);
       vert_free(&vertlist); /* free any excess */
       msg();
     }
   return count;
 }
case 13: /* selected centers */
 {
   sprintf(msgbuf,"Centers: ");
   if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) != NULL)
     {
       trace=vertlist;
       do {
	 ctr=p->packR_ptr[trace->v].center;
	 sprintf(buff,"%d (% .12e % .12e),\n |.|=% .12e ; ",
		 trace->v,(singledouble)ctr.re,(singledouble)ctr.im,
		 (singledouble)cAbs(ctr));
	 strcat(msgbuf,buff);
	 count++;
	 trace=trace->next;
       } while (trace!=NULL && count < 3);
       vert_free(&vertlist); /* free any excess */
       msg();
     }
   return count;
 }
case 20: /* ?screendata */
 {
   sprintf(msgbuf,
	   "Screen %d: lower left=(%.6e, %.6e), upper right=(%.6e,%.6e).\n"
	   "  width=%.6e.",
	   pnum,(singledouble)(scr->box.lx),(singledouble)(scr->box.ly),
	   (singledouble)(scr->box.rx),(singledouble)(scr->box.ry),
	   (singledouble)(scr->box.rx-scr->box.lx));
   msg();
   return 1;
 }
case 25: /* ?pk_status: curvature/overlap errors */
 {
/* Collect data, set vertex marks with codes: 
  -1 = not subject to change
  0 = packed within tolerance
  1 = incompat, too small in some faces
  2 = incompat, too large in some faces
  3 = incompat, mixed (too large in some faces, too small in others)
*/

   double accum_ic_err=0.0,ic_mix_max=0.0,ic_large_max=0.0;
   double ic_small_max=0.0,worst_as_err=0.0,accum_as_err=0.0;
   int ic_small_vert=0,ic_large_vert=0,ic_mix_vert=0;
   int free_count=0,worst_as_vert=0;

   for (i=1;i<=p->nodecount;i++)
     {
       if (p->packR_ptr[i].aim<0) p->packK_ptr[i].mark=-1;
       else 
	 {
	   free_count++;
	   if (p->hes<0)
	     h_as_overlap(p,i,p->packR_ptr[i].rad,
	       &p->packR_ptr[i].curv,&p->packK_ptr[i].mark,&ic_err);

	   if (!p->packK_ptr[i].mark)
	     {
	       diff=fabs(p->packR_ptr[i].curv-p->packR_ptr[i].aim);
	       if (diff > worst_as_err) 
		 {
		   worst_as_err=diff;
		   worst_as_vert=i;
		 }
	       accum_as_err += diff;
	     }
	   else if (p->packK_ptr[i].mark==1)
	     {
	       if (ic_err>ic_small_max) 
		 {
		   ic_small_max=ic_err;
		   ic_small_vert=i;
		 }
	       accum_ic_err += ic_err;
	     }
	   else if (p->packK_ptr[i].mark==2)
	     {
	       if (ic_err>ic_large_max)
		 {
		   ic_large_max=ic_err;
		   ic_large_vert=i;
		 }
	       accum_ic_err += ic_err;
	     }
	   else if (p->packK_ptr[i].mark==3)
	     {
	       if (ic_err>ic_mix_max)
		 {
		   ic_mix_max=ic_err;
		   ic_mix_vert=i;
		 }
	       accum_ic_err += ic_err;
	     }
	 }
     }
   over_err=0.0;
   if (p->overlap_status)
     for (i=1;i<p->nodecount;i++)
       for (j=0;j<p->packK_ptr[i].num;j++)
	 {
	   if (i<(v=p->packK_ptr[i].flower[j]))
	     {
	       diff=fabs(p->packK_ptr[i].overlaps[j]-
			 comp_inv_dist(p,i,v));
	       if (diff > okerr && diff > over_err)
		 {n1=i;n2=v;over_err=diff;} 
	       /* worst, so far */
	     }
	 }
 
 /* printing results */

   sprintf(msgbuf,"Pack status, p%d: %d verts subject to change.",
	   pack_num(p),free_count);
   msg();
   sprintf(format_buf,"   %% .%de = accum ang sum err; "
	   "worst = %% .%de at v%%d.",digits,digits);
   sprintf(msgbuf,format_buf,(singledouble)accum_as_err,
	   (singledouble)worst_as_err,worst_as_vert);
   msg();
   if (p->overlap_status && accum_ic_err > 0.0)
     {
       sprintf(msgbuf,"   INCOMPATIBILITIES have occurred; verts "
	       "marked:\n          1 = too small,2 = too large, 3 = mixed.");
       msg();
       sa_window_bell();
       sprintf(msgbuf,"   %f = accum incompatibility err",
	       (singledouble)accum_ic_err);
       if (ic_small_vert)
	 {
	   sprintf(buff,"\n     worst too small = %f at v%d",
		   (singledouble)ic_small_max,ic_small_vert);
	   strcat(msgbuf,buff);
	 }
       if (ic_large_vert)
	 {
	   sprintf(buff,"\n     worst too large = %f at v%d",
		   (singledouble)ic_large_max,ic_large_vert);
	   strcat(msgbuf,buff);
	 }
       if (ic_mix_vert)
	 {
	   sprintf(buff,"\n     worst mixed = %f at v%d.",
		   (singledouble)ic_mix_max,ic_mix_vert);
	   strcat(msgbuf,buff);
	 }
       msg();
       if (over_err>0.0)
	 {
	   sprintf(msgbuf,"   OVERLAP ERRORS: worst is %f, edge (%d,%d).",
	       (singledouble)over_err,n1,n2);
	   msg();
	   sa_window_bell();
	 }
       else
	 {
	   sprintf(msgbuf,"   Overlaps within tolerances.");
	   msg();
	 }
     }
   return 1;
 }
case 26: /* distance to bdry */
 {
	h_to_e_data(p->packR_ptr[p->alpha].center,
		p->packR_ptr[p->alpha].rad,&ctr,&dum);
	bdry_dist(p,ctr,&b_dist,&b_vert_dist);
	sprintf(msgbuf,"Euclidean distance from alpha to the "
		"boundary polygon is % .5e.",(singledouble)b_dist);
	msg();
	return 1;
 }
case 27: /* edge_pairing. Give mobius transforms */
 {
    n1=0;
    int hit=0;
    int e=1;
    while (p->edge_pair[e].edge) {
		if ((n1=find_mate_index(p,e))>0) { /* found mate, print mobius */
      		if (!hit) { // first pair?
				sprintf(msgbuf,"Edge pairings for p%d:\n",pnum);
			    msg();
			    hit++;
      		}
		    sprintf(msgbuf,"  Pairing %d <--> %d as 2x2 matrix:\n",e,n1);
		    msg();
		    msg_mobius_info(p->edge_pair[e].mob);
      	}
      	e++;
    }
	if (!hit) {
      sprintf(msgbuf,"No edge pairing for p%d.",pnum);
      msg();
    }
    return 1;
 }
case 28: /* ratio_ftn <p,q>. Gives (eucl radius of alpha vert of q)
divided by (eucl rad of alpha vert of p) */
 {
	if (sscanf(datastr,"%d %d",&n1,&n2)!=2 
		|| n1< 0 || n1>=NUM_PACKS || !packdata[n1].status 
		|| n2< 0 || n2>=NUM_PACKS || !packdata[n2].status)
	 {sprintf(msgbuf,"Specify two active packs.");emsg();return 0;}
	p1=&packdata[n1];p2=&packdata[n2];
	if (p1->hes>0 || p2->hes>0)
	 {sprintf(msgbuf,"One of packs is spherical.");emsg();return 0;}
	if (p1->hes<0) h_to_e_data(p1->packR_ptr[p1->alpha].center,
	    p1->packR_ptr[p1->alpha].rad,&e_cent,&CPrad1);
	else CPrad1=p1->packR_ptr[p1->alpha].rad;
	if (p2->hes<0) h_to_e_data(p2->packR_ptr[p2->alpha].center,
	    p2->packR_ptr[p2->alpha].rad,&e_cent,&CPrad2);
	else CPrad2=p2->packR_ptr[p2->alpha].rad;
	sprintf(msgbuf,"Radii comparisons (for alpha vert):\n  "
		"Eucl:  ratio p%d/p%d = % .5e;     "
		"difference p%d-p%d = % .5e.",
		n2,n1,(singledouble)(CPrad2/CPrad1),n2,n1,
		(singledouble)(CPrad2-CPrad1));
	msg();
	if (p->hes<0)
	 {
		CPrad2=-log(p2->packR_ptr[p2->alpha].rad);
		CPrad1=-log(p1->packR_ptr[p1->alpha].rad);
		sprintf(msgbuf,"  Hyp:   ratio p%d/p%d = % .5e;   "
			"difference p%d-p%d = % .5e.",
		n2,n1,(singledouble)(CPrad2/CPrad1),n2,n1,
			(singledouble)(CPrad2-CPrad1));
		msg();
	 }
	return 1;
 }
case 29: /* give conductances, transition probabilities for nodes. */
 {
	if (p->hes>=0) return 0;  /* Only for hyp case at this time. */
	if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) )
	 {
		trace=vertlist;
		do {
		   v=trace->v;
		   num=pK_ptr[v].num;
		   sprintf(msgbuf,"Random walk data, pack %d, "
			   "vert %d: neighbors = ",
		   	pnum,v);
		   for (i=0;i<=num;i++) 
		    {
			sprintf(buff," %d",pK_ptr[v].flower[i]);
			strcat(msgbuf,buff);
		    }
		   msg();
		   sprintf(msgbuf,"  Conductances:  total node = % .5e; "
			   "twin = % .5e",
			   (singledouble)(nc=node_conductance(p,v)),
			   (singledouble)(tc=twin_conductance(p,v)));
		   msg();
		   dumf=Cvw(pR_ptr[v].rad,0.0,
			pR_ptr[pK_ptr[v].flower[0]].rad,
			pR_ptr[pK_ptr[v].flower[1]].rad,1,p->hes);
		   tf=dumf/nc;
		   duml = Cvw(pR_ptr[v].rad,
			pR_ptr[pK_ptr[v].flower[num-1]].rad,
			pR_ptr[pK_ptr[v].flower[num]].rad,0.0,2,p->hes);
		   tl=duml/nc;
		   if (!pK_ptr[v].bdry_flag) {dumf += duml;tf += tl;}
		   sprintf(msgbuf,"  % .5e",(singledouble)dumf);
		   sprintf(buff2,"  % .5e",(singledouble)tf);
		   for (i=1;i<num;i++)
		    {
			dum=Cvw(pR_ptr[v].rad,
				pR_ptr[pK_ptr[v].flower[i+1]].rad,
				pR_ptr[pK_ptr[v].flower[i]].rad,
				pR_ptr[pK_ptr[v].flower[i-1]].rad,0,p->hes);
			tf=dum/nc;
			sprintf(buff,"  % .5e",(singledouble)dum);
			strcat(msgbuf,buff);
			sprintf(buff,"  % .5e",(singledouble)tf);
			strcat(buff2,buff);
		    }
		   if (pK_ptr[v].bdry_flag)
		    {
			sprintf(buff,"  % .5e",(singledouble)duml);
			strcat(msgbuf,buff);
			sprintf(buff,"  % .5e",(singledouble)tl);
			strcat(buff2,buff);
		    }
		   msg(); 	/* print conductances */
		   sprintf(msgbuf,"  Transition Probabilities:  "
			   "leakage = % .5e\n",
			(singledouble)(tc/nc));
		   strcat(msgbuf,buff2);
		   msg(); 	/* print transition probabilities */
		   count++;
		   trace=trace->next;
		 } while (trace!=NULL && count < 5);
		vert_free(&vertlist); /* free any excess */
	 }
	return count;
 }
case 31: /* length of bdry */
 {
	if ( (vertlist=node_link_parse(p,datastr,&endptr,&hits)) != NULL)
	 {
	   sprintf(msgbuf,"Bdry counts:");
	   trace=vertlist;
	   do {
	     if ((hits=bdry_comp_count(p,trace->v)))
	       {
		 sprintf(buff," v%d --> count=%d, ",trace->v,hits);
		 strcat(msgbuf,buff);
		 count++;
	       }
	     trace=trace->next;
	   } while (trace!=NULL && count < 5);
	   vert_free(&vertlist);
	   msg();
	   return 1;
	 }
	sprintf(msgbuf,"Bdry length? Given vert doesn't seem to be "
		"in the boundary.");
	msg();
	return 0;
 }
case 35: /* vertex_map indices */
case 36: /* vertex_map reverse indices */
 {
   if (!p->vertex_map)
     {
       sprintf(msgbuf,"map? No 'vertex_map' recorded for this packing.");
       msg();
       return 0;
     }
   sprintf(msgbuf,"Vertex_map:");
   if (f==35 && (vertlist=node_link_parse(p,datastr,&endptr,&hits)) )
     {
       /* keep only up to 12 in list */
       trace=vertlist;
       do 
	 {
	   trace=trace->next;
	   count++;
	 }
       while (trace && count<12);
       if (trace) vert_free(&trace); 
     }
   else if (f==36 && sscanf(datastr,"%d",&v))
     {
       vertlist=(struct Vertlist *)calloc((size_t)1,sizeof(struct Vertlist));
       vertlist->v=v;
     }
   trace=vertlist;
   if (!trace) return 0;
   count=0;
   do
     {
       if ( (f==35 && (v=vertex_translation(p->vertex_map,trace->v))>0)
	    || (f==36 
		&& (v=vertex_rev_translation(p->vertex_map,trace->v))>0) )
	 {
	   if (f==35) sprintf(buff," %d --> %d,\0",trace->v,v);
	   else  sprintf(buff," %d <-- %d,\0",trace->v,v);
	   strcat(msgbuf,buff);
	   strcat(msgbuf,"\0");
	   count++;
	 }
       trace=trace->next;
     }
   while (trace);
   vert_free(&vertlist); /* free any excess */
   if (!count)
     {
       if (f==35) sprintf(msgbuf,"?map. No corresponding indices found.");
       else sprintf(msgbuf,"?map_rev. No indices with these "
		    "images found.");
     }
   msg();
   return count;
 }
case 37: /* edge_x_ratio */
 {
   sprintf(msgbuf,"Edge cross_ratios, p%d:",pnum);
   if ((edgelist=node_pair_link(p,datastr,&endptr,&hits))!=NULL)
     {
       track=edgelist;
       do {
	 if ((j=nghb(p,(v1=track->v),(v3=track->w)))>=0 
	     && (k=nghb(p,v3,v1))>=0 
	     && j!=pK_ptr[v1].num && k!=pK_ptr[v3].num){
	   v4=pK_ptr[v1].flower[j+1];
	   v2=pK_ptr[v3].flower[k+1];
	   if (p->overlap_status){ /* have overlaps to use */
	     e1=pK_ptr[v1].overlaps[nghb(p,v1,v2)];
	     e2=pK_ptr[v3].overlaps[k];
	     e3=pK_ptr[v3].overlaps[nghb(p,v3,v4)];
	     e4=pK_ptr[v1].overlaps[j+1];
	     e5=pK_ptr[v1].overlaps[j];
	   }
	   else e1=e2=e3=e4=e5=1.0;
	   z=quad_cross_ratio(pR_ptr[v1].rad,pR_ptr[v2].rad,
		   pR_ptr[v3].rad,pR_ptr[v4].rad,
		   e1,e2,e3,e4,e5);		 
	   sprintf(msgbuf," edge (%d,%d): cross_ratio = (%lf + i %lf).\n",
		   track->v,track->w,
		   (singledouble)z.re,(singledouble)z.im);
	   msg();
	   count++;
	 }
	 track=track->next;
       } while (track!=NULL && count<5);
     }
   else {
     sprintf(msgbuf,"?edge_x_ration: need list of edges.");
     msg();
   }
   edge_free(&edgelist);
   return count;
 }
	 } /* end of switch */
	return 0;
} /* print_inq_data */

int fexec_call(char *datastr)
/* open file, execute commands */
{
	int count=0;
	char filename[256],command[BUFSIZE];
	FILE *fp;
	
	sscanf(datastr,"%s",filename);
	if ((fp=fopen(filename,"r"))==NULL)
	 {
	 	sprintf(msgbuf,"fexec error opening %s.",
	 		filename);
	 	emsg();
	 	return count;
	 }
	while (fgets(command,BUFSIZE-1,fp) && command!='\0')
	 	count += handle_cmd(command,&current_p);
	fclose(fp);
	return count;
} /* fexec_call */
	
/* ====================== shmem routines ======================
shared memory routines from Monica Hurdal and Kurt Shaper (Minn)
to pass commands between CirclePack and IDL (or other external
program) */

int c_initmem()
{
	#ifndef WIN32 // FA
  int shmid;
  key_t key;
  int blockSize;
  char *keyName;

  /* create the key */
  keyName = (char *) getenv("SHMEMKEY");
  if (keyName == NULL)
    keyName = "shmem.key";
  key = ftok(keyName,'c');
  if ((int)key < 0) {
    perror("error creating key");
    return(1);
  }
  
  /* create share memory block and write id to interface file */
  blockSize = 3*sizeof(int) + MAXCMD*sizeof(char);
  shmid = shmget(key, blockSize, IPC_CREAT|SHMODE);
  if (shmid < 0) {
    perror("error getting shared memory block");
    return(2);
  }
  
  /* map the variable pointers to the shared memory block */
  shmBlock = (char *) shmat(shmid, 0, 0);
  if (!((long)shmBlock+1)) {
    perror("error attaching to shared memory block");
    return(3);
  }
  lockFlg = (int *) shmBlock;
  newFlag = (int *) (shmBlock+1*sizeof(int));
  vertNum = (int *) (shmBlock+2*sizeof(int));
  cmdStr = (char *) (shmBlock+3*sizeof(int));
  *lockFlg = 0;
#endif // FA
  return(0);
} /* c_initmem */

int c_rdmem(int *vnum,char *vstr)
{
#ifndef WIN32 // FA
  int loopCount;

  /* if not initialized, error */
  if (lockFlg == NULL) {
    perror("shared memory not initialized");
    return(1);
  }
  
  /* if nothing new, just bail */
  if (*newFlag == 0) {
    *vnum = *vertNum;
    strncpy(vstr,cmdStr,MAXCMD);
    return(-1);
  }
  
  /* lock memory area */
  loopCount = 0;
  while (*lockFlg != 0) {
    usleep(100);
    loopCount++;
    if (loopCount > 100) {
      perror("error locking memory segment");
      return(2);
    }
  }
  
  /* ---------------------- critical section ------------------- */
  *lockFlg = 1;
  *vnum = *vertNum;
  strncpy(vstr,cmdStr,MAXCMD);
  *newFlag = 0;
  *lockFlg = 0;
  /* ---------------------- critical section ------------------- */
#endif
  return(0);
} /* c_rdmem */

int c_wrmem(int *vnum,char *vstr)
{
	#ifndef WIN32 // FA
  int loopCount;
  
  /* if not initialized, error */
  if (lockFlg == NULL) {
    perror("shared memory not initialized");
    return(1);
  }
  
  /* lock memory area */
  loopCount = 0;
  while (*lockFlg != 0) {
    usleep(100);
    loopCount++;
    if (loopCount > 100) {
      perror("error locking memory segment");
      return(2);
    }
  }
  
  /* ---------------------- critical section ------------------- */
  *lockFlg = 1;
  *vertNum = *vnum;
  strncpy(cmdStr,vstr,MAXCMD);
  *newFlag = 1;
  *lockFlg = 0;
  /* ---------------------- critical section ------------------- */
#endif //FA
  return(0);
} /* c_wrmem */

int cmd_import(char *datastr)
/* import a string from shared memory */
{
  int stat;
  int vnum = 0;
  
  if (init_shmem == 0) {
    stat = c_initmem(); 
    if (stat > 0) {
      if (stat == 1) {
	sprintf(msgbuf,"File shmem.key not found.");
	emsg();
      }
      sprintf(msgbuf,"Not able to share memory with "
	      "external application.");
      emsg();
      return 0;
    }
  } else init_shmem = 1;
  stat = c_rdmem(&vnum,buf);
  if (stat > 0) {
    sprintf(msgbuf,"Error importing command from "
	    "external application.");
    emsg();
    return 0;
  } else if (stat < 0) {
    sprintf(msgbuf,"No change - nothing to do.");
    msg();
    return 1;
  } else {
    sprintf(msgbuf,"Importing %s",buf);
    msg();
    return(sort_cmd(buf,&current_p));
  }
} /* cmd_import */
         
int cmd_export(struct p_data *p,char *datastr)
/* export a string to shared memory */
{
  int stat,vnum = 0,hits;
  char *dpoint,cmd[128],outstr[128],*endptr;
  struct Vertlist *holdv;
  
  if (init_shmem == 0) {
    stat = c_initmem();
    if (stat > 0) {
      if (stat == 1) {
	sprintf(msgbuf,"File shmem.key not found.");
	emsg();
      }
      sprintf(msgbuf,"Not able to share memory with "
	      "external application.");
      emsg();
      return 0;
    }
  } else init_shmem = 1;
  dpoint=datastr;
  /* pre-processing datastr */
  if (!grab_next(&dpoint,cmd)) return 0; /* string was empty */
  switch(cmd[0]) {
  case 'c':
    {
      if (!strncmp(cmd,"center_vert",11)) /* ======= center_vert */
	{
	  if ((holdv=node_link_parse(p,dpoint,&endptr,&hits))!=NULL)
	    {
	      sprintf(outstr,"center_vert %d",holdv->v);
	      vnum = holdv->v;
	      vert_free(&holdv);
	    }
	  else goto BAIL;
	  break;
	}
      break;
    }
  }/* end of switch */
  if ((stat = c_wrmem(&vnum,outstr))>0) goto BAIL;
  sprintf(msgbuf,"Exporting %s",outstr);
  msg();
  return 1;	
 BAIL:
  sprintf(msgbuf,"Error exporting command to "
	  "external application.");
  emsg();
  return 0;
  
} /* cmd_export */

/* ---------------------------------- some technical routines ---------- */
	
int stat_file(char *filename)
/* returns 1 if filename exists */
{
  struct stat statbuf;
  return( (strlen(filename)<=NAME_MAX) && (stat(filename, &statbuf)>=0) );
} /* stat_file */

int post_canvas(struct p_data *p)
/* circles/triangles/edges/labels/etc using standard flags. */
{
	int cflag,fflag;
	char buff[NAME_MAX+128];

	if (!p->status) return 0;
	if (post_fp)
	  {
	    fprintf(post_fp,"\ngrestore\nend\nshowpage"); /* wrap up */
	    fclose(post_fp);
	    post_fp=NULL;
	  }
	switch (fill_mode)
	 {
		case 0: {graylevel=.5;break;/* grey */}
		case 1: {graylevel=1;break; /* white */}
		case 2: {graylevel=0;break; /* black */}
	 }
	ps_linewidth=(1+2*(line_thick-1))*PS_UNIT_LINEWIDTH;
	cflag=p->screen->cflag_opt;
	fflag=p->screen->fflag_opt;
	sprintf(buff," -o %s ",print_file_name);
	if (p->screen->unitcircle) strcat(buff," -u ");
    
	if (cflag & 1) /* circle options */
	  {
	    strcat(buff," -c");
	    if (cflag & 8) strcat(buff,"c");
	    if (!(cflag & 16) && (cflag & 4)) strcat(buff,"bg");
	    else if (!(cflag & 16) && (cflag & 2)) strcat(buff,"fg");
	    else if (cflag & 2) strcat(buff,"f");
	    strcat(buff," a");
	    if (cflag & 32) strcat(buff," -nc a");
	  }
	if (fflag & 1) /* face options */
	  {
	    strcat(buff," -f");
	    if (fflag & 8) strcat(buff,"c");
	    if (!(fflag & 16) && (fflag & 4)) strcat(buff,"bg");
	    else if (!(fflag & 16) && (fflag & 2)) strcat(buff,"fg");
	    else if (fflag & 2) strcat(buff,"f");
	    strcat(buff," a");
	    if (fflag & 32) strcat(buff," -nf a");
	  }
	print_call(&post_fp,p,print_file_name,
		   buff,user_name,CPVersion,print_cmd);
	fprintf(post_fp,"\ngrestore\nend\nshowpage"); /* wrap up */
	fclose(post_fp);
	post_fp=NULL;
	/* to save paper, pop up ghostview image; user can then print */
	/* print_psfile(&post_fp); */
	strcpy(buff,print_file_name);
	sysfp=popen("/bin/sh","w");
	fprintf(sysfp," %s %s &\n",ps_viewer,buff);
	pclose(sysfp);

	return 1;
} /* post_canvas */

int print_psfile(FILE **FP)
/* print canvas (flag=1) or custom (flag=2) file */
{
  char buff[64];
  FILE *sysfp;

  if (FP==&post_fp) 
    {
      strcpy(buff,print_file_name);
    }
  else if (FP==&custom_fp)
    {
      if (custom_fp) close_psfile(&custom_fp);
      strcpy(buff,ps_file_name);
    }
  else return 0;

  strcpy(buf,print_cmd);
  strcat(buf," ");
  strcat(buf,buff);
  sysfp=popen("/bin/sh","w");
  fprintf(sysfp,"%s\n",buf);
  pclose(sysfp);
  sprintf(msgbuf,"Have sent `%s' to the printer.",buff);
  msg();
  return 1;
} /* print_psfile */

int msg_mobius_info(Mobius mob)
     /* display info on given mobius transformation. */
{
  int m,inf_flag;
  double angle,length;
  complex z1,z2;

  if (!(m=mobius_fixed_pts(mob,&z1,&z2,&inf_flag,&angle,&length))) 
    return 0; /* improper mobius */
  /* print mob itself */
  sprintf(msgbuf,"     [ %f + %f*i, %f + %f*i;\n",
	  (singledouble)mob.a.re,(singledouble)mob.a.im,
	  (singledouble)mob.b.re,(singledouble)mob.b.im);
  msg();
  sprintf(msgbuf,"      %f + %f*i, %f + %f*i ]\n",
	  (singledouble)mob.c.re,(singledouble)mob.c.im,
	  (singledouble)mob.d.re,(singledouble)mob.d.im);
  msg();

  /* info depending on type */
  if (m==1) { /* elliptic */
    sprintf(msgbuf,"Elliptic: rotation angle %f.",
	    (singledouble)angle);
    msg();
  }
  else if (m==2) { /* parabolic */
    if (inf_flag) /* fixes infinity */
      sprintf(msgbuf,"Parabolic: euclidean translation by distance %f.",
	      (singledouble)length);
    else sprintf(msgbuf,"Parabolic.");
    msg();
  }
  else if (m==3) { /* hyperbolic */
    sprintf(msgbuf,"Hyperbolic: hyperbolic translation length %f.",
	    (singledouble)length);
    msg();
  }
  else if (m==4) { /* strict loxodromic */
    sprintf(msgbuf,"Strict Loxodromic: angle %f, factor %f.",
	    (singledouble)angle,(singledouble)length);
    msg();
  }
  /* fixed point(s) */
  if (inf_flag) {
    sprintf(msgbuf,"  Fixed points: infinity, %f + %f*i.",
	    (singledouble)z1.re,(singledouble)z1.im);
    msg();
    return 1;
  }
  sprintf(msgbuf,"  Fixed points: %f + %f*i,   %f + %f*i.",
	  (singledouble)z1.re,(singledouble)z1.im,
	  (singledouble)z2.re,(singledouble)z2.im);
  msg();
  return 1;
} /* msg_mobius_info */

/* Version from web to replace stdlib.h 'itoa' function,
   which is not in linux */

char* itoA(int value, char*  str, int radix)
{
    int  rem = 0;
    int  pos = 0;
    char ch  = '!' ;
    do
    {
        rem    = value % radix ;
        value /= radix;
        if ( 16 == radix )
        {
            if( rem >= 10 && rem <= 15 )
            {
                switch( rem )
                {
                    case 10:
                        ch = 'a' ;
                        break;
                    case 11:
                        ch ='b' ;
                        break;
                    case 12:
                        ch = 'c' ;
                        break;
                    case 13:
                        ch ='d' ;
                        break;
                    case 14:
                        ch = 'e' ;
                        break;
                    case 15:
                        ch ='f' ;
                        break;
                }
            }
        }
        if( '!' == ch )
        {
            str[pos++] = (char) ( rem + 0x30 );
        }
        else
        {
            str[pos++] = ch ;
        }
    }while( value != 0 );
    str[pos] = '\0' ;
    /* now have to reverse the string (but 'strrev' not on linux) */
    return strrev(str);
} /* itoA */

char *strrev(char *str)
{
      char *p1, *p2;

      if (! str || ! *str)
            return str;
      for (p1 = str, p2 = str + strlen(str) - 1; p2 > p1; ++p1, --p2)
      {
            *p1 ^= *p2;
            *p2 ^= *p1;
            *p1 ^= *p2;
      }
      return str;
} /* strrev */
