/*
 * separate_p_light_components:  takes a light packing pl as its first parameter,
 * fills in the componentList parameter with a list of light packings representing
 * the connected components of pl, and returns the size of the list.
 */

#include "sa_types.h"
#include "sa_proto.h"

int separate_p_light_components(struct p_light *pl, struct p_light ***componentListParm)

{
    int i, cursor;
    int numComponents;
    struct p_light **componentList;
    int *flower_heads, *flower_heads_backup, *visited;

    flower_heads = (int *) malloc(sizeof(int) * (pl->counts[2] + 1));
    memset(flower_heads, 0, sizeof(int) * (pl->counts[2] + 1));
    visited = (int *) malloc(sizeof(int) * (pl->counts[2] + 1));
    memset(visited, 0, sizeof(int) * (pl->counts[2] + 1));

    cursor = 1;
    for(i = 1; i <= pl->counts[2]; i++)  {
        flower_heads[i] = cursor;
        cursor += flower_heads[i] + 2;
    }

    flower_heads_backup = pl->flower_heads;
    pl->flower_heads = flower_heads;

    numComponents = 1;
    for(i = 1; i <= pl->counts[2]; i++)  {
        if(!visited[i])  {
            mark_all_neighbors_recursive(i, pl, visited, numComponents);
            numComponents++;
        }
    }

    componentList = (p_light **) malloc(sizeof(struct p_light *) * numComponents);

    free(flower_heads);
    free(visited);
    pl->flower_heads = flower_heads_backup;

    *componentListParm = componentList;
    return numComponents;
}

/*
 * mark_all_neighbors_recursive:  Helps find a connected component of the
 * given light packing by marking all the neighbors of the specified cell
 * as belong to component number compNo.  This function expects the
 * flower_heads field of pl to be filled in (this is not normally done when
 * a light packing is created).
 */

void mark_all_neighbors_recursive(int cell, struct p_light *pl, int *visited, int compNo)

{
    int i;
    int neighbor;

    if(visited[cell])  return;

    visited[cell] = compNo;
    for(i = 0; i <= pl->var_indices[pl->flower_heads[cell]]; i++)  {
        neighbor = pl->var_indices[pl->flower_heads[cell] + i + 1];
        mark_all_neighbors_recursive(neighbor, pl, visited, compNo);
    }
}
