#include "sa_types.h"
#include "sa_proto.h"

/* Routines for setting overlaps (this includes inversive distances). */

int set_xyz_overlaps(struct p_data *p,struct xyzPoint *xyz_list,
		     int count,char *datastr,int flag)
     /* Use xyz data to set overlaps of pack p. Allocate space if
necessary. 
This command is still evolving. Eventually, e.g., datastr may contain
list of overlaps to set; for now, all are set.

	flag=1: find max/min edge lengths in xyz data, 
	set overlaps as though circle radii were all min/2. 

	flag=2: for each vertex v find min edgelength, set
	as though radius of v was half that. 

Radii are NOT actually changed --- only setting overlaps.
Return count of overlaps set, 0 on error. */
{
  int i,j,k;
  double *tmprads,miN=1000000000.0,maX=0.0;
  double ovlp,dist,rad,ovlpmax=okerr,ovlpmin=1000;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;

  if (!p->status || count<=0 || !xyz_list
      || !alloc_overlaps(p)) return 0;

  if (flag==1) { /* treat as though all radii were miN/2. */

  /* compute maX/miN distances */
    for (i=1;i<=p->nodecount;i++)
    for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++)
      if ((k=pK_ptr[i].flower[j])>i
	  && i<= count && k<=count) {
	dist=xyz_dist(xyz_list[i], xyz_list[k]);
	miN = (dist<miN) ? dist : miN;
	maX = (dist>maX) ? dist : maX;
      }
    if (miN==1000000000.0 || maX==0.0 || (miN < okerr)) return 0; /* ?? */
    rad=miN/2.0; 
    /* compute and store overlaps */
    for (i=1;i<=p->nodecount;i++)
      for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++)
	if ((k=pK_ptr[i].flower[j])>i
	    && i<= count && k<=count ) {
	  ovlp=xyz_inv_dist(xyz_list[i],xyz_list[k],rad,rad);
	  set_single_overlap(p,i,j,ovlp);
	  ovlpmax = (ovlp>ovlpmax) ? ovlp : ovlpmax;
	  ovlpmin = (ovlp<ovlpmin) ? ovlp : ovlpmin;
	}
    sprintf(msgbuf,"Set overlaps using euclidean 3-space"
	    "distances.\n"
	    "  side lengths: max = %f, max/min = %f\n"
	    "  radius %f\n"
	    "  overlaps: max = %f, min =%f.",
	    (singledouble)maX,(singledouble)(maX/miN),
	    (singledouble)rad,(singledouble)ovlpmax,(singledouble)ovlpmin);
    msg();
    return 1;
  }
  if (flag==2) { /* set as though each vertex v radius was 1/2 the 
		    minimum eldg length from v. */
    /* compute temporary radii */
    tmprads=(double *)calloc((size_t)(p->nodecount+2),sizeof(double));
    for (i=1;i<=p->nodecount;i++) {
      miN=100000000;
      for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++) {
	if ((k=pK_ptr[i].flower[j])
	    && i<= count && k<=count ) {
	  dist=
	    sqrt((xyz_list[i].x-xyz_list[k].x)*(xyz_list[i].x-xyz_list[k].x)
		 +(xyz_list[i].y-xyz_list[k].y)*(xyz_list[i].y-xyz_list[k].y)
		 +(xyz_list[i].z-xyz_list[k].z)*(xyz_list[i].z-xyz_list[k].z));
	  miN= (dist<miN) ? dist : miN;
	}
      }
      tmprads[i]=miN/2.0;
    }
    /* compute and store overlaps */
    for (i=1;i<=p->nodecount;i++) {
      for (j=0;j<(pK_ptr[i].num+pK_ptr[i].bdry_flag);j++) {
	if ((k=pK_ptr[i].flower[j])>i
	    && i<= count && k<=count ) {
	  ovlp=xyz_inv_dist(xyz_list[i],xyz_list[k],tmprads[i],tmprads[k]);
	  set_single_overlap(p,i,j,ovlp);
	  ovlpmax = (ovlp>ovlpmax) ? ovlp : ovlpmax;
	  ovlpmin = (ovlp<ovlpmin) ? ovlp : ovlpmin;
	}
      }
    }
    sprintf(msgbuf,"Set overlaps using euclidean 3-space "
	    "distances and for each vert v, \n"
	    "  use 1/2 the length of the shortest edge from v .\n"
	    "  overlaps: max = %f, min =%f.",
	    (singledouble)ovlpmax,(singledouble)ovlpmin);
    msg();
    free(tmprads);
    return 1;
  }

  return 0;

  } /* set_xyz_overlaps */

int set_single_overlap(struct p_data *p,int v,int j,double angle)
     /* Set one overlap or inv. dist value in all approp places. 
Given vert v and index j of petal. Return 0 on error or if space 
not allocated. */
{
  int w,indx;
  struct K_data *pK_ptr;

  pK_ptr=p->packK_ptr;
  if (!p->overlap_status) return 0;
  w=p->packK_ptr[v].flower[j];
  indx=nghb(p,v,w);
  pK_ptr[v].overlaps[indx]=angle;
  if (indx==0 && !pK_ptr[v].bdry_flag) 
    pK_ptr[v].overlaps[pK_ptr[v].num]=angle;
  indx=nghb(p,w,v);
  pK_ptr[w].overlaps[indx]=angle;
  if (indx==0 && !pK_ptr[w].bdry_flag) 
    pK_ptr[w].overlaps[pK_ptr[w].num]=angle;
  return 1;
} /* set_single_overlap */
