#include "sa_types.h"
#include "sa_proto.h"


/* The 'plot_flag's in K-data and in f_data generally indicate 
   if a reasonable value for circle centers has been computed 
   when a packing is laid out; if zero or negative, the circle/face 
   will not be drawn in certain cases, this lets us override (generally, 
   plot_flag = 0 or 1). Default is for circles (-c or -v) and
   -f is for faces. */

int set_plot_flags(struct p_data *p,char *datastr)
{
  int i,hits,flag,cf_flag=0;
  char *nextpoint,*endptr,next[BUFSIZE];
  struct Vertlist *vertlist,*facelist,*trace;
  struct K_data *pK_ptr=p->packK_ptr;
  f_data *faces=p->faces;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) /* first check for -c, -v, or -f;
				      -c (or -v) for circles is default */
    return 0;
  if (*next=='-')
    {
      if (*(next+1)=='f') cf_flag=1; /* want faces */
      if (!grab_next(&nextpoint,next))
	return 0;
    }
  if (!sscanf(next,"%d",&flag)) return 0; /* didn't get a flag value */
  if (!cf_flag)
    {
      if (!(vertlist=Node_link_parse(p,nextpoint,&endptr,&hits,
				 &Vlist,&Elist,&Flist,
				 &region,pathlist,pathlength)))
	{
	  for (i=1;i<=p->nodecount;i++) pK_ptr[i].plot_flag=flag;
	  return 1;
	}
      trace=vertlist;
      while (trace)
	{
	  pK_ptr[trace->v].plot_flag=flag;
	  trace=trace->next;
	}
      vert_free(&vertlist);
      return 1;
    }
  else /* for faces */
    {
      if (!(facelist=Face_link_parse(p,nextpoint,&endptr,&hits,
				 &Vlist,&Elist,&Flist,
				 &region,pathlist,pathlength)))
	{
	  for (i=1;i<=p->facecount;i++) faces[i].plot_flag=flag;
	  return 1;
	}
      trace=facelist;
      while (trace)
	{
	  faces[trace->v].plot_flag=flag;
	  trace=trace->next;
	}
      vert_free(&facelist);
      return 1;
    }
} /* set_plot_flags */
