#include "sa_types.h"
#include "sa_proto.h"

/* slit complex along edge chain, which is created from list of vertices. 
Clone the vertices along the slit so a copy appears on each side of slit.
Chain may start at bdry or interior vert and can continue while still in
the interior; shouldn't disconnect complex.

Not very solid -- bad path may screw up complex. For interior cut,
need at least two edges. Lose overlap data. */

int slit_complex(struct p_data *p,struct Vertlist *vertlist)
{
  int ind,back,forward,newval,k,v,w,next,num,wnum,u;
  int *oldflower,*newflower;
  struct Vertlist *vtrace;
  struct Edgelist *edgelist=NULL,*trace;
  struct K_data *pK_ptr;
  struct R_data *pR_ptr;

  pK_ptr=p->packK_ptr;
  pR_ptr=p->packR_ptr;
  free_overlaps(p); /* fixup:want to save overlaps, if possible */
  /* make up edgelist */
  vtrace=vertlist;
  v=vtrace->v;
  if (vtrace->next!=NULL && nghb(p,v,(w=vtrace->next->v))>=0)
    edgelist=(struct Edgelist *)calloc((size_t)1,
				       sizeof(struct Edgelist));
  else return 0; /* no first edge */
  trace=edgelist;
  trace->v=v;
  trace->w=w;
  trace->next=NULL;
  v=w;
  vtrace=vtrace->next;
  while (vtrace->next!=NULL && nghb(p,v,(w=vtrace->next->v))>=0
	 && !pK_ptr[w].bdry_flag)
    /* stops if path disconnects or hits bdry */
    {
      trace->next=(struct Edgelist *)calloc((size_t)1,
					    sizeof(struct Edgelist));
      trace=trace->next;
      trace->v=v;
      trace->w=w;
      v=w;
      vtrace=vtrace->next;
    }
  /* use it */
  trace=edgelist;
  v=trace->v;
  next=trace->w;
  if (pK_ptr[v].bdry_flag && open_edge(p,v,next)==0) 
    {edge_free(&edgelist);return 0;}  /* failed to open first edge */
  else if (!pK_ptr[v].bdry_flag) 	/* opening interior edge */
    {
      if ((trace=trace->next)==NULL) /* need at least two edges to start */ 
	{edge_free(&edgelist);return 0;} 
      w=next;
      next=trace->w;
      /* three verts are (in order) v,w,next, slitting towards next*/
      /* need one new vert (w splits into two) */
      newval=p->nodecount+1;
      if (newval>p->sizelimit-1 && !alloc_pack_space(p,newval+1,1)) {
	sprintf(msgbuf,
		"Space allocation problem with adding vertex.");
	emsg();
	return 0;
      }
      pK_ptr=p->packK_ptr;
      pR_ptr=p->packR_ptr;
      /* fix v */
      ind=nghb(p,v,w);
      newflower=(int *)calloc((size_t)(pK_ptr[v].num+1),
			      sizeof(int));
      for (k=1;k<=pK_ptr[v].num;k++)
	newflower[k]=pK_ptr[v].flower[(k+ind) % pK_ptr[v].num];
      free(pK_ptr[v].flower);
      pK_ptr[v].flower=newflower;
      pK_ptr[v].flower[0]=newval;
      pK_ptr[v].bdry_flag=1;
      /* fix w */
      back=nghb(p,w,v);
      forward=nghb(p,w,next);
      num=(forward+pK_ptr[w].num-back) % (wnum=pK_ptr[w].num);
      oldflower=(int *)calloc((size_t)(pK_ptr[w].num+1),
			      sizeof(int));
      for (k=0;k<=pK_ptr[w].num;k++)
	oldflower[k]=pK_ptr[w].flower[k];
      newflower=(int *)calloc((size_t)(num+1),sizeof(int));
      for (k=0;k<=num;k++)
	newflower[k]=oldflower[(k+back) % wnum];
      free(pK_ptr[w].flower);
      pK_ptr[w].flower=newflower;
      pK_ptr[w].num=num;
      pK_ptr[w].bdry_flag=1;
      /* fix next */
      ind=nghb(p,next,w);
      newflower=(int *)calloc((size_t)(pK_ptr[next].num+1),
			      sizeof(int));
      for (k=0;k<pK_ptr[next].num;k++)
	newflower[k]=pK_ptr[next].flower
	  [(k+ind) % pK_ptr[next].num];
      free(pK_ptr[next].flower);
      pK_ptr[next].flower=newflower;
      pK_ptr[next].flower[pK_ptr[next].num]=newval;
      pK_ptr[next].bdry_flag=1;
      /* fix newval */
      num=(back+wnum-forward) % wnum;
      pK_ptr[newval].flower=(int *)calloc((size_t)(num+1),
				       sizeof(int));
      for (k=1;k<num;k++)
	pK_ptr[newval].flower[k]=oldflower[(k+forward) % wnum];
      free(oldflower);
      pK_ptr[newval].flower[0]=next;
      pK_ptr[newval].flower[num]=v;
      pK_ptr[newval].num=num;
      pK_ptr[newval].bdry_flag=1;
      for (k=1;k<num;k++)
	{
	  back=nghb(p,(u=pK_ptr[newval].flower[k]),w);
	  pK_ptr[u].flower[back]=newval;
	  if (back==0 && pK_ptr[u].flower[pK_ptr[u].num]==w)
	    pK_ptr[u].flower[pK_ptr[u].num]=newval;
	}
      pR_ptr[newval].rad=pR_ptr[w].rad;
      pR_ptr[newval].center=pR_ptr[w].center;
      p->nodecount++;
    } /* end of else */
  /* now, continue opening prescribed edges */
  while ((v=next) && (trace=trace->next)!=NULL
	 && open_edge(p,v,(next=trace->w)));
  complex_count(p,0);
  facedraworder(p,0);
  if (trace==NULL)
    {
      edge_free(&edgelist);
      return 1;
    }
  edge_free(&edgelist);
  return 0;
} /* slit_complex */
