#include "sa_types.h"
#include "sa_head.h"

#define SEPARATOR 1 /* separator symbol is ';' */
#define PACK 2
#define FLAG 3
#define STRING 4
#define INTEGER 5
#define FLOAT 6

StrTok *string_tokenizer(char *datastr,int *allocation,int *N)
  /* Break input string into tokens and put them in a vector.
   * Created here, but calling routine must do cleanup, so
   * allocation is important to keep track of. N reflects the
   * number of actual items found.  */
{
  int indx=0,i,j;
  size_t len;
  double val;
  char *nextpoint,*oldnext,next[BUFSIZE],c,*sepptr,*spot;
  StrTok *strTok=NULL;

  nextpoint=oldnext=datastr;
  stripsp(nextpoint);

  // first just count number of strings, add one for
  //   each separator.
  while (grab_next(&nextpoint,next)) indx++;
  if (!indx) return NULL;  	
  nextpoint=datastr;
  len=strlen(datastr);
  for (i=0;i<len;i++) 
  	if (datastr[i]==';') indx++;
  strTok=(StrTok *)calloc(indx+2,sizeof(StrTok));
  *allocation=indx+2;
  
  indx=0;
  while(grab_next(&nextpoint,next)) {
//printf("next is %s\n",next);
    /* first, look through oldnext for separator. */
    if ((sepptr=strchr(oldnext,';'))) {
      if (next[0]==';') { /* first character in next? */
	i=0;
	if (strTok[indx].key!=SEPARATOR) { /* current is separator? */
	  next[0]=' ';
	  stripsp(next);
	  strTok[indx].key=SEPARATOR;
	  indx++;
	  i=1;
	}
	while (next[i]==';') /* eat any contiguous separators */
	  next[i++]=' ';
	stripsp(next);
	if (!strlen(next)) goto CONTIN; /* if empty, new grab_next */
      }
      else if((spot=strchr(next,';'))) { 
	/* that separator we found is, in fact, in next; have
	   to terminate next, yet set up nextpoint so ';' is 
	   first after next grab_next (so its not lost).*/
	*spot='\0'; 
	nextpoint=sepptr;
      }
    }
    /* now continue processing 'next' */
    if (next[0]=='-') { /* flag or number */
      if (!strncmp(next,"-p",2)) {
	if (sscanf(next+2,"%d",&j)==1 && j>=0
	    && j<NUM_PACKS) { 
	  strTok[indx].key=PACK;
	  strTok[indx].intvalue=j;
	  indx++;
	}
	else goto CONTIN; /* error, skip and continue parsing */
      }

      /* Is it a number? integer or floating point? */
      else if (((c=next[1])=='.') || ((c>='0') && (c<='9'))) {
	if (sscanf(next,"%d",&j)) { /* integer */
//printf("- branch: integer? j=%d\n",j);
	  strTok[indx].key=INTEGER;
	  strTok[indx].intvalue=j;
	  indx++;
	}
	else if (sscanf(next,"%lf",&val)) { /* float */
//printf("- branch: float? val=%lf\n",val);
	  strTok[indx].key=FLOAT;
	  strTok[indx].floatvalue=val;
	  indx++;
	}
	else { /* seems to be non-numeric, treat as string */
//printf("- branch: string? 1. str=%s\n",next);
	  strTok[indx].key=STRING;
	  size_t len=strlen(next)+1;
	  len = (len>(size_t)BUFSIZE) ? (size_t)(BUFSIZE-1) : len;
	  strTok[indx].str=(char *)calloc(len,sizeof(char));
	  strncpy(strTok[indx].str,next,len);
	  indx++;
	}
      }
      /* else, must be a flag indicator: drop the '-'. */
      else {
//printf("- branch: drop the -\n");
 	strTok[indx].key=FLAG;
	next[0]=' ';
	stripsp(next);
	size_t len=strlen(next);
//sprintf(msgbuf,"got flag key: length=%d, next = %s\n",len,next);
//debugmsg();
	len = (len>(size_t)BUFSIZE) ? (size_t)(BUFSIZE-1) : len;
	if (!len) { /* don't want zero length */
	  len=1; 
	  strTok[indx].str=(char *)calloc(len,sizeof(char));
	  strTok[indx].str[0]=' ';
	}
	else {
	  strTok[indx].str=(char *)calloc(len,sizeof(char));
	  strncpy(strTok[indx].str,next,len);
	}
	indx++;
      }
    } /* end for precessing with initial '-' */
    else { /* could be integer, float, or a string */
//sprintf(msgbuf,"enter 'else' after '-' processing: next=%s",next);
//debugmsg();
      if (((c=next[0])=='.') || ((c>='0') && (c<='9')) || c=='+') { 
	if (sscanf(next,"%d",&j)) { /* integer */
//printf("integer? j=%d\n",j);
	  strTok[indx].key=INTEGER;
	  strTok[indx].intvalue=j;
	  indx++;
	}
	else if (sscanf(next,"%lf",&val)) { /* float */
//printf("float? val=%lf\n",val);
	  strTok[indx].key=FLOAT;
	  strTok[indx].floatvalue=val;
	  indx++;
	}
	else { /* take as a string */
//printf("string?  str=%s\n",next);
	  strTok[indx].key=STRING;
	  size_t len=strlen(next)+1;
	  len = (len>(size_t)BUFSIZE) ? (size_t)(BUFSIZE-1) : len;
	  strTok[indx].str=(char *)calloc(len,sizeof(char));
	  strncpy(strTok[indx].str,next,len);
	  indx++;
	}
      }
      else { /* is a string */
//printf("final try: string? str=%s\n",next);
	strTok[indx].key=STRING;
	size_t len=strlen(next)+1;
	len = (len>(size_t)BUFSIZE) ? (size_t)(BUFSIZE-1) : len;
	strTok[indx].str=(char *)calloc(len,sizeof(char));
	strncpy(strTok[indx].str,next,len);
	indx++;
      }
    }
  CONTIN:
    oldnext=nextpoint; /* need 'oldtext' to search for ';' */
  } /* end of while, go do another grab_next */
  *N=indx;
  if (indx>0) return strTok;
  return NULL;
} /* string_parse */

/**
 * If token is a flag, and flag is given character, return 1; 
 * if it is a flag, but is NOT the give character, return -1; 
 * if not a flag, return 0.
 */
int tok_is_flag(StrTok tok,char c) {
	if (tok.str && tok.str[0]==c) return 1;
	if (tok.key==FLAG) return -1;
	else return 0;
} /* tok_is_flag */

/**
 * Put all tokens including start, up to (not including) end or
 * the next FLAG token into space-separated string, converting 
 * integers, floats, etc. Limit size to 256.
 */
char *flag_to_flag(StrTok *token,int start,int end) {
	int n=start;
	if (!token || start>=end) return NULL;
	char *format;
	char *result=(char *)calloc(280,sizeof(char));
	result[0]='\0';
	int hit=0;
	
	while (n<end && token[n].key!=FLAG && token[n].key!=SEPARATOR
		&& token[n].key!=PACK && strlen(result)<256) {
		if(token[n].key==STRING) {
			sprintf(format,"%%s%d ",256-strlen(result));
			if (token[n].str) sprintf(result,format,token[n].str);
		}
		else if (token[n].key==INTEGER) sprintf(result,"%d ",token[n].intvalue);
		else if (token[n].key==FLOAT) sprintf(result,"%lf ",token[n].floatvalue);
		n++;
		hit++;
	}
	if (hit) return result;
	return NULL;
} /* flag_to_flag */
