#include "sa_types.h"
#include "sa_proto.h"

/* write 2D or 3D faces in "*.jvx" form for display by JavaView. 
See 'write_carrier.cpp' for output (of all faces) in "common OFF 
(Object File Format)" for use by Geomview. */

int surf_write(struct p_data *p,char *datastr)
{
  int i,j,fcount=0,ccount=0,col,*new2old,hits,xyzflag=0;
  int v,f,v0,v1,v2,*flist=NULL;
  double V[3];
  complex ctr;
  struct tm *t;
  time_t clicks=time(NULL);
  char *nextpoint,next[BUFSIZE],*endptr;
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;
  struct Vertlist *facelist=NULL,*ftrace;
  FILE *fpout=NULL;

  nextpoint=datastr;
  stripsp(nextpoint);
  if (!strlen(nextpoint)) return 0;
  while (nextpoint[0]=='-') { /* read flags */
    grab_next(&nextpoint,next);
    stripsp(nextpoint);
    if (!strncmp(next,"-3D",3)) { /* use 3D xyzpoint data */
      if (!p->xyzpoint){ 
	sprintf(msgbuf,"Usage: surf_write. No 3D data available.");
	emsg();
	return 0;
      }
      xyzflag=1; 
    }
    else if (!strncmp(next,"-o",2)) { /* output filename */
      if (!grab_next(&nextpoint,next)
	  || !(fpout=fopen(next,"w"))) {
	sprintf(msgbuf,"Usage: surf_write. Didn't get filename.");
	emsg();
	return 0;
      }
    }
  }

  if (!fpout) { /* use default outputfile */
    sprintf(next,"surf_write.jvx");
    fpout=fopen(next,"w");
    sprintf(msgbuf,"surf_write output to file %s.",next);
    msg();
  }
  
  /* rest of command line should specify face list; default to 'all' */

  if (!strlen(nextpoint)) strcpy(next,"a\0");
  if (!(facelist=Face_link_parse(p,nextpoint,&endptr,&hits,
	&Vlist,&Elist,&Flist,&region,pathlist,pathlength))) {
    sprintf(msgbuf,"surf_write: no facelist was obtained.");
    emsg();
    return 0;
  }

  /* have to mark qualifying faces and vertices */
  for (i=1;i<=p->nodecount;i++) pK_ptr[i].util_flag=-1;
  flist=(int *)calloc((size_t)(p->facecount+1),sizeof(int));
  ftrace=facelist;
  while (ftrace) {
    f=ftrace->v;
    v0=p->faces[f].vert[0];
    v1=p->faces[f].vert[1];
    v2=p->faces[f].vert[2];
    if (!flist[f] && (xyzflag ||
	  (pK_ptr[v0].plot_flag
	  && pK_ptr[v1].plot_flag
	  && pK_ptr[v2].plot_flag
	  && !isnan(pR_ptr[v0].center.re) && !isnan(pR_ptr[v0].center.im)
	  && !isnan(pR_ptr[v1].center.re) && !isnan(pR_ptr[v1].center.im)
	  && !isnan(pR_ptr[v2].center.re) && !isnan(pR_ptr[v2].center.im)))) {
      pK_ptr[v0].util_flag=1;
      pK_ptr[v1].util_flag=1;
      pK_ptr[v2].util_flag=1;
      flist[f]=1;
      fcount++;
    }
    ftrace=ftrace->next;
  }
  vert_free(&facelist);
      
  new2old=(int *)calloc((size_t)(p->nodecount+1),sizeof(int));
  for (i=1;i<=p->nodecount;i++) /* store new/old indices for target circles */
    if (pK_ptr[i].util_flag>=0) {
      pK_ptr[i].util_flag=ccount;
      new2old[ccount++]=i;
    }

/* ---- writing the file */

  /* heading stuff */
  fprintf(fpout,"<\?xml version=\"1.0\" encoding=\"ISO-8859-1\""
	  " standalone=\"no\"\?>\n");
  fprintf(fpout,"<!DOCTYPE jvx-model SYSTEM \"http://www.javaview.de/"
	  "rsrc/jvx.dtd\">\n");
  fprintf(fpout,"<jvx-model>\n"
	  "  <meta generator=\"CirclePack\"/>\n");
  t=localtime(&clicks); /* get time/date */
  strftime(buff,256,"%c",t);
  fprintf(fpout,"  <meta date=\"%s\" />\n",buff);
  fprintf(fpout,"  <description>\n    circle packing carrier\n"
	  "  </description>\n");

  fprintf(fpout,
	  "  <geometries>\n"
	  "    <geometry name=\"carrier\">\n"
	  "      <pointSet dim=\"3\""
	  " point=\"hide\" color=\"hide\">\n"
	  "        <points num=\"%d\">\n",ccount);

/* now cycle through the points */
  for (i=0;i<ccount;i++) {
    v=new2old[i];
    if (xyzflag) { /* using xyzpoint data */
      fprintf(fpout,"          <p>%lf %lf %lf</p>\n",
	      (singledouble)p->xyzpoint[v].x,
	      (singledouble)p->xyzpoint[v].y,
	      (singledouble)p->xyzpoint[v].z);
    }
    else if (p->hes>0) { /* spherical case */
      s_pt_to_vec(p->packR_ptr[v].center,V);
      fprintf(fpout,"         <p>%lf %lf %lf</p>\n",
	      (singledouble)V[0],(singledouble)V[1],
	      (singledouble)V[2]);
    }
    else if (p->hes<=0) { /* eucl case */
      ctr=p->packR_ptr[v].center;
      fprintf(fpout,"          <p>%lf %lf 0.0</p>\n",
	      (singledouble)ctr.re,(singledouble)ctr.im);
      }
  }
  fprintf(fpout,
	  "          <thickness>1.0</thickness>\n"
	  "          <color type=\"rgb\">255 0 0</color>\n"
	  "          <colorTag type=\"rgb\">255 0 255</colorTag>\n"
	  "        </points>\n"
	  "      </pointSet>\n");
  
  /* now add the faces; recall that indices start at 0. */
  fprintf(fpout,"      <faceSet face=\"show\" color=\"show\" edge=\"show\">\n"
	  "        <faces num=\"%d\">\n",fcount);
  for (f=1;f<=p->facecount;f++) {
    if (flist[f]) {
      v0=pK_ptr[p->faces[f].vert[0]].util_flag;
      v1=pK_ptr[p->faces[f].vert[1]].util_flag;
      v2=pK_ptr[p->faces[f].vert[2]].util_flag;
      fprintf(fpout,"          <f>%d %d %d</f>\n",v0,v1,v2);
    }
  }
  fprintf(fpout,
	  "        </faces>\n"
	  "        <colors type=\"rgb\" num=\"%d\">\n",fcount);
  for (f=1;f<=p->facecount;f++) {
    if (flist[f]) {
      col=p->faces[f].color;
      fprintf(fpout,"        <c> %lf %lf %lf </c>\n",
	      (singledouble)((double)red[col]/255),
	      (singledouble)((double)green[col]/255),
	      (singledouble)((double)blue[col]/255));
    }
  }
  fprintf(fpout,
	  "        </colors>\n"
      	  "      </faceSet>\n"
	  "    </geometry>\n"
	  "  </geometries>\n"
	  "</jvx-model>\n");
  fclose(fpout);
  free(new2old);
  free(flist);
  return fcount;
} /* surf_write */
