#include "sa_types.h"
#include "sa_proto.h"

/* update the mobius transformations associated with side pairings
   in multiply connected cases. Return 0 if errors occur in the
   process -- particularly arith errors in routines of mobius_math.c */

int update_pair_mob(struct p_data *p)
{
  int j,length=0,err_flag=0;

  if (!p->redfaces) return 0;
  while (p->edge_pair[length+1].edge) length++;
  if (length>1)
    for (j=1;j<=length;j++)
      if (!pair_mobius(p,j)) 
	err_flag++;
  if (err_flag) return 0;
  return 1;
} /* update_pair_mob */

/**
 * Return index in p->edge_pair for mate to edge with index e. 
 * Return -1 on failure. (Note: indexes start at 1.)
 */
int find_mate_index(struct p_data *p,int e) 
{
	int i;
    int length=0;
    for (i=1;p->edge_pair[i].edge;i++) length=i;
     
	if (e<1 || e>length || !p->edge_pair[e].edge || !p->edge_pair[e].mate) return -1;
	for (i=1;i<=length;i++) 
	  if (p->edge_pair[i].edge==p->edge_pair[e].mate
		&& p->edge_pair[i].edge_indx==p->edge_pair[e].mate_indx) 
		return i;
	return -1;
} /* find_mate */
	
	