#include "sa_types.h"
#include "sa_proto.h"

/* write carrier or xyzpoint data in "common OFF (Object File Format)" 
form for use, e.g., by Geomview, or in "*.jvx" form for display by
JavaView. */

/* NOTE: This is partly deprecated. I've written routine 'surf_write' 
to allow option of specifying faces. That output is only in *.jvx. */

int write_carrier_file(struct p_data *p,char *datastr)
{
  int i,j,fcount=0,jvxflag=0,ccount=0,col,*newnum,vert[3],xyzflag=0;
  double V[3];
  complex ctr;
  struct tm *t;
  time_t clicks=time(NULL);
  char *nextpoint,next[BUFSIZE];
  struct K_data *pK_ptr=p->packK_ptr;
  struct R_data *pR_ptr=p->packR_ptr;
  FILE *fpout=NULL;

  nextpoint=datastr;
  if (!grab_next(&nextpoint,next)) return 0;
  while (next[0]=='-') { /* read any flags */
    if (!strncmp(next,"-3D",3) && p->xyzpoint){ /* use 3D xyzpoint data */
      xyzflag=1; 
    }
    else if (!strncmp(next,"-jvx",4)) { /* use Javaview jvx format */
      jvxflag++;
    }
    /* .. grab_next to get  */
    if (!grab_next(&nextpoint,next)) return 0; /* should get filename */
  }

  if (!(fpout=fopen(next,"w")))  {
    sprintf(msgbuf,"write_carrier: failed to open %s for output.",next);
    emsg();
    return 0;
  }

  newnum=(int *)calloc((size_t)(p->nodecount+1),sizeof(int));
  for (i=1;i<=p->nodecount;i++) { /* store new index for plotted circles */
    if (xyzflag || pK_ptr[i].plot_flag 
	|| !isnan(pR_ptr[i].center.re) || !isnan(pR_ptr[i].center.im))
      newnum[i]=ccount++;
    else newnum[i]=-1;
  }

  if (!xyzflag) { /* use carrier locations; use only plotted circles */ 
    if((i=p->nodecount-ccount)) {
      sprintf(msgbuf,"write_carrier: %d vertices w/o plot_flag set.",i);
      msg();
    }
    for (i=1;i<=p->facecount;i++) {
      for (j=0;j<3;j++) vert[j]=p->faces[i].vert[j];
      if (p->faces[i].plot_flag
	  && newnum[vert[0]]>=0 && newnum[vert[1]]>=0 && newnum[vert[2]]>=0)
	fcount++;
    }
  }
  else { /* use xyzpoint data */
    ccount=p->nodecount;
    fcount=p->facecount;
  }

  if (!jvxflag) { /* want 'OFF' format */
    fprintf(fpout,"OFF\n%d %d %d\n",ccount,fcount,
	    ccount+fcount-1); /* -1 due to indexing from 0 vs 1 */
    if (xyzflag) { /* using xyzpoint data */
      for (i=1;i<=p->nodecount;i++) {
	fprintf(fpout,"%lf %lf %lf\n",
		(singledouble)p->xyzpoint[i].x,
		(singledouble)p->xyzpoint[i].y,
		(singledouble)p->xyzpoint[i].z);
      }
    }
    else if (p->hes>0) { /* spherical case */
      for (i=1;i<=p->nodecount;i++) /* store center, r/g/b/a */
	if (newnum[i]>=0) {
	  s_pt_to_vec(p->packR_ptr[i].center,V);
	  fprintf(fpout,"%lf %lf %lf\n",
		  (singledouble)V[0],(singledouble)V[1],
		  (singledouble)V[2]);
	}
    }
    else if (p->hes<=0) { /* eucl case */
      for (i=1;i<=p->nodecount;i++) /* centers (x_3=0), r/g/b/a */
	if (newnum[i]>=0) {
	  ctr=p->packR_ptr[i].center;
	  fprintf(fpout,"%lf %lf 0.0\n",
		  (singledouble)ctr.re,(singledouble)ctr.im);
	}
    }
    fprintf(fpout,"\n");

    /* now add the faces (with plot_flags); have to recall that
       indices start at 0 in the *.off file face output. */
    for (i=1;i<=p->facecount;i++) {
      for (j=0;j<3;j++) vert[j]=p->faces[i].vert[j];
      if (xyzflag 
	  || (p->faces[i].plot_flag
	      && newnum[vert[0]]>=0 && newnum[vert[1]]>=0 
	      && newnum[vert[2]]>=0))
	fprintf(fpout,"3   %d %d %d",
		newnum[p->faces[i].vert[0]],
		newnum[p->faces[i].vert[1]],
		newnum[p->faces[i].vert[2]]);
      col=p->faces[i].color;
      fprintf(fpout,"  %lf %lf %lf 1.0 \n",
	      (singledouble)((double)red[col]/255),
	      (singledouble)((double)green[col]/255),
	      (singledouble)((double)blue[col]/255));
    }
    fclose(fpout);
    free(newnum);
    return ccount;
  }

  /* output in jvx format */

  /* heading stuff */
  fprintf(fpout,"<\?xml version=\"1.0\" encoding=\"ISO-8859-1\""
	  " standalone=\"no\"\?>\n");
  fprintf(fpout,"<!DOCTYPE jvx-model SYSTEM \"http://www.javaview.de/"
	  "rsrc/jvx.dtd\">\n");
  fprintf(fpout,"<jvx-model>\n"
	  "  <meta generator=\"CirclePack\"/>\n");
  t=localtime(&clicks); /* get time/date */
  strftime(buff,256,"%c",t);
  fprintf(fpout,"  <meta date=\"%s\" />\n",buff);
  fprintf(fpout,"  <description>\n    circle packing carrier\n"
	  "  </description>\n");

  fprintf(fpout,
	  "  <geometries>\n"
	  "    <geometry name=\"carrier\">\n"
	  "      <pointSet dim=\"3\""
	  " point=\"show\" color=\"show\">\n"
	  "        <points num=\"%d\">\n",ccount);

/* now cycle through the points */
  if (xyzflag) { /* using xyzpoint data */
    for (i=1;i<=p->nodecount;i++) {
      fprintf(fpout,"          <p>%lf %lf %lf</p>\n",
		(singledouble)p->xyzpoint[i].x,
		(singledouble)p->xyzpoint[i].y,
		(singledouble)p->xyzpoint[i].z);
      }
  }
  else if (p->hes>0) { /* spherical case */
    for (i=1;i<=p->nodecount;i++) /* store center, r/g/b/a */
      if (newnum[i]>=0) {
	s_pt_to_vec(p->packR_ptr[i].center,V);
	fprintf(fpout,"         <p>%lf %lf %lf</p>\n",
		(singledouble)V[0],(singledouble)V[1],
		(singledouble)V[2]);
      }
  }
  else if (p->hes<=0) { /* eucl case */
    for (i=1;i<=p->nodecount;i++) /* centers (x_3=0), r/g/b/a */
      if (newnum[i]>=0) {
	ctr=p->packR_ptr[i].center;
	fprintf(fpout,"          <p>%lf %lf 0.0</p>\n",
		(singledouble)ctr.re,(singledouble)ctr.im);
      }
  }
  fprintf(fpout,
	  "          <thickness>1.0</thickness>\n"
	  "          <color type=\"rgb\">255 0 0</color>\n"
	  "          <colorTag type=\"rgb\">255 0 255</colorTag>\n"
	  "        </points>\n"
	  "      </pointSet>\n");
  
/* now add the faces; recall that indices start at 0. */
  fprintf(fpout,"      <faceSet face=\"show\" edge=\"show\">\n"
	  "        <faces num=\"%d\">\n",fcount);
  for (i=1;i<=p->facecount;i++) {
    for (j=0;j<3;j++) vert[j]=p->faces[i].vert[j];
    if (xyzflag 
	|| (p->faces[i].plot_flag
	    && newnum[vert[0]]>=0 && newnum[vert[1]]>=0 
	    && newnum[vert[2]]>=0))
      fprintf(fpout,"          <f>%d %d %d</f>\n",
		newnum[p->faces[i].vert[0]],
		newnum[p->faces[i].vert[1]],
		newnum[p->faces[i].vert[2]]);
  }
  fprintf(fpout,
	  "        </faces>\n"
	  "        <colors type=\"rgb\" num=\"%d\">\n",fcount);
  for (i=1;i<=p->facecount;i++) {
    for (j=0;j<3;j++) vert[j]=p->faces[i].vert[j];
    if (xyzflag 
	|| (p->faces[i].plot_flag
	    && newnum[vert[0]]>=0 && newnum[vert[1]]>=0 
	    && newnum[vert[2]]>=0)) {
      col=p->faces[i].color;
      fprintf(fpout,"        <c> %lf %lf %lf </c>\n",
	      (singledouble)((double)red[col]/255),
	      (singledouble)((double)green[col]/255),
	      (singledouble)((double)blue[col]/255));
    }
  }
  fprintf(fpout,
	  "        </colors>\n"
      	  "      </faceSet>\n"
	  "    </geometry>\n"
	  "  </geometries>\n"
	  "</jvx-model>\n");
  fclose(fpout);
  free(newnum);
  return ccount;
} /* write_carrier_file */

