#include "sa_types.h"
#include "sa_proto.h"

/* Write 'light' pack data to open filepointer fp. If flag set,
then want 'raw' data, ie., no keywords, etc. Return 0 on error. */

extern char buf[];

int write_light_as_full(FILE *fp,struct p_light *pl,int flag)
{
  int i,j,n,tick,digits,count,vert=1,petalNo,start,nghb_flag,nextPetal;
  int *flower;
  int flowersize;
  char format_buf[32];
  int *numInternal, *numTotal;
  int *flower_heads;

  //fprintf(stderr, "doing write_light on pl = %p with\n"
                  //"pl->counts[0] = %d\n"
                  //"pl->counts[1] = %d\n"
                  //"pl->counts[2] = %d\n"
                  //"pl->counts[3] = %d\n"
                  //"pl->counts[4] = %d\n"
                  //"pl->counts[5] = %d\n",
                   //pl, pl->counts[0], pl->counts[1], pl->counts[2],
                       //pl->counts[3], pl->counts[4], pl->counts[5]);

  count = pl->counts[2];
  if (!flag)
    {
      fprintf(fp,"NODECOUNT:   %d\n",pl->counts[2]);
      if (pl->counts[1]<0) sprintf(buf,"hyperbolic");
      else if (pl->counts[1]>0) sprintf(buf,"spherical");
      else sprintf(buf,"euclidean");
      fprintf(fp,"\nGEOMETRY:   %s\n",buf);
    }
  else for (n=0;n<12;n++) fprintf(fp,"%d ",pl->counts[n]);
     /* note: only first 6 spots currently in use. */

  fprintf(fp,"\n");
  if (!flag)
    fprintf(fp,"FLOWERS: \n"); /* interior flowers */

  numInternal = (int *) malloc(sizeof(int) * (pl->counts[2] + 1));
  memset(numInternal, 0, sizeof(int) * (pl->counts[2] + 1));
  numTotal = (int *) malloc(sizeof(int) * (pl->counts[2] + 1));
  memset(numTotal, 0, sizeof(int) * (pl->counts[2] + 1));
  flower_heads = (int *) malloc(sizeof(int) * (pl->counts[2] + 1));
  memset(flower_heads, 0, sizeof(int) * (pl->counts[2] + 1));

  tick=1;
  for(n = 1; n <= pl->counts[2]; n++)
    {
      flower_heads[n] = tick;
      numTotal[n]=pl->var_indices[tick++];
      for (i=0;i<=numTotal[n];i++)  {
        if(pl->var_indices[tick+i] <= pl->counts[2])  {
          numInternal[n]++;
          //fprintf(stderr, "flower %3d: petal %3d (vertex %3d) is %s\n",
                  //vert - 1, i, pl->var_indices[tick + i],
                 //(pl->var_indices[tick + i] <= pl->counts[2]) ? "internal" : "external");
        }
      }
      //fprintf(stderr, "flower %3d: done and found %3d interior petals\n", vert - 1, numInternal[n]);
      tick += (numTotal[n] + 1);
    }

  flowersize = -1;
  flower = NULL;

  for (n = 1; n <= pl->counts[2]; n++) 
    {
      if (!flag) fprintf(fp,"\n%d ",vert++);
      if(numTotal[n] > flowersize)  {
          flowersize = numTotal[n];
          //fprintf(stderr, "growing flower to new size of %d\n", flowersize + 1);
          flower = (int *) realloc(flower, sizeof(int) * (flowersize + 1));
      }
      memset(flower, 0, sizeof(int) * flowersize);
      petalNo = 0;
      for (i=0;i<=numTotal[n];i++)  {
        if(pl->var_indices[flower_heads[n]+i+1] <= pl->counts[2])  {
          flower[petalNo] = pl->var_indices[flower_heads[n]+i+1];
          petalNo++;
        }
      }
      if(petalNo != numInternal[n])  fprintf(stderr, "warning: mismatch in count of #petals on flower #%d (petalNo is %d and numInternal is %d)\n", n, petalNo,  numInternal[n]);

      if(numTotal[n] != numInternal[n] - 1)  { /* external node */
        if(numInternal[n] != 1 && flower[0] == flower[numInternal[n] - 1])
          fprintf(fp, "%d   ", numInternal[n] - 2);
        else
          fprintf(fp, "%d   ", numInternal[n] - 1);

        /* Find the two consecutive petals in the list that are not neighbors of one another.
           The second of this pair should be the first petal given in this vertex's flower. */
        start = -1;
        for(i = 0; i < numInternal[n]; i++)  {
          nghb_flag = 0;
          if(flower[i] > pl->counts[2])  {
            fprintf(stderr, "on flower #%d: petal %d (flower %d) is out of range, skipping\n", n, i, flower[i]);
            continue;
          }
          nextPetal = (i + 1) % numInternal[n];
          //if(pl->var_indices[flower_heads[flower[i]]] > 9)  fprintf(stderr, "on flower #%d: petal %d (flower %d) has %d petals: what gives?\n", n, i, flower[i], pl->var_indices[flower_heads[flower[i]]]);
          for(j = 1; j <= pl->var_indices[flower_heads[flower[i]]] + 1; j++)  {
            if(pl->var_indices[flower_heads[flower[i]]+j] == flower[nextPetal])  {
              if(nextPetal == i)  continue;
              //fprintf(stderr, "on flower #%d: petal %d (flower %d) is the %dth petal of petal %d (flower %d)\n", n, nextPetal, flower[nextPetal], j, i, flower[i]);
              nghb_flag = 1;
              break;
            }
          }
          if(!nghb_flag)  {
            start = nextPetal;
            break;
          }
        }
        if(start == -1)  {
          //fprintf(stderr, "warning: flower %d is external but all its consecutive petals are neighbors, using petal 0 as start\n", n);
          //exit(1);
          start = 0;
        }
        //fprintf(stderr, "on flower #%d: using petal %d (flower %d) as start petal\n", n, start, flower[start]);
        
        for(i = 0; i < numInternal[n]; i++)  {
          j = (start + i) % numInternal[n];
          if(j == 0 || flower[j] != flower[0])
            fprintf(fp,"%d ",flower[j]);
        }
      }
      else  { /* internal node */
        //fprintf(stderr, "on flower #%d: using petal %d (flower %d) as start petal since flower is internal\n", n, start, flower[start]);
        fprintf(fp, "%d   ", numInternal[n] - 1);
        for(i = 0; i < numInternal[n]; i++)
          fprintf(fp,"%d ",flower[i]);
      }

    }
  free(flower);
  free(flower_heads);

  digits=1;
  while((pow(10.,(double)digits))*toler < 0.1 && digits<MAX_ACCUR)
    digits++;
  sprintf(format_buf,"%% .%de ",digits);

  fprintf(fp,"\n");
//  if (!flag) 
//    fprintf(fp,"\nORIG_INDICES: \n  "); /* original packing indices */
//  for (i=1;i<=count;i++)
//    {
//      fprintf(fp,"%d  ",pl->orig_indices[i]);
//      if (!flag && (i % 10)==0) fprintf(fp,"\n");
//    }

//  fprintf(fp,"\n");
  if (!flag) 
    fprintf(fp,"\nRADII:\n"); /* radii for all */

  for (i=1;i<=count;i++)
    {
      fprintf(fp,format_buf,(singledouble)radius_pl(pl,i));
      if (!flag && (i % 4)==0) fprintf(fp,"\n");
    }
//  fprintf(fp,"\n");

//  if (pl->counts[5]) /* have non-default aims? */
//    {
//      fprintf(fp,"\n");
//      if (!flag) fprintf(fp,"\nAIM_INDICES: %d \n",pl->counts[5]);
//      for (i=1;i<=pl->counts[5];i++)
//	fprintf(fp,"%d  ",pl->aim_indices[i]);
//      fprintf(fp,"\n");
//      if (!flag) fprintf(fp,"\nAIMS:\n");
//      for (i=1;i<=pl->counts[5];i++)
//	fprintf(fp,format_buf,(singledouble)pl->aims[i]);
//    }

  if (!flag) fprintf(fp,"\n\nEND\n");

  return 1;
} /* write_light_as_full */
