#include "sa_types.h"
#include "sa_proto.h"

/* Write out file with triples of triangle vertices from a packing;
to be used as input to other programs. File format:

N 1
u1 v1 w1
u2 v2 w2
...
uN vN wN

where N is the total number of triangles, 1 indicates triangles (versus
other shapes), and then a list of N triples. datastr is the file
name, including path. Return 0 on error. */

int write_tri_verts(struct p_data *p,char *datastr)
{
  int i;
  char filename[NAME_MAX];
  FILE *fp;

  if (!p->status || !p->faces || !(sscanf(datastr,"%s",filename)))
      return 0;
  if (!(fp=fopen(filename,"a")) )
    {
      sprintf(msgbuf,"tri_verts failed to open %s for appending.",
	      filename);
      msg();
      return 0;
    }
  fprintf(fp,"%d 1\n",p->nodecount);
  for (i=1;i<=p->nodecount;i++)
    fprintf(fp,"%d %d %d\n", p->faces[i].vert[0],
	    p->faces[i].vert[1],p->faces[i].vert[2]);
  fclose(fp);
  sprintf(msgbuf,"Vertices of %d triangles recorded in %s.",
	  p->nodecount,filename);
  msg();
  return 1;
} /* write_tri_verts */
